/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import <appkit/appkit.h>
#import "PeripheralCard.h"
#import <stdio.h>
#import <mach/cthreads.h>

#define DISK_MAGIC_NUMBER 0x04136359   /* This means file has a valid header.*/
                                       /* If not found, it's Prodos blocks.  */
#define NUM_TRACKS         40        /* Normally 35, but some drives had 40. */
#define NUM_HALF_TRACKS    (2 * NUM_TRACKS)
#define NUM_QUARTER_TRACKS (4 * NUM_TRACKS)
#define SECTORS_PER_TRACK  16
#define BYTES_PER_SECTOR   256
#define BYTES_PER_TRACK (BYTES_PER_SECTOR * SECTORS_PER_TRACK)

typedef enum {
  DISK_NO_DISK   = 0,   /* For internal use only; do not use in a file! */
  DISK_PRODOS    = 1,   /* Do not change these numbers, as they are     */
  DISK_DOS33     = 2,   /* present in files which already exist.  If    */
  DISK_RAW_BYTES = 3,   /* you wish to add a type, choose a new number. */
} DiskType;

enum ProdosDriverCall {
  PRODOS_STATUS = 0,
  PRODOS_READ   = 1,
  PRODOS_WRITE  = 2,
  PRODOS_FORMAT = 3
};

enum DosRWTSCall {
  DOS_SEEK   = 0,
  DOS_READ   = 1,
  DOS_WRITE  = 2,
  DOS_FORMAT = 4
};

enum DosErrCode {
  DOS_ERR_NO_ERROR        = 0,
  DOS_ERR_INITIALIZATION  = 0x8,
  DOS_ERR_WRITE_PROTECTED = 0x10,
  DOS_ERR_VOLUME_MISMATCH = 0x20,
  DOS_ERR_IO_ERROR        = 0x40
};


enum ProdosErrCode {
  PRODOS_ERR_NO_ERROR            = 0,
  PRODOS_ERR_BAD_CALL            = 1,
  PRODOS_ERR_IO_ERROR            = 0x27,
  PRODOS_ERR_NO_DEVICE_CONNECTED = 0x28,
  PRODOS_ERR_WRITE_PROTECTED     = 0x2B
};

typedef struct {
  DiskType type;        /* Specifies the type of disk.                  */
  char diskname[64];    /* Name of disk; /tmp/foo.a2disk -> foo, etc.   */
  char diskpath[MAXPATHLEN + 1];  /* Path of original disk. */
  BOOL dirty;           /* Set to true iff disk is modified.            */
  int fd;               /* File descriptor for the file.                */
  NXStream *stream;     /* Stream for this disk file.                   */
  BOOL write_protected; /* 'nuff said.                                  */
  union {
    struct {
      long num_blocks;          /* Total # of blocks on the disk.            */
      long current_block;       /* Current block being worked on (needed?)   */
      long first_block_offset;  /* File offset where the first block begins. */
    } prodos;
    struct {
      int num_tracks;           /* Either 35 or 40. */
      int current_sector;
      long first_track_offset;  /* File offset where first track begins. */
    } dos33;
    struct {
      int num_tracks;                  /* Either 35 or 40.                 */
      struct {
	int track_length;           /* Length of this track (0 == no track) */
	unsigned long track_offset; /* File offset where this track begins. */
      } track[NUM_QUARTER_TRACKS];

      /* Housekeeping stuff for raw mode. */
      unsigned long last_read_cycles;  /* Cycle count at last read.        */
      int head_position;               /* "bytes" into current track.      */
    } raw_bytes;
  } u;
  void *auxdata;  /* This can be anything. */
} Disk;


/* This information really only makes sense for a Disk ][. */
typedef struct {
  int qtrack;      // Current quartertrack number.
  int cached_qtrack_number;
  unsigned char *cached_qtrack_data;
  BOOL magnet[4];  // State of the stepper motor magnets.
  BOOL on;         // Whether or not this drive is spinning.
  BOOL write_mode; // Whether or not we are in write mode.
  unsigned char write_register;    // Write register for disk writes.
} DriveState;


#define SECTOR_OFFSET(t,s) ((t) * BYTES_PER_TRACK + ((s) * BYTES_PER_SECTOR))
#define NO_QTRACK (-1)

@interface DiskCard:PeripheralCard
{
  id diskList1;
  id diskList2;
  id accessLight1;
  id accessLight2;
  
  int drive;                      // Currently selected disk.
  DriveState drive_state[2];      // State of the drive hardware.
  Disk disk[2];                   // Actual disk data.
  mutex_t drive_mutex;            // Mutex to protect disk data.
  unsigned char prom_image[256];  // Default values the Apple // sees
}

- changeDisk:sender;
- (const char *)insertDisk:(const char *)path inDrive:(int) driveNumber;
- ejectDisk:(int)driveNumber;

/* Handlers for Apple ][ actions. */
- (BOOL)deviceDriverCalled;
- (BOOL)bootUp;
- (BOOL)rwts;
- (BOOL)cardReadSector;
- (int)stepperMotorReferenced:(AccessType)type :(MemoryElement *)m;
- (int)selectDrive:(AccessType)type :(MemoryElement *)m;
- (int)handleDataLatch:(AccessType)type :(MemoryElement *)m;
- (int)undoWrite:(AccessType)t :(MemoryElement *)m;

@end
