/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import "PeripheralCard.h"
#import "handlers.h"
#import "memory.h"

@interface PeripheralCard(Private)
- _setupPeripheralCardWithBundle:aBundle;
@end


@implementation PeripheralCard

+ (int)preferredSlot:(int)availableSlotsMask
{
  int                 possibleSlot;

  for (possibleSlot = 0; possibleSlot < 7; possibleSlot++)
  {
    if (availableSlotsMask % 1)
      return possibleSlot + 1;
    availableSlotsMask >>= 1;
  }

  return NO_SLOT;
}

- initFromBundle:aBundle
{
  [super init];
  return[self _setupPeripheralCardWithBundle:aBundle];
}

- awake
{
  return[self _setupPeripheralCardWithBundle:[NXBundle bundleForClass:
					      [self class]]];
}

- _setupPeripheralCardWithBundle:aBundle
{
  char                path[MAXPATHLEN + 1];

  bundle = aBundle;
  slot_number = NO_SLOT;
  state = NULL;
  computer = nil;

  if ([bundle getPath:path forResource:".dir" ofType:"tiff"])
    iconImage = [[NXImage allocFromZone:[self zone]] initFromFile:path];
  else
    iconImage = [NXImage findImageNamed:"PeripheralCard"];

  if ([bundle getPath:path forResource:"Options" ofType:"nib"])
    [NXApp loadNibFile:path owner:self];

  return self;
}

- free
{
  [self removeFromSlot];
  return[super free];
}

- read:(NXTypedStream *) stream
{
  [super read:stream];
  computer = NXReadObject(stream);

  return self;
}

- write:(NXTypedStream *) stream
{
  [super write:stream];
  NXWriteObjectReference(stream, computer);

  return self;
}

- insert:(MachineState *) st inComputer:(Computer *) comp inSlot:(int)slot
{
  if (slot_number != NO_SLOT)	/* If in slot */
    [self removeFromSlot];

 /* Remember where we were inserted. */
  computer = comp;
  state = st;
  slot_number = slot;

  return self;
}

- removeFromSlot
{
  int                 clearIndex;
  MemoryElement      *clearElement;

  if (slot_number != NO_SLOT)
  {
  /* Remove all address/routine handlers. */
    [self removeAddressHandlers:SWITCHES_FOR_SLOT(slot_number)
      :SWITCHES_FOR_SLOT_END(slot_number)
      :BOTH];
    [self removeAddressHandlers:ROM_FOR_SLOT(slot_number)
      :ROM_FOR_SLOT_END(slot_number)
      :BOTH];
    [self removeRoutineHandlers:ROM_FOR_SLOT(slot_number)
      :ROM_FOR_SLOT_END(slot_number)];

  /* Clear out our card's main memory range. */
    clearElement = &state->sysmem[ROM_FOR_SLOT_END(slot_number)];
    clearIndex = 0x100;
    while (clearIndex--)
      [computer writeMemoryElement:clearElement--:0];

    state->peripheral_card[slot_number] = nil;
  }
  slot_number = NO_SLOT;
  return self;
}


/* Returns the current slot, or NO_SLOT if this card is not
 * currently inserted.
 */
- (int)currentSlot
{
  return slot_number;
}


/* Returns the Computer object this card is plugged into. */
- computer
{
  return computer;
}


/* Traps memory accesses to a given address.  You can specify which type
 * of accesses you'd like to trap by passing READ, WRITE, or BOTH as
 * the type argument.  Whenever that trap is hit the card object will be
 * sent the specified message; the value your method returns is the value
 * the 6502 will see at that address.  The selector should be prototyped as:
 *
 * -(int)myHandler:(AccessType)type :(MemoryElement *)m;
 *
 * Note that the addresses passed here are *NOT* the "actual" Apple ][
 * addresses, but the indices of the affected MemoryElements.  In other words,
 * you should pass offsets from the constants in memory.h for lowAddress and
 * highAddress.
 */
- addAddressHandler:(int)lowAddress :(int)highAddress :(SEL)selector
   :(AccessType) type
{
  add_slot_address_handler(slot_number, state, lowAddress, highAddress,
			   selector, type);
  return self;
}


/* Traps any attempt to execute code in a given memory range.
 * If this code is executed, the card object will be sent a "selector"
 * message.  The method to be trapped to should be prototyped as:
 *
 * -(BOOL) myRoutineHandler; 
 *
 * Returning YES indicates that an RTS should be executed on return to the
 * main emulator.
 *
 * Note that the addresses passed here are *NOT* the "actual" Apple ][
 * addresses, but the indices of the affected MemoryElements.  In other words,
 * you should pass offsets from the constants in memory.h for lowAddress and
 * highAddress.
 */

- addRoutineHandler:(int)lowAddress :(int)highAddress :(SEL)selector
{
  add_slot_routine_handler(slot_number, state, lowAddress,
			   highAddress, selector);
  return self;
}


/* This removes all address handlers of a specific type in a given memory
 * range.  Note that this removes ALL handlers, and not just those installed
 * by this card, so be careful!
 *
 * Note also that the addresses passed here are *NOT* the "actual" Apple ][
 * addresses, but the indices of the affected MemoryElements.  In other words,
 * you should pass offsets from the constants in memory.h for lowAddress and
 * highAddress.
 */
- removeAddressHandlers:(int)lowAddress :(int)highAddress :(AccessType) type
{
  remove_address_handlers(state, lowAddress, highAddress, type);
  return self;
}


/* This removes all routine handlers of a specific type in a given memory
 * range.  Note that this removes ALL handlers, and not just those installed
 * by this card, so be careful!
 *
 * Note that the addresses passed here are *NOT* the "actual" Apple ][
 * addresses, but the indices of the affected MemoryElements.  In other words,
 * you should pass offsets from the constants in memory.h for lowAddress and
 * highAddress.
 */
- removeRoutineHandlers:(int)lowAddress :(int)highAddress
{
  remove_routine_handlers(state, lowAddress, highAddress);
  return self;
}


/* Handy when you want a memory access to be ignored. */
- (int)emptyHandler:(AccessType) t :(MemoryElement *) m
{
  return 0;
}

- iconImage
{
  return iconImage;
}

- optionView
{
  return optionView;
}

- bundle
{
  return bundle;
}

#if 1
/* This is a horrible last second hack to make the dynamically loadable
 * DiskCard link correctly.  If realpath() isn't used somewhere here the
 * realpath() in the DiskCard fails to link.  I don't understand this yet,
 * but I want to get something out to the rest of the people who want to
 * work on this.  Someone can show me the error of my ways then.
 */
static void
temporary_hack()
{
  realpath();
  strcasecmp(0, 0);
  fgetc(0);
  fgetpos(0, 0);
  fsetpos(0, 0);
}

#endif

@end
