/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import "SerialCard.h"
#import <sys/param.h>    /* for MAXPATHLEN */
#import <assert.h>
#import "memory.h"


@implementation SerialCard


/* This maps from the ACIA baud rates to the NeXT baud rate #define's. */
static const int baud_map[] = {
  EXTB /* FIXME - nothing analagous to external clock - assume 38400 */,
  B50, B75, B110, B134, B150, B300, B600, B1200, B1800, B2400,
  B2400 /* FIXME - no 3600 */, B4800,
  B4800 /* FIXME - no 7200 */, B9600, EXTA
  };


- initFromBundle:(NXBundle *)b
{
  char path[MAXPATHLEN + 1];
  NXStream *s;

  [super initFromBundle:b];

  serial_port_fd = -1;

  /* Load up our memory image. */
  if ([b getPath:path forResource:"card" ofType:"prom"] && (s = NXMapFile (path, NX_READONLY)) != NULL) {
    NXRead (s, prom_image, 256);
    NXCloseMemory (s, NX_FREEBUFFER);
  }
  else {
    NXRunAlertPanel (NULL, "Fatal error: unable to load file \"card.prom\".", NULL, NULL, NULL);
    return nil;
  }

  /* Set up default serial device information. */
  strcpy (serial_dev, "/dev/ttya");
  default_baud_rate = 15;	/* 9600 */

  return self;
}

- free
{
  return [super free];
}

- insert:(MachineState *)st inComputer:(Computer *)comp inSlot:(int)slot
{
  char path[MAXPATHLEN + 1];
  MemoryElement *m;


  [super insert:st inComputer:comp inSlot:slot];

  /* Copy prom image to Apple2 memory. */
  m = &state->sysmem[ROM_FOR_SLOT(slot_number)];
  for (i = 0; i < 256; i++)
    [computer writeMemoryElement:m++ :prom_image[i]];

  /* Write protect our memory image. */
  [self addAddressHandler:IO_SELECT_ADDRESS_BEGIN(slot_number)
  			 :IO_SELECT_ADDRESS_END(slot_number)
                         :@selector (undoWrite::)
                         :WRITE];

  /* Trap ACIA registers. */
  [self addAddressHandler:0xC088 + slot_number * 0x10
                         :0xC088 + slot_number * 0x10
                         :@selector (ACIARTRegister::)
                         :BOTH];

  [self addAddressHandler:0xC089 + slot_number * 0x10
                         :0xC089 + slot_number * 0x10
                         :@selector (ACIAStatusRegister::)
                         :BOTH];

  [self addAddressHandler:0xC08A + slot_number * 0x10
                         :0xC08A + slot_number * 0x10
                         :@selector (ACIACommandRegister::)
                         :BOTH];

  [self addAddressHandler:0xC08B + slot_number * 0x10
                         :0xC08B + slot_number * 0x10
                         :@selector (ACIAControlRegister::)
                         :BOTH];

  /* The last thing we do should be to actually install ourselves. */
  st->peripheral_card[slot] = self;
  return self;
}

- (BOOL)initCard
{
  struct sgttyb s;
  struct tchars t;

#define INIT_ERR(z) \
  do { \
    serial_port_fd = -1; \
    state->x = (z); \
    state->c = 1; \
    return YES; \
  } while (0)

  /* Is it already open?  Close it! */
  if (serial_port_fd >= 0)
    close (serial_port_fd);

  /* Get file descriptor associated with serial port A. */
  serial_port_fd = open (serial_dev, O_RDWR, 0666);
  if (serial_port_fd < 0)
    INIT_ERR (SERIAL_ERR_WHO_KNOWS);

  /* Allow no one else to access the serial port. */
  if (ioctl (serial_port_fd, TIOCEXCL, NULL) == -1)
    {
      close (serial_port_fd);
      INIT_ERR (SERIAL_ERR_WHO_KNOWS);
    }

  /* Disable all "special" interrupt characters. */
  t.t_intrc = t.t_quitc = t.t_startc = t.t_stopc = t.t_eofc = t.t_brkc = -1;
  if (ioctl (serial_port_fd, TIOCSETC, &t) == -1)
    {
      close (serial_port_fd);
      INIT_ERR (SERIAL_ERR_WHO_KNOWS);
    }

  /* Set serial port to 9600 baud in RAW mode. */
  s.sg_ispeed = baud_map[default_baud_rate];
  s.sg_ospeed = baud_map[default_baud_rate];
  s.sg_erase  = -1;
  s.sg_kill   = -1;
  s.sg_flags  = RAW;
  if (ioctl (serial_port_fd, TIOCSETP, &s) == -1)
    {
      close (serial_port_fd);
      INIT_ERR (SERIAL_ERR_WHO_KNOWS);
    }

  /* Set up control register default values. */
  ctl_reg_data.baud_rate   = default_baud_rate;
  ctl_reg_data.stop_bit    = 0;
  ctl_reg_data.word_length = 0;

#undef INIT_ERR

  return YES;
}

- (int)ACIARTRegister:(AccessType)t :(MemoryElement *)m
{
  /* Make sure it's already open. */
  if (serial_port_fd < 0)
    {
      [self initCard];
      if (serial_port_fd < 0)   /* Some problem opening it, whoops. */
	return 0;
    }
}

- (int)ACIAStatusRegister:(AccessType)t :(MemoryElement *)m
{
  /* Make sure it's already open. */
  if (serial_port_fd < 0)
    {
      [self initCard];
      if (serial_port_fd < 0)   /* Some problem opening it, whoops. */
	return 0;
    }
}

- (int)ACIACommandRegister:(AccessType)t :(MemoryElement *)m
{
  int v;

  /* Make sure it's already open. */
  if (serial_port_fd < 0)
    {
      [self initCard];
      if (serial_port_fd < 0)   /* Some problem opening it, whoops. */
	return 0;
    }

  if (t == READ)
    {
      return ((cmd_reg_data.parity << 5)
	      | (cmd_reg_data.echo << 4)
	      | (cmd_reg_data.xmit_ctl << 2)
	      | (cmd_reg_data.rcv_int_enable << 1)
	      | cmd_reg_data.dtr);
    }


  /* Update whatever bits they changed. */
  v = m->mem.byte;
  if (((v >> 5) & 0x7) != cmd_reg_data.parity)
    {
      cmd_reg_data.parity = ((v >> 5) & 7);
      
    }
  if (((v >> 4) & 0x1) != cmd_reg_data.echo)
    {
      cmd_reg_data.echo = ((v >> 4) & 0x1);
    }
  if (((v >> 2) & 0x3) != cmd_reg_data.xmit_ctl)
    {
      cmd_reg_data.xmit_ctl = ((v >> 2) & 0x3);
    }
  if (((v >> 1) & 0x1) != cmd_reg_data.rcv_int_enable)
    {
      cmd_reg_data.rcv_int_enable = ((v >> 1) & 0x1);
    }
  if ((v & 0x1) != cmd_reg_data.dtr)
    {
      cmd_reg_data.dtr = v & 0x1;
      ioctl (serial_port_fd, cmd_reg_data.dtr ? TIOCSDTR : TIOCCDTR, NULL);
    }

  return 0;
}

- (int)ACIAControlRegister:(AccessType)t :(MemoryElement *)m
{
  struct sgttyb s;
  int v;

  /* Make sure it's already open. */
  if (serial_port_fd < 0)
    {
      [self initCard];
      if (serial_port_fd < 0)   /* Some problem opening it, whoops. */
	return 0;
    }

  /* If they were reading, return the status register. */
  if (t == READ)
    {
      return ((ctl_reg_data.stop_bit << 7)
	      | (ctl_reg_data.word_length << 5)
	      | (1 << 4)
	      | ctl_reg_data.baud_rate);
    }

  /* Get the value they just wrote. */
  v = m->mem.byte;
  
  /* See if they were trying to change the baud rate. */
  if ((v & 0xF) != ctl_reg_data.baud_rate)
    {
      ctl_reg_data.baud_rate = v & 0xF;

      /* Set the baud rate. */
      ioctl (serial_port_fd, TIOCGETP, &s);
      s.sg_ispeed = s.sg_ospeed = baud_map[ctl_reg_data.baud_rate];
      ioctl (serial_port_fd, TIOCSETP, &s);
    }

  /* See if they were changing word length or stop bits. */
  if (((v >> 5) & 3) != ctl_reg_data.word_length
      || (v >> 7) != ctl_reg_data.stop_bit)
    {
      ctl_reg_data.word_length = (v >> 5) & 3;
      ctl_reg_data.stop_bit    = (v >> 7);
    }

  return 0;
}

@end
