/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * UnixShare.c - this handles mapping of the UNIX
 * 		 filesystem into a ProDOS compatible format
 * written by Chris Osborn <fozztexx%nvcc.uucp@groucho.sonoma.edu>
 */

#import "UnixShare.h"
#import "memory.h"
#import <stdio.h>
#import <sys/param.h>
#import <strings.h>
#import <sys/file.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/errno.h>
#import <sys/dir.h>

#define READLOC(x)	[computer readByte:x]
#define READWORD(x)	[computer readShort:x]
#define WRITELOC(x,y)	[computer writeByte:x :y]
#define LDA(v) (void)({int _v = v;state->a=_v;state->n_flag = _v > 0x7f;state->z_flag = _v == 0;})

#define UNIX_START	0x90
#define UNIX_PDOSOP	(UNIX_START + 4)
#define UNIX_COMMAND	(UNIX_START + 6)
#define UNIX_PARM	(UNIX_START + 7)

#define UNIX_DIRECTORY 0x0100
#define UNIX_DOCRLFCNV 0x0200

#define UNIX_SLOTDRIVE 0x20

#define UPPER(a) ({int _a1 = a; _a1 >= 'a' && _a1 <= 'z' ? _a1 - 32 : _a1;})
#define LOWER(a) ({int _a1 = a; _a1 >= 'A' && _a1 <= 'Z' ? _a1 + 32 : _a1;})

#define PDOSERR_NOERROR		0x00
#define PDOSERR_INVALIDMLICMD	0x01
#define PDOSERR_INVALIDPARMCNT	0x04
#define PDOSERR_INTRPTTBLFULL	0x25
#define PDOSERR_IOERROR		0x27
#define PDOSERR_NODEVICECNCTD	0x28
#define PDOSERR_WRITEPROTECTED	0x2b
#define PDOSERR_VOLUMESWITCHED	0x2e
#define PDOSERR_INVALIDPATHSYN	0x40
#define PDOSERR_TOOMANYOPEN	0x42
#define PDOSERR_INVALIDREFNUM	0x43
#define PDOSERR_NOTADIR		0x44
#define PDOSERR_VOLUMENTMNTD	0x45
#define PDOSERR_FILENOTFOUND	0x46
#define PDOSERR_DUPFILENAME	0x47
#define PDOSERR_DISKFULL	0x48
#define PDOSERR_VOLDIRFULL	0x49
#define PDOSERR_INCOMPFILETYPE	0x4a
#define PDOSERR_UNSUPSTORTYPE	0x4b
#define PDOSERR_ENDOFFILE	0x4c
#define PDOSERR_POSPASTEOF	0x4d
#define PDOSERR_ACCESSERROR	0x4e
#define PDOSERR_FILEALRDYOPEN	0x50
#define PDOSERR_FILECNTBAD	0x51
#define PDOSERR_NONPRODOS	0x52
#define PDOSERR_BADPARM		0x53
#define PDOSERR_VCBOVERFLOW	0x55
#define PDOSERR_BADBUFADDR	0x56
#define PDOSERR_DUPVOLUME	0x57
#define PDOSERR_BADVOLBITMAP	0x5a

typedef struct fileInfoRec
{
  int                 accessBits;
  int                 fileType;
  int                 auxType;
  int                 storageType;
  int                 blocksUsed;
  int                 fileLength;
  int                 dateLastMod;
  int                 timeLastMod;
  int                 dateCreate;
  int                 timeCreate;
}                   fileInfo;

@implementation UnixShare

- initFromBundle:aBundle
{
  char                path[MAXPATHLEN + 1];


  [super initFromBundle:aBundle];

  unixPrefix[0] = 0;
  lastCmd = 0;
  readbuf = NULL;
  readbuflength = 0;

  [self setUnixVolume:NXHomeDirectory()];

  unixDoTXTConv = YES;
  [txtConvButton setState:unixDoTXTConv];

  return self;
}

- free
{
  return[super free];
}

- fuzzball
{
  char                buf[MAXPATHLEN + 1];


  [super awake];

  if ([bundle getPath:buf forResource:"UnixShare" ofType:"nib"])
    [NXApp loadNibFile:buf owner:self];

  unixPrefix[0] = 0;
  lastCmd = 0;
  readbuf = NULL;
  readbuflength = 0;

  if (unixVolumePath[0])
    sprintf(buf, "%s/%s", unixVolumePath, unixVolumeName);
  else
    strcpy(buf, unixVolumeName);
  [unixField setStringValue:buf];
  [txtConvButton setState:unixDoTXTConv];

  return self;
}

- read:(NXTypedStream *) stream
{
  [super read:stream];
  NXReadArray(stream, "c", sizeof(unixVolumeName), unixVolumeName);
  NXReadArray(stream, "c", sizeof(unixVolumePath), unixVolumePath);
  NXReadTypes(stream, "i", &unixDoTXTConv);

  return self;
}

- write:(NXTypedStream *) stream
{
  [super write:stream];
  NXWriteArray(stream, "c", sizeof(unixVolumeName), unixVolumeName);
  NXWriteArray(stream, "c", sizeof(unixVolumePath), unixVolumePath);
  NXWriteTypes(stream, "i", &unixDoTXTConv);

  return self;
}

- setUnixVolume:(const char *)name
{
  char               *p;


  strcpy(unixVolumePath, name);
  [unixField setStringValue:unixVolumePath];

  if (p = rindex(unixVolumePath, '/'))
  {
    strcpy(unixVolumeName, p + 1);
    *p = 0;
  }
  else
  {
    strcpy(unixVolumeName, unixVolumePath);
    unixVolumePath[0] = 0;
  }

  return self;
}

- insert:(MachineState *) st inComputer:aComputer inSlot:(int)slot
{
  int                 i;


  [super insert:st inComputer:aComputer inSlot:slot];

  rom_image[0] = 'U';
  rom_image[1] = 'X';
  rom_image[2] = 'S';
  rom_image[3] = UNIX_START;

  i = UNIX_START;
  rom_image[i++] = 0x30;
  rom_image[i++] = 0x01;
  rom_image[i++] = 0x60;
  rom_image[i++] = 0x20;
  rom_image[i++] = 0;
  rom_image[i++] = 0;
  rom_image[i++] = 0;
  rom_image[i++] = 0;
  rom_image[i++] = 0;
  rom_image[i++] = 0x8D;
  rom_image[i++] = UNIX_START;
  rom_image[i++] = 0xC0 | slot_number;
  rom_image[i++] = 0x60;

  [self addAddressHandler:ROM_FOR_SLOT(slot_number)
    :ROM_FOR_SLOT_END(slot_number)
    :@selector(romTwiddle::) :BOTH];
  [self addRoutineHandler:ROM_FOR_SLOT(slot_number)
   + UNIX_START
   :ROM_FOR_SLOT_END(slot_number)
   + UNIX_START
   :@selector(unixShare)];

  return self;
}

- setTXTConv:sender
{
  unixDoTXTConv = [sender state];

  return self;
}

- setUnixVol:sender
{
  id                  myOP;
  const char         *const types[] = {NULL};
  char               *p;


  myOP = [OpenPanel new];
  [myOP setDelegate:self];

  [myOP setTitle:"Choose directory"];
  [myOP chooseDirectories:YES];
  if ([myOP runModalForTypes:types])
  {
    [unixField setStringValue:[myOP filename]];
    strcpy(unixVolumePath,[unixField stringValue]);
    if (p = rindex(unixVolumePath, '/'))
    {
      strcpy(unixVolumeName, p + 1);
      *p = 0;
    }
    else
    {
      strcpy(unixVolumeName, unixVolumePath);
      unixVolumePath[0] = 0;
    }
  }
  [myOP setDelegate:nil];
  [myOP chooseDirectories:NO];

  return self;
}

- (int)paccess:(char *)filename
{
  int                 r, w, d;
  char                buf[MAXPATHLEN + 1];
  char               *p;
  int                 bits;


  r = !access(filename, R_OK);
  w = !access(filename, W_OK);
  strcpy(buf, filename);
  if (p = rindex(buf, '/'))
    *p = 0;
  else
    strcpy(buf, ".");
  d = !access(buf, W_OK);

  bits =
    (d << 7) |			/* DESTROY */
    (d << 6) |			/* RENAME  */
    (1 << 5) |			/* BACKUP. No equivalent. Just set it. Maybe put in AD info */
    (0 << 4) |			/* unused  */
    (0 << 3) |			/* unused  */
    (0 << 2) |			/* unused  */
    (w << 1) |			/* WRITE   */
    (r << 0);			/* READ    */

  return bits;
}

- getAppleDoubleInfo:(fileInfo *) info for:(char *)
    filename
{
  char                path[MAXPATHLEN + 1];
  char                file[MAXPATHLEN + 1];
  char                buf[MAXPATHLEN + 1];
  struct stat         sbuf;
  FILE               *ainfo;
  char               *p;
  int                 i, j;

  int                 mkdir();


  strcpy(path, filename);
  if (p = rindex(path, '/'))
  {
    strcpy(file, p + 1);
    *p = 0;
  }
  else
  {
    path[0] = 0;
    strcpy(file, filename);
  }

  if (stat(filename, &sbuf))
  {
    return nil;
  }
  else
  {
    info->accessBits = [self paccess:filename];
    info->storageType = 0x01;	/* Seedling. What does AppleShare do? */
    info->blocksUsed = sbuf.st_size / 512 + (sbuf.st_size % 512 > 0);
    info->fileLength = sbuf.st_size;
    info->dateLastMod = ([self ptime:sbuf.st_mtime] >> 16) & 0xffff;
    info->timeLastMod = [self ptime:sbuf.st_mtime] & 0xffff;
    info->dateCreate = ([self ptime:sbuf.st_ctime] >> 16) & 0xffff;
    info->timeCreate = [self ptime:sbuf.st_ctime] & 0xffff;
  }

  if (sbuf.st_mode & S_IFDIR)
  {
    info->storageType = 0x0d;
    info->fileType = 0x0f;
    info->auxType = 0x0000;
    if (!strcmp(path, unixVolumePath) && !strcmp(file, unixVolumeName))
    {
      sprintf(buf, "/bin/df %s/%s\n", unixVolumePath, unixVolumeName);
      ainfo = popen(buf, "r");
      fgets(buf, sizeof(buf), ainfo);
      fscanf(ainfo, "%*[^ ] %i %*i %i", &i, &j);
      pclose(ainfo);
      info->auxType = i / 1024;	/* FIXME - Should be disk block size. But only got 16 bits */
      info->blocksUsed = ((i - j) + 1023) / 1024;	/* FIXME - Should be disk blocks used */
    }
  }
  else
  {
    sprintf(buf, "%s/.AppleDouble/%s", path, file);
    if (!(ainfo = fopen(buf, "r")))
    {				/* FIXME */
    /*
     * Can't open? Just set some defaults then. Really, it should
     * pop a panel and ask, but since this will be called from
     * the CPU thread, we can't 
     */
      info->fileType = 0x04;	/* TXT */
      info->auxType = 0x00;
    }
    else
    {				/* FIXME - Make this read apple double later */
      fscanf(ainfo, "%X\n%X\n", (unsigned int *)&info->fileType, (unsigned int *)&info->auxType);
      fclose(ainfo);
    }
  }

  return self;
}

- (int)setAppleDoubleInfo:(fileInfo *) finfo for:(char *)
    filename
{
  char                path[MAXPATHLEN + 1];
  char                file[MAXPATHLEN + 1];
  char                buf[MAXPATHLEN + 1];
  struct stat         sbuf;
  FILE               *ainfo;
  char               *p;
  int                 np, op, um;

  int                 mkdir();


  strcpy(path, filename);
  if (p = rindex(path, '/'))
  {
    strcpy(file, p + 1);
    *p = 0;
  }
  else
  {
    path[0] = 0;
    strcpy(file, filename);
  }

  if (finfo->fileType != 0x0f)
  {
    sprintf(buf, "%s/.AppleDouble", path);
    if (stat(buf, &sbuf))
    {
      if (mkdir(buf, 0777))
	return 0;
    }

    sprintf(buf, "%s/.AppleDouble/%s", path, file);
    if (!(ainfo = fopen(buf, "w")))
      return 0;

    fprintf(ainfo, "%X\n%X\n", (unsigned int)finfo->fileType, (unsigned int)finfo->auxType);
    fclose(ainfo);

    sprintf(buf, "%s/%s", path, file);
  }

  np = 0444 * (finfo->accessBits & 1 > 0) |	/* Set all reads if READ */
    0222 * (finfo->accessBits & 2 > 0) |	/* Set all writes if WRITE */
    0111;			/* Set all executes. Mask out later */
  stat(filename, &sbuf);
  op = sbuf.st_mode | 0666;
  um = umask(0);
  umask(um);
  np &= op;			/* Mask out execution */
  np &= ~um;			/* Mask out umask */
  chmod(filename, np);

 /* FIXME */
 /*
  * About all we can do is try to change permissions. I don't see
  * any way to change the date. Oh well. 
  */
 /*
  * Well, actually, I guess utime lets you change modify and
  * access times. 
  */
 /*
  * Also, I've noticed that on the NeXT, the create time is
  * almost always wrong. 
  */
 /* I'll set it up later */

  return 1;
}

- (int)removeAppleDoubleInfoFor:(char *)filename
{
  char                path[MAXPATHLEN + 1];
  char                file[MAXPATHLEN + 1];
  char                buf[MAXPATHLEN + 1];
  struct stat         sbuf;
  char               *p;

  int                 unlink();


  strcpy(path, filename);
  if (p = rindex(path, '/'))
  {
    strcpy(file, p + 1);
    *p = 0;
  }
  else
  {
    path[0] = 0;
    strcpy(file, filename);
  }

  sprintf(buf, "%s/.AppleDouble", path);
  if (stat(buf, &sbuf))
  {
    return 0;
  }

  sprintf(buf, "%s/.AppleDouble/%s", path, file);
  if (!stat(buf, &sbuf))
    unlink(buf);

  return 1;
}

- (char *)validPName:(char *)path
{
  static char         file[MAXPATHLEN + 1];
  int                 i, j, k;


  strcpy(file, path);
  for (i = 0; file[i]; i++)
    file[i] = UPPER(file[i]);

  for (i = k = 0, j = 1; file[i]; i++, k++)
  {
    if (file[i] != '/')
    {
      if (j && (file[i] < 'A' || file[i] > 'Z'))
	return NULL;
      if (!index(".0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ", file[i]))
	return NULL;
      j = 0;
    }
    else
    {
      if (k > 15)
	return NULL;
      k = -1;
      j = 1;
    }
  }
  if (k > 15)
    return NULL;

  return file;
}

- (char *)mapPName:(char *)file inDir:(char *)path
 /*
  * This will read through the directory <path> to find a match
  * for the Prodos name <file>. If it finds one, it will return
  * it, otherwise returns <file>. Returns NULL on err. 
  */
{
  static char         fname[MAXPATHLEN + 1];
  DIR                *dirp;
  struct direct      *ent;
  int                 len;


  if (!(dirp = opendir(path)))
    return NULL;

  len = strlen(file);

  for (ent = readdir(dirp); ent; ent = readdir(dirp))
  {
    strcpy(fname, ent->d_name);	/* Copy now before gets overwritten */
    if (ent->d_namlen == len && [self validPName:fname] && !strcasecmp(fname, file))
    {
      closedir(dirp);
      return fname;
    }
  }

  closedir(dirp);

  return file;
}

- (char *)makeUnixPath:(int)fnameLoc
 /*
  * This returns NULL if the path won't be in our UNIX dir. fname
  * is a pointer to the actual Prodos filename in SysMem. 
  */
{
  static char         fullname[MAXPATHLEN + 1];
  char                pname[129], pname2[129];
  char               *p, *q;
  int                 i;


  strcpy(pname,[self makeCString:fnameLoc]);
  i = strlen(pname);
  if (i > 1 && pname[i - 1] == '/')
    pname[i - 1] = 0;

  if (pname[0] != '/')
  {
    strcpy(fullname, pname);
    sprintf(pname, "%s%s", unixPrefix, fullname);
  }

  if (pname[0] != '/' || ![self validPName:pname])
    return NULL;

  strcpy(fullname, &pname[1]);
  if (p = index(fullname, '/'))
    *p = 0;
  if (strcasecmp(fullname, unixVolumeName))
    return NULL;
  sprintf(fullname, "%s/%s", unixVolumePath, unixVolumeName);
  if (p = index(&pname[1], '/'))
    p++;
  while (p && *p)
  {
    strcpy(pname2, p);
    if (q = index(pname2, '/'))
      *q = 0;
    if (!(q = [self mapPName:pname2 inDir:fullname]))
      return NULL;
    strcat(fullname, "/");
    strcat(fullname, q);
    if (p = index(p, '/'))
      p++;
  }

  return fullname;
}

- (char *)loadDir:(char *)path length:(int *)length
 /*
  * This will load an entire directory into ram, converting into
  * the ProDOS directory structure. 
  */
{
  char               *dirBlock;
  DIR                *dirp;
  struct direct      *ent;
  int                 i, j, k, l;
  int                 cb;
  fileInfo            finfo;
  char                buf[MAXPATHLEN + 1];
  char               *p;

  void               *malloc();


  if (!(dirp = opendir(path)))
    return NULL;

 /* Find out how many entries there are */
  k = 43;
  i = 1;
  while (ent = readdir(dirp))
  {
    sprintf(buf, "%s/%s", path, ent->d_name);
    if ([self validPName:ent->d_name] && [self getAppleDoubleInfo:&finfo for:buf])
    {
      i++;
      k += 0x27;
      if ((512 - k % 512) < 0x27)
	k += (512 - k % 512) + 4;
    }
  }
  rewinddir(dirp);

  k = ((k + 511) / 512) * 512;
  dirBlock = malloc(k);
  memset(dirBlock, 0, k);

  strcpy(buf, path);
  if (buf[strlen(buf) - 1] == '/')
    buf[strlen(buf) - 1] = 0;

  if (p = rindex(buf, '/'))
    p++;
  else
    p = path;

  k = 4;
  cb = 0;
  dirBlock[k - 0x04] = cb & 0xff;
  dirBlock[k - 0x03] = (cb >> 8) & 0xff;
  dirBlock[k - 0x02] = (cb + 1) & 0xff;
  dirBlock[k - 0x01] = ((cb + 1) >> 8) & 0xff;
  [self getAppleDoubleInfo:&finfo for:path];
  dirBlock[k + 0x00] = (strcmp(path, unixVolumeName) ? 0xe0 : 0xf0) | strlen(p);
  for (l = 0; l < (dirBlock[k + 0x00] & 0xf); l++)
    dirBlock[k + 0x01 + l] = UPPER(p[l]);
  dirBlock[k + 0x10] = 0x75;
  dirBlock[k + 0x18] = finfo.dateCreate & 0xff;
  dirBlock[k + 0x19] = (finfo.dateCreate >> 8) & 0xff;
  dirBlock[k + 0x1a] = finfo.timeCreate & 0xff;
  dirBlock[k + 0x1b] = (finfo.timeCreate >> 8) & 0xff;
  dirBlock[k + 0x1c] = 0;
  dirBlock[k + 0x1d] = 0;
  dirBlock[k + 0x1e] = finfo.accessBits;
  dirBlock[k + 0x1f] = 0x27;
  dirBlock[k + 0x20] = 0x0d;
  dirBlock[k + 0x21] = (i - 1) & 0xff;
  dirBlock[k + 0x22] = ((i - 1) >> 8) & 0xff;
  dirBlock[k + 0x23] = dirBlock[k + 0x24] = dirBlock[k + 0x25] = dirBlock[k + 0x26] = 0;
  k += 0x27;

  for (j = 1; j < i;)
  {
    ent = readdir(dirp);
    sprintf(buf, "%s/%s", path, ent->d_name);
    if ([self validPName:ent->d_name] && [self getAppleDoubleInfo:&finfo for:buf])
    {
      if ((512 - k % 512) < 0x27)
      {
	k += (512 - k % 512) + 4;
	cb++;
	dirBlock[k - 0x04] = cb & 0xff;
	dirBlock[k - 0x03] = (cb >> 8) & 0xff;
	dirBlock[k - 0x02] = (cb + 1) & 0xff;
	dirBlock[k - 0x01] = ((cb + 1) >> 8) & 0xff;
      }

      dirBlock[k + 0x00] = (finfo.storageType << 4) | (ent->d_namlen <= 15 ? ent->d_namlen : 15);
      strcpy(&dirBlock[k + 0x01],[self validPName:ent->d_name]);
      dirBlock[k + 0x10] = finfo.fileType;
      dirBlock[k + 0x11] = 0;
      dirBlock[k + 0x12] = 0;
      dirBlock[k + 0x13] = finfo.blocksUsed & 0xff;
      dirBlock[k + 0x14] = (finfo.blocksUsed >> 8) & 0xff;
      dirBlock[k + 0x15] = finfo.fileLength & 0xff;
      dirBlock[k + 0x16] = (finfo.fileLength >> 8) & 0xff;
      dirBlock[k + 0x17] = (finfo.fileLength >> 16) & 0xff;
      dirBlock[k + 0x18] = finfo.dateCreate & 0xff;
      dirBlock[k + 0x19] = (finfo.dateCreate >> 8) & 0xff;
      dirBlock[k + 0x1a] = finfo.timeCreate & 0xff;
      dirBlock[k + 0x1b] = (finfo.timeCreate >> 8) & 0xff;
      dirBlock[k + 0x1c] = 0;
      dirBlock[k + 0x1d] = 0;
      dirBlock[k + 0x1e] = finfo.accessBits;
      dirBlock[k + 0x1f] = finfo.auxType & 0xff;
      dirBlock[k + 0x20] = (finfo.auxType >> 8) & 0xff;
      dirBlock[k + 0x21] = finfo.dateLastMod & 0xff;
      dirBlock[k + 0x22] = (finfo.dateLastMod >> 8) & 0xff;
      dirBlock[k + 0x23] = finfo.timeLastMod & 0xff;
      dirBlock[k + 0x24] = (finfo.timeLastMod >> 8) & 0xff;
      dirBlock[k + 0x25] = 0;
      dirBlock[k + 0x26] = 0;
      k += 0x27;

      j++;
    }
  }
  closedir(dirp);
  dirBlock[k + 2 - k % 512] = dirBlock[k + 3 - k % 512] = 0;
  k = ((k + 511) / 512) * 512;
  *length = k;

  return dirBlock;
}

- (int)mapErrnoToPdos:(int)err
 /*
  * This is very incomplete. Some of these may be better suited
  * to different Prodos errors as well. 
  */
{
  switch (err)
  {
  case EPERM:
    return PDOSERR_ACCESSERROR;
  case ENOENT:
    return PDOSERR_FILENOTFOUND;
  case EIO:
    return PDOSERR_IOERROR;
  case ENXIO:
    return PDOSERR_IOERROR;
  case EBADF:
    return PDOSERR_IOERROR;
  case EACCES:
    return PDOSERR_ACCESSERROR;
  case EEXIST:
    return PDOSERR_DUPFILENAME;
  case ENOTDIR:
    return PDOSERR_NOTADIR;
  case EISDIR:
    return PDOSERR_ACCESSERROR;
  case EMFILE:
    return PDOSERR_TOOMANYOPEN;
  case ETXTBSY:
    return PDOSERR_IOERROR;
  case EFBIG:
    return PDOSERR_IOERROR;
  case ENOSPC:
    return PDOSERR_DISKFULL;
  case EROFS:
    return PDOSERR_ACCESSERROR;
  default:
    return PDOSERR_IOERROR;
  }

  return 0;
}

- setReadbufLength:(int)length
{
  if (length > readbuflength)
  {
    length = 0xffff;
    if (!readbuf)
      readbuf = malloc(length);
    else
      readbuf = realloc(readbuf, length);
    readbuflength = length;
  }

  return self;
}

- (int)unixOpen:(char *)path
 /* This will handle opens on directories as well as files */
{
  fileInfo            finfo;
  int                 i;


  pdoserrno = 0;

  for (i = 0; unixFile[i]; i++);
  if (i > 7)
  {
    pdoserrno = PDOSERR_TOOMANYOPEN;
    return 0;
  }

  if ([self getAppleDoubleInfo:&finfo for:path])
  {
    if (finfo.fileType == 0x0f)
    {				/* Opening a directory? */
      unixFile[i] = (FILE *)[self loadDir:path length:&unixFileLength[i]];	/* Address of dir buf */
      unixFileMode[i] = UNIX_DIRECTORY | [computer readByte:0xbf94];
      unixFilePos[i] = 0;
      unixFileNewlineMask[i] = 0;
    }
    else
    {				/* Just a file */
      if (!(unixFile[i] = fopen(path, "r+")) && !(unixFile[i] = fopen(path, "r")))
      {
	pdoserrno = PDOSERR_IOERROR;	/* does global errno get set? */
	return 0;
      }
      unixFileMode[i] = READLOC(0xbf94) | (UNIX_DOCRLFCNV * (unixDoTXTConv && finfo.fileType == 0x04));
      unixFileNewlineMask[i] = 0;
    }
  }
  else
  {
    pdoserrno = [self mapErrnoToPdos:errno];
    return 0;
  }

  return i + 1;
}

- (int)unixSetNewline:(int)fnum :(int)nmask :(int)nchar
{
  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  unixFileNewlineMask[fnum - 1] = nmask;
  unixFileNewlineChar[fnum - 1] = nchar;

  return fnum;
}

- (int)unixRead:(int)fnum :(int)length :(int)bufaddr
 /*
  * This handles reads from unix files into SysMem, including
  * directories and newline mapping 
  */
{
  int                 i, j;
  char               *p;


  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
  {
    i = length;
    [self setReadbufLength:i];
    if (i + unixFilePos[fnum - 1] > unixFileLength[fnum - 1])
      i = unixFileLength[fnum - 1] - unixFilePos[fnum - 1];
    p = (char *)unixFile[fnum - 1];
    memcpy(readbuf, (char *)&p[unixFilePos[fnum - 1]], i);

    for (j = 0; j < i; j++)
    {
      [computer writeByte:bufaddr + j:readbuf[j]];
      if (unixFileNewlineMask[fnum - 1] && ((readbuf[j] & unixFileNewlineMask[fnum - 1]) == unixFileNewlineChar[fnum - 1]))
      {
	j++;
	break;
      }
    }

    unixFilePos[fnum - 1] += j;
  }
  else
  {
    for (j = 0; j < length; j++)
    {
      i = fgetc(unixFile[fnum - 1]);
      if (unixFileMode[fnum - 1] & UNIX_DOCRLFCNV && i == 10)
	i = 13;
      if (i == EOF)
	break;
      WRITELOC(bufaddr + j, i);
      if (unixFileNewlineMask[fnum - 1] && ((i & unixFileNewlineMask[fnum - 1]) == unixFileNewlineChar[fnum - 1]))
      {
	j++;
	break;
      }
    }
  }

  if (!j)
  {
    pdoserrno = PDOSERR_ENDOFFILE;
    return 0;
  }

  return j;
}

- (int)unixWrite:(int)fnum :(int)length :(int)bufaddr
{
  int                 i, j;


  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
  {
    pdoserrno = PDOSERR_ACCESSERROR;
    return 0;
  }
  else
  {
    [self setReadbufLength:length];
    for (i = 0; i < length; i++)
    {
      readbuf[i] = READLOC(bufaddr + i);
      if (unixFileMode[fnum - 1] & UNIX_DOCRLFCNV && readbuf[i] == 13)
	readbuf[i] = 10;
    }
  }

  j = fwrite(readbuf, 1, length, unixFile[fnum - 1]);

  return j;
}

- (int)unixClose:(int)fnum
{
  int                 i, j;

  void                free();


  pdoserrno = 0;

  if (!fnum)
  {
    j = READLOC(0xbf94);
    for (i = 0; i < 8; i++)
      if ((unixFileMode[i] & 0xff) >= j && unixFile[i])
	[self unixClose:i + 1];
    return -1;
  }
  else
  {
    if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
    {
      pdoserrno = PDOSERR_INVALIDREFNUM;
      return 0;
    }

    if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    {
      free(unixFile[fnum - 1]);
    }
    else
    {
      fclose(unixFile[fnum - 1]);
    }

    unixFileMode[fnum - 1] = 0;
    unixFile[fnum - 1] = NULL;
  }

  return fnum;
}

- (int)unixFlush:(int)fnum
{
  fflush(unixFile[fnum - 1]);
  return fnum;
}

- (int)unixSetMark:(int)fnum :(int)where
{
  int                 i;


  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    unixFilePos[fnum - 1] = where;
  else
  {
    if (fseek(unixFile[fnum - 1], where, SEEK_SET))
    {
      pdoserrno = [self mapErrnoToPdos:errno];
      return 0;
    }
    i = ftell(unixFile[fnum - 1]);
    if (i != where)
    {
      pdoserrno = PDOSERR_IOERROR;
      return 0;
    }
  }

  return fnum;
}

- (int)unixGetMark:(int)fnum
{
  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    return unixFilePos[fnum - 1];
  else
    return ftell(unixFile[fnum - 1]);

  return fnum;
}

- (int)unixSetEOF:(int)fnum :(int)length
{
  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
  {
    pdoserrno = PDOSERR_ACCESSERROR;
    return 0;
  }
  else
  {
    [self unixFlush:fnum];
    if (ftruncate(fileno(unixFile[fnum - 1]), length))
    {
      pdoserrno = [self mapErrnoToPdos:errno];
      return 0;
    }
  }

  return fnum;
}

- (int)unixGetEOF:(int)fnum
{
  int                 i;
  fpos_t              pos;


  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1])
  {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }

  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    return unixFileLength[fnum - 1];
  else
  {
    fgetpos(unixFile[fnum - 1], &pos);
    fseek(unixFile[fnum - 1], 0, SEEK_END);
    i = ftell(unixFile[fnum - 1]);
    fsetpos(unixFile[fnum - 1], &pos);
    return i;
  }

  return fnum;
}

- (BOOL)doUnixShare
{
  int                 retAddr = [computer readShort:0x100 + state->sp + 1];
  int                 cmd = [computer readByte:retAddr + 1], parms = [computer readShort:retAddr + 2];
  char               *up;
  char                up2[MAXPATHLEN + 1];
  fileInfo            finfo;
  FILE               *file;
  int                 i, j, l;


/*  fprintf(stderr, "Command: %04X %02X\n", (unsigned) retAddr, (unsigned) cmd); */

  lastCmd = cmd;

 /* Move return address past parameters */
  retAddr += 3;
  WRITELOC(0x100 + state->sp + 1, retAddr & 0xff);
  WRITELOC(0x100 + state->sp + 2, (retAddr >> 8) & 0xff);

 /* Will do JSR to Prodos inside us */
  rom_image[UNIX_COMMAND] = cmd;
  rom_image[UNIX_PARM] = parms & 0xff;
  rom_image[UNIX_PARM + 1] = (parms >> 8) & 0xff;

  state->n_flag = 0;		/* If negative is set, not us, pass it on to Prodos */

  switch (cmd)
  {
  case 0x80:
    if (READLOC(parms + 1) == UNIX_SLOTDRIVE)
    {
      state->c_flag = 1;
      LDA(PDOSERR_NODEVICECNCTD);
      return NO;
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0x81:
    if (READLOC(parms + 1) == UNIX_SLOTDRIVE)
    {
      state->c_flag = 1;
      LDA(PDOSERR_NODEVICECNCTD);
      return NO;
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc0:			/* Create */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      if ([self getAppleDoubleInfo:&finfo for:up])
      {				/* Can't exist */
	state->c_flag = 1;
	LDA(PDOSERR_DUPFILENAME);
	return NO;
      }
      else
      {
	finfo.accessBits = READLOC(parms + 3);
	finfo.fileType = READLOC(parms + 4);
	finfo.auxType = READWORD(parms + 5);
	finfo.storageType = READLOC(parms + 7);
	finfo.dateCreate = READWORD(parms + 8);
	finfo.timeCreate = READWORD(parms + 10);
	if (finfo.fileType == 0x0f)
	{			/* Directory */
	  if (mkdir(up, 0777))
	  {
	    state->c_flag = 1;
	    LDA([self mapErrnoToPdos:errno]);
	    return NO;
	  }
	  [self setAppleDoubleInfo:&finfo for:up];
	  state->c_flag = 0;
	  LDA(0);
	}
	else
	{
	  if (!(file = fopen(up, "w")))
	  {
	    state->c_flag = 1;
	    LDA(PDOSERR_IOERROR);
	    return NO;
	  }
	  fclose(file);
	  [self setAppleDoubleInfo:&finfo for:up];
	  state->c_flag = 0;
	  LDA(0);
	}
      }
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc1:			/* Destroy */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      if (unlink(up))
      {
	state->c_flag = 1;
	LDA([self mapErrnoToPdos:errno]);
	return NO;
      }
      [self removeAppleDoubleInfoFor:up];
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc2:			/* Rename */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      strcpy(up2, up);
      if (up = [self makeUnixPath:READWORD(parms + 3)])
      {
	[self getAppleDoubleInfo:&finfo for:up2];
	if (rename(up2, up))
	{
	  state->c_flag = 1;
	  LDA([self mapErrnoToPdos:errno]);
	  return NO;
	}
	[self setAppleDoubleInfo:&finfo for:up];
	[self removeAppleDoubleInfoFor:up2];
      }
      else
      {
	state->c_flag = 1;
	LDA(PDOSERR_INVALIDPATHSYN);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc3:			/* Set File Info */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      finfo.accessBits = READLOC(parms + 3);
      finfo.fileType = READLOC(parms + 4);
      finfo.auxType = READWORD(parms + 5);
      finfo.dateLastMod = READWORD(parms + 10);
      finfo.timeLastMod = READWORD(parms + 12);
      [self setAppleDoubleInfo:&finfo for:up];
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc4:			/* Get File Info */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      if ([self getAppleDoubleInfo:&finfo for:up])
      {
	WRITELOC(parms + 3, finfo.accessBits);
	WRITELOC(parms + 4, finfo.fileType);
	WRITELOC(parms + 5, finfo.auxType & 0xff);
	WRITELOC(parms + 6, (finfo.auxType >> 8) & 0xff);
	WRITELOC(parms + 7, finfo.storageType);
	WRITELOC(parms + 8, finfo.blocksUsed & 0xff);
	WRITELOC(parms + 9, (finfo.blocksUsed >> 8) & 0xff);
	WRITELOC(parms + 10, finfo.dateLastMod & 0xff);
	WRITELOC(parms + 11, (finfo.dateLastMod >> 8) & 0xff);
	WRITELOC(parms + 12, finfo.timeLastMod & 0xff);
	WRITELOC(parms + 13, (finfo.timeLastMod >> 8) & 0xff);
	WRITELOC(parms + 14, finfo.dateCreate & 0xff);
	WRITELOC(parms + 15, (finfo.dateCreate >> 8) & 0xff);
	WRITELOC(parms + 16, finfo.timeCreate & 0xff);
	WRITELOC(parms + 17, (finfo.timeCreate >> 8) & 0xff);
      }
      else
      {
	state->c_flag = 1;
	LDA([self mapErrnoToPdos:errno]);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc5:			/* Online */
    if (!READLOC(parms + 1))
    {
    /* What does AppleShare do about this?? */
      oldAddress = parms;
      state->n_flag = 1;
    }
    else
    if (READLOC(parms + 1) == UNIX_SLOTDRIVE)
    {
      j = READWORD(parms + 2);
      [self writePString:unixVolumeName to:j];
      WRITELOC(j, READLOC(j) | UNIX_SLOTDRIVE);

      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc6:			/* Set Prefix */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      if (!([self getAppleDoubleInfo:&finfo for:up]))
      {
	state->c_flag = 1;
	LDA(PDOSERR_NOTADIR);
	return NO;
      }
      if (finfo.fileType != 0x0f)
      {
	state->c_flag = 1;
	LDA(PDOSERR_INCOMPFILETYPE);
	return NO;
      }
      up = [self makeCString:READWORD(parms + 1)];
      if (up[0] == '/')
	strcpy(unixPrefix, up);
      else
	strcat(unixPrefix, up);
      if (unixPrefix[strlen(unixPrefix) - 1] != '/')
	strcat(unixPrefix, "/");
      WRITELOC(0xbf9a, 0xff);
      state->c_flag = 0;
      LDA(0);
    }
    else
    {
    /* If prodos changes it, then unset us. */
      state->n_flag = 1;
    }
    return NO;
    break;
  case 0xc7:			/* Get Prefix */
    if (unixPrefix[0])
    {
      [self writePString:unixPrefix to:READWORD(parms + 1)];
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc8:			/* Open */
    if (up = [self makeUnixPath:READWORD(parms + 1)])
    {
      if (!(i = [self unixOpen:up]))
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      WRITELOC(parms + 5, i | 0x80);
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xc9:			/* Newline */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      if (![self unixSetNewline:j :READLOC(parms + 2) :READLOC(parms + 3)])
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xca:			/* Read */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      l = [self unixRead:j :READWORD(parms + 4) :READWORD(parms + 2)];
      WRITELOC(parms + 6, l % 256);
      WRITELOC(parms + 7, l / 256);
      if (!l)
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xcb:			/* Write */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      l = [self unixWrite:j :READWORD(parms + 4) :READWORD(parms + 2)];
      WRITELOC(parms + 6, l % 256);
      WRITELOC(parms + 7, l / 256);
      if (!l)
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xcc:			/* Close */
    j = READLOC(parms + 1);
    if ((!j) || (j & 0x80))
    {
      j &= 0x7f;
      i = [self unixClose:j];
      if (!i)
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      if (i < 0)
      {
	state->n_flag = 1;
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xcd:			/* Flush */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      [self unixFlush:j];
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xce:			/* Set Mark */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      if (![self unixSetMark:j :READWORD(parms + 2) | (READLOC(parms + 4) << 16)])
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xcf:			/* Get Mark */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      i = [self unixGetMark:j];
      WRITELOC(parms + 2, i % 256);
      WRITELOC(parms + 3, i / 256);
      WRITELOC(parms + 4, i >> 16);
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xd0:			/* Set EOF */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      if (![self unixSetEOF:j :READWORD(parms + 2) | (READLOC(parms + 4) << 16)])
      {
	state->c_flag = 1;
	LDA(pdoserrno);
	return NO;
      }
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xd1:			/* Get EOF */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      j &= 0x7f;
      i = [self unixGetEOF:j];
      WRITELOC(parms + 2, i % 256);
      WRITELOC(parms + 3, i / 256);
      WRITELOC(parms + 4, i >> 16);
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xd2:			/* Set Buf */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  case 0xd3:			/* Get Buf */
    if ((j = READLOC(parms + 1)) & 0x80)
    {
      state->c_flag = 0;
      LDA(0);
    }
    else
      state->n_flag = 1;
    return NO;
    break;
  default:			/* We don't want it. Let Prodos handle it */
    state->n_flag = 1;
    return NO;
    break;
  }

  return NO;
}

- (BOOL)unixShare
{
  int                 ret = [self doUnixShare];


/*  fprintf(stderr, "N = %i, C = %i, A = %i\n", state->n_flag, state->c_flag, state->a); */

  if (state->n_flag)
  {				/* Call ProDOS instead */
    state->pc += 3;
  }
  else
  {
    WRITELOC(0xBF30, UNIX_SLOTDRIVE);
    return YES;
  }

  return NO;
}

- unixReturnVal:(int)val
{
  int                 i, j;
  int                 parms, buf;


  if (lastCmd == 0xc6 && !val)
    unixPrefix[0] = 0;
  if (lastCmd == 0xc5 && oldAddress && !val)
  {
    parms = oldAddress;
    buf = READWORD(parms + 2);
    for (i = 0; i < 16; i++)
      if (!(j = READLOC(buf + i * 16)) || (j == UNIX_SLOTDRIVE))
	break;
    if (i < 16)
    {
      [self writePString:unixVolumeName to:buf + i * 16];
      WRITELOC(buf + i * 16, READLOC(buf + i * 16) | UNIX_SLOTDRIVE);
      if (!j && i < 15)
	WRITELOC(buf + i * 16 + 16, 0);
    }
    oldAddress = 0;
  }
  lastCmd = 0;

  return self;
}

- (int)romTwiddle:(AccessType) t :(MemoryElement *) m
{
  int                 address = (m - &state->sysmem[ROM_FOR_SLOT(slot_number)]) & 0xFF;
  int                 val = m->mem.byte;


/*  fprintf(stderr, "Twiddle:%i %04X:%02X\n", t, (unsigned) address, (unsigned) val); */

  if (address == UNIX_START)
    [self unixReturnVal:val];

  if (address >= UNIX_PDOSOP && address <= (UNIX_PDOSOP + 1))
  {
    rom_image[address] = m->mem.byte;
    unixPrefix[0] = 0;
  }

  return m->mem.byte = rom_image[address];
}

/* Various Utility Methods */
- (long)ptime:(time_t) t
{
  struct tm          *clock;
  long                theTime = 0;


  clock = localtime(&t);
  theTime = (clock->tm_year << 9) + ((clock->tm_mon + 1) << 5) + (clock->tm_mday << 0);
  theTime <<= 16;
  theTime |= (clock->tm_hour << 8) + (clock->tm_min << 0);

  return theTime;
}

- (char *)makeCString:(int)fnameLoc
{
  static char         pname[129];
  int                 i, j;


  j = READLOC(fnameLoc);
  for (i = 0; i < j; i++)
    pname[i] = LOWER(READLOC(fnameLoc + i + 1) & 0x7f);
  pname[i] = 0;

  return pname;
}

- writePString:(char *)fname to:(int)where
{
  int                 i;


  WRITELOC(where, strlen(fname));

  for (i = 0; fname[i]; i++)
    WRITELOC(where + 1 + i, UPPER(fname[i]));

  return self;
}

/* OpenPanel Delegate Methods */

- (BOOL)panel:sender filterFile:(const char *)filename inDirectory:(const char *)directory
{
  struct stat         sbuf;
  static char         buf[MAXPATHLEN + 1];


  sprintf(buf, "%s/%s", directory, filename);
  stat(buf, &sbuf);
  if (sbuf.st_mode & S_IFDIR)
    return YES;

  return NO;
}

@end
