/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import "WozView.h"
#import "memory.h"
#import "hgrconv.h"
#import "drawstring.h"
#import "Computer.h"
#import <string.h>
#import <assert.h>

@implementation WozView


- initFrame:(const NXRect *)frameRect
{
  static const float  doublewidth[] = {2.0, 0.0, 0.0, 1.0, 0.0, 0.0};

  [super initFrame:frameRect];

  [self setFlipped:NO];

 /* Load up the Apple ][ fonts */
  font40 = [Font newFont:"Apple2" size:16 matrix:doublewidth];
  if (font40 == nil)
  {
    NXRunAlertPanel(NULL, "Unable to load font \"Apple2\".  Using Ohlfs.  "
		    "This will look really bad.",
		    NULL, NULL, NULL);
    font40 = [Font newFont:"Ohlfs" size:16 matrix:NX_IDENTITYMATRIX];
  }
  font80 = [Font newFont:"Apple2" size:16 matrix:NX_IDENTITYMATRIX];
  if (font80 == nil)
  {
  /*
   * We should have already printed the warning message for
   * font40, so don't bother notifying them again. 
   */
    font80 = [Font newFont:"Ohlfs" size:12 matrix:NX_IDENTITYMATRIX];
  }

 /* Initialize the shadow screens (not strictly necessary). */
  memset(text_shadow, 0, sizeof text_shadow);
  memset(hgr_shadow, 0, sizeof hgr_shadow);

 /* Since we will be repeatedly focused on, allocate a gstate. */
  [self allocateGState];

 /* No memory image or display flags yet. */
  display_flags = NULL;
  sysmem = NULL;
  hgrbuf = NULL;
  bitmapRep = nil;

  return self;
}

- free
{
  [bitmapRep free];
  NX_FREE(real_hgrbuf);
  return[super free];
}

- awake
{
  [super awake];

  [self setFlipped:NO];

 /* Initialize the shadow screens (not strictly necessary). */
  memset(text_shadow, 0, sizeof text_shadow);
  memset(hgr_shadow, 0, sizeof hgr_shadow);

 /* Since we will be repeatedly focused on, allocate a gstate. */
  [self allocateGState];

 /* No memory image or display flags yet. */
  display_flags = NULL;
  sysmem = NULL;
  hgrbuf = NULL;
  bitmapRep = nil;

  return self;
}

- read:(NXTypedStream *) stream
{
  [super read:stream];
  NXReadTypes(stream, "i", &we_are_color);
  bitmapRep = NXReadObject(stream);
  font40 = NXReadObject(stream);
  font80 = NXReadObject(stream);
  computer = NXReadObject(stream);
  return self;
}

- write:(NXTypedStream *) stream
{
  [super write:stream];
  NXWriteTypes(stream, "i", &we_are_color);
  NXWriteObject(stream, bitmapRep);
  NXWriteObject(stream, font40);
  NXWriteObject(stream, font80);
  NXWriteObjectReference(stream, computer);
  return self;
}

- awakeFromNib
{

 /*
  * Set up the NXBitmapImageRep to hold the hires/double-hires
  * data. 
  */
  we_are_color = [self shouldDrawColor];

  if (!we_are_color)
  {
  /* Allocate space for frame buffer and clear to black. */
    NX_MALLOC(real_hgrbuf, unsigned char, 384 * 560 / 4);
    hgrbuf = real_hgrbuf;
    memset(hgrbuf, 0, 384 * 560 / 4);

    bitmapRep = [[NXBitmapImageRep alloc] initData:hgrbuf
		 pixelsWide:560 pixelsHigh:384 bitsPerSample:2
		 samplesPerPixel:1 hasAlpha:NO isPlanar:NO
		 colorSpace:NX_OneIsWhiteColorSpace
		 bytesPerRow:140 bitsPerPixel:2];
  }
  else
  {
    int                 i;

  /* Allocate space for frame buffer and clear to black. */
  /*
   * Make sure that the high nibble is 0xF, as per NeXT
   * documentation. 
   */
    NX_MALLOC(real_hgrbuf, unsigned char, 384 * 560 * 2 + 15);

  /* Make sure that hgrbuf is aligned on a cache line. */
    hgrbuf = (unsigned char *)(((unsigned long)real_hgrbuf) & ~15);

    for (i = 384 * 560 / 2; i > 0; i--)
/*	((unsigned long *)hgrbuf)[i] = 0xF000F000L; */
      ((unsigned long *)hgrbuf)[i] = 0x000F000FL;

    bitmapRep = [[NXBitmapImageRep alloc] initData:hgrbuf
		 pixelsWide:560 pixelsHigh:384 bitsPerSample:4
		 samplesPerPixel:3 hasAlpha:NO isPlanar:NO
		 colorSpace:NX_RGBColorSpace
		 bytesPerRow:(280 * 2 * 2) bitsPerPixel :16];
  }

  return self;
}

static const int    text_base_offset[] = {
					  0x000, 0x080, 0x100, 0x180, 0x200, 0x280, 0x300, 0x380,
					  0x028, 0x0A8, 0x128, 0x1A8, 0x228, 0x2A8, 0x328, 0x3A8,
					  0x050, 0x0D0, 0x150, 0x1D0, 0x250, 0x2D0, 0x350, 0x3D0
};


- update40columnText:(const MemoryElement *) screenbase :(double)now
  isMixed:(BOOL)mixed isAltChar:(BOOL)altcharset
  redrawAll:(BOOL)redraw_all
{
#define MAX_40_COL_STRINGS (((40 / 2) + 1) * 24)	/* max # possible */
  unsigned char       nstring[40 * 24], istring[40 * 24];
  unsigned char      *np = &nstring[0], *ip = &istring[0];
  unsigned char      *shadow;
  const MemoryElement *m;
  NXPoint             nxy[MAX_40_COL_STRINGS], ixy[MAX_40_COL_STRINGS];
  int                 nlen[MAX_40_COL_STRINGS], ilen[MAX_40_COL_STRINGS];
  int                 nindices[MAX_40_COL_STRINGS], iindices[MAX_40_COL_STRINGS];
  int                 nnum = 0, inum = 0;
  int                 x, y;
  BOOL                is_continuing_string, last_was_inverse = FALSE;
  int                 flashfactor = (((unsigned long)(now * 4)) & 1) ? 0x40 : -0x40;
  int                 altbits = (altcharset ? 0x80 : 0x0);

#undef MAX_40_COL_STRINGS

  for (y = (mixed ? 20 : 0); y < 24; y++)
  {
    shadow = &text_shadow[y][0];
    m = &screenbase[text_base_offset[y]];
    is_continuing_string = FALSE;

    for (x = 0; x < 40; m++, x++)
    {
      int                 c = m->mem.byte;

    /* Make flashing chars be either normal or inverse */
      if (!altcharset && (c & 0xC0) == 0x40)
	c += flashfactor;

      if (c == shadow[x] && !redraw_all)
	is_continuing_string = FALSE;
      else
      {
	BOOL                is_normal = ((c & 0x80)
/*				|| (altcharset && (c & 0xE0) == 0x40) */
	);

	shadow[x] = c;

	if (is_normal)		/* Is this a non-inverse string? */
	{
	  if (is_continuing_string && !last_was_inverse)
	    nlen[nnum - 1]++;
	  else
	  {
	    nlen[nnum] = 1;
	    nindices[nnum] = (np - nstring);
	    nxy[nnum].x = x * 14;
	    nxy[nnum].y = (384 - 16 - y * 16);
	    nnum++;
	  }

	  *np++ = (c & 0x7F) | (altbits & ~(c & 0x80));
	  last_was_inverse = FALSE;
	}
	else			/* This is an inverse string. */
	{
	  if (is_continuing_string && last_was_inverse)
	    ilen[inum - 1]++;
	  else
	  {
	    ilen[inum] = 1;
	    iindices[inum] = (ip - istring);
	    ixy[inum].x = x * 14;
	    ixy[inum].y = (384 - 16 - y * 16);
	    inum++;
	  }
	  *ip++ = c | altbits;
	  last_was_inverse = TRUE;
	}
	is_continuing_string = TRUE;
      }
    }
  }

  if (nnum || inum)
  {
    [self lockFocus];

    [font40 set];

  /* Draw non-inverse strings, if there are any. */
    if (nnum)
      draw_str(nstring, nlen, nindices, 14, (const float *)nxy, NX_WHITE,
	       NX_BLACK, nnum, np - nstring, nnum * 2);

  /* Draw inverse strings, if there are any. */
    if (inum)
      draw_str(istring, ilen, iindices, 14, (const float *)ixy, NX_BLACK,
	       NX_WHITE, inum, ip - istring, inum * 2);

    [self unlockFocus];
  }

  return self;
}


- update80columnText:(const MemoryElement *) screen_even_columns
   :(const MemoryElement *) screen_odd_columns
   :(double)now isMixed:(BOOL)mixed isAltChar:(BOOL)altcharset
  redrawAll:(BOOL)redraw_all
{
#define MAX_80_COL_STRINGS (((80 / 2) + 1) * 24)	/* max # possible */
  unsigned char       nstring[80 * 24], istring[80 * 24];
  unsigned char      *np = &nstring[0], *ip = &istring[0];
  unsigned char      *shadow;
  const MemoryElement *m1, *m2;
  NXPoint             nxy[MAX_80_COL_STRINGS], ixy[MAX_80_COL_STRINGS];
  int                 nlen[MAX_80_COL_STRINGS], ilen[MAX_80_COL_STRINGS];
  int                 nindices[MAX_80_COL_STRINGS], iindices[MAX_80_COL_STRINGS];
  int                 nnum = 0, inum = 0;
  int                 x, y;
  BOOL                is_continuing_string, last_was_inverse = FALSE;
  int                 flashfactor = (((unsigned long)(now * 4)) & 1) ? 0x40 : -0x40;
  int                 altbits = (altcharset ? 0x80 : 0x0);

#undef MAX_80_COL_STRINGS

  for (y = (mixed ? 20 : 0); y < 24; y++)
  {
    shadow = &text_shadow[y][0];
    m1 = &screen_odd_columns[text_base_offset[y]];
    m2 = &screen_even_columns[text_base_offset[y]];
    is_continuing_string = FALSE;

    for (x = 0; x < 80; x++)
    {
      int                 c = (x & 1) ? (m1++)->mem.byte : (m2++)->mem.byte;

    /* Make flashing chars be either normal or inverse */
      if (!altcharset && (c & 0xC0) == 0x40)
	c += flashfactor;

      if (c == shadow[x] && !redraw_all)
	is_continuing_string = FALSE;
      else
      {
	BOOL                is_normal = ((c & 0x80)
/*				|| (altcharset && (c & 0xE0) == 0x40) */
	);

	shadow[x] = c;

	if (is_normal)		/* Is this not an inverse string? */
	{
	  if (is_continuing_string && !last_was_inverse)
	    nlen[nnum - 1]++;
	  else
	  {
	    nlen[nnum] = 1;
	    nindices[nnum] = (np - nstring);
	    nxy[nnum].x = x * 7;
	    nxy[nnum].y = (384 - 16 - y * 16);
	    nnum++;
	  }
	  *np++ = (c & 0x7F) | (altbits & ~(c & 0x80));
	  last_was_inverse = FALSE;
	}
	else			/* This is an inverse string. */
	{
	  if (is_continuing_string && last_was_inverse)
	    ilen[inum - 1]++;
	  else
	  {
	    ilen[inum] = 1;
	    iindices[inum] = (ip - istring);
	    ixy[inum].x = x * 7;
	    ixy[inum].y = (384 - 16 - y * 16);
	    inum++;
	  }
	  *ip++ = c | altbits;
	  last_was_inverse = TRUE;
	}
	is_continuing_string = TRUE;
      }
    }
  }

  if (nnum || inum)
  {
    [self lockFocus];

    [font80 set];

  /* Draw non-inverse strings, if there are any. */
    if (nnum)
      draw_str((const char *)nstring, nlen, nindices, 7,
	       (const float *)nxy, NX_WHITE, NX_BLACK, nnum,
	       np - nstring, nnum * 2);

  /* Draw inverse strings, if there are any. */
    if (inum)
      draw_str((const char *)istring, ilen, iindices, 7,
	       (const float *)ixy, NX_BLACK, NX_WHITE, inum,
	       ip - istring, inum * 2);

    [self unlockFocus];
  }

  return self;
}



/* This routine updates either normal or double hires.  If you want to
 * display normal hires, doubleHires == NO && auxscreen == NULL.
 */

- updateHires:(const MemoryElement *) mainscreen
   :(const MemoryElement *) auxscreen
  isDouble:(BOOL)doubleHires
  isMixed:(BOOL)mixed
  isMono:(BOOL)monochrome
  redrawAll:(BOOL)drawAll
{
  NXRect              nxr[MAX_UPDATE_RECTS];
  int                 num_rects;

  if (!doubleHires)
  {
    if (!we_are_color)
      num_rects = hgr_update_framebuffer(mainscreen, hgr_shadow, hgrbuf,
					 nxr, drawAll, mixed, monochrome);
    else
      num_rects = hgr_update_color_framebuffer(mainscreen, hgr_shadow,
					       hgrbuf, nxr, drawAll,
					       mixed, monochrome);
  }
  else
    num_rects = dblhgr_update_framebuffer(mainscreen, auxscreen, hgr_shadow,
					  auxhgr_shadow, hgrbuf,
					  nxr, drawAll, mixed, monochrome);

  if (num_rects > 0)
  {
    [self lockFocus];
    NXRectClipList(&nxr[0], num_rects);	/* Just to optimize blitting */
    [bitmapRep draw];
    [self unlockFocus];
  }

  return self;
}


- updateLores:(const MemoryElement *) mainscreen
   :(const MemoryElement *) auxscreen :(BOOL)mixed :(BOOL)drawAll
{
  return self;
}



- updateScreen:(BOOL)drawAll :(double)now
{
 /* FIXME!  This is all a BIG hack! */


  if (last_flags.text != display_flags->text
      || last_flags.mixed != display_flags->mixed
      || ((display_flags->text || display_flags->mixed)
	  && (last_flags.col80 != display_flags->col80
	      || last_flags.altcharset != display_flags->altcharset))
      || (display_flags->hires
	  && (last_flags.col80 != display_flags->col80
	      || last_flags.an3 != display_flags->an3)))
    drawAll = TRUE;

  last_flags = *display_flags;

  if (!last_flags.text && last_flags.hires)
  {
    int                 index = last_flags.page2 ? HIRES2 : HIRES1;

  /* Double hires?  FIXME, this is wrong. */
    if (last_flags.col80 && !last_flags.an3)
    {
      [self updateHires:&sysmem[HIRES1] :&sysmem[AUXHIRES1]
       isDouble:YES isMixed:last_flags.mixed isMono:YES
       redrawAll:drawAll];
    }
    else
      [self updateHires:&sysmem[index] :NULL isDouble:NO
       isMixed:last_flags.mixed isMono:NO redrawAll:drawAll];
  }

  if (last_flags.text || last_flags.mixed)
  {
    int                 index;

    index = (!last_flags.store80 && last_flags.page2) ? TEXT2 : TEXT1;


    if (!last_flags.col80)
    {
      [self update40columnText:&sysmem[index]
        :now
       isMixed:!last_flags.text && last_flags.mixed
       isAltChar:display_flags->altcharset
       redrawAll:drawAll];
    }
    else
    {
      [self update80columnText:&sysmem[index + AUXOFFSET] :&sysmem[index]
        :now
       isMixed:!last_flags.text && last_flags.mixed
       isAltChar:display_flags->altcharset
       redrawAll:drawAll];
    }
  }

  [[self window] flushWindow];
 /* PSflushgraphics(); */

  return self;
}

#ifdef CUTPASTE
- setSel:(int)start :(int)end
{
  int                 i;
  int                 x, y;
  int                 w = 7 + 7 * !Vid80On;
  int                 ll = 40 + 40 * Vid80On;
  int                 d;
  int                 c;
  NXRect              r;
  NXPoint             p;


  if (end < start)
  {
    i = start;
    start = end;
    end = i;
  }
  start = start >= 0 ? (start <= ll * 40 ? start : ll * 40) : 0;
  end = end >= 0 ? (end <= ll * 40 ? end : ll * 40) : 0;

  [self lockFocus];
  if (Vid80On)
  {
    for (i = start; i < end; i++)
    {
      if (!(i >= currentSelStart && i < currentSelEnd))
      {
	x = (i % ll) * w;
	y = (23 - i / ll) * 16;
	if (TextVisible(x / w, 23 - y / 16))
	{
	  d = wuzthere[(23 - (y / 16)) * 80 + (x / w)];
	  c = d & 0xff;
	  r.origin.x = (c % 32) * 7 + 224 * ((d & MODE_MOUSE) > 0);
	  r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
	  r.size.width = 7;
	  r.size.height = 16;
	  p.x = x;
	  p.y = y;
	  [col80imagesHL composite:NX_COPY fromRect:&r toPoint:&p];
	}
	else
	  start++;
      }
    }

    for (i = currentSelStart; i < currentSelEnd; i++)
    {
      if (!(i >= start && i < end))
      {
	x = (i % ll) * w;
	y = (23 - i / ll) * 16;
	if (TextVisible(x / w, 23 - y / 16))
	{
	  d = wuzthere[(23 - (y / 16)) * 80 + (x / w)];
	  c = d & 0xff;
	  r.origin.x = (c % 32) * 7 + 224 * ((d & MODE_MOUSE) > 0);
	  r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
	  r.size.width = 7;
	  r.size.height = 16;
	  p.x = x;
	  p.y = y;
	  [col80images composite:NX_COPY fromRect:&r toPoint:&p];
	}
      }
    }
  }
  else
  {
    for (i = start; i < end; i++)
    {
      if (!(i >= currentSelStart && i < currentSelEnd))
      {
	x = (i % ll) * w;
	y = (23 - i / ll) * 16;
	if (TextVisible(x / w, 23 - y / 16))
	{
	  d = wuzthere[(23 - (y / 16)) * 80 + (x / 14) * 2];
	  c = d & 0xff;
	  r.origin.x = (c % 32) * 14 + 448 * ((d & MODE_MOUSE) > 0);
	  r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
	  r.size.width = 14;
	  r.size.height = 16;
	  p.x = x;
	  p.y = y;
	  [col40imagesHL composite:NX_COPY fromRect:&r toPoint:&p];
	}
	else
	  start++;
      }
    }

    for (i = currentSelStart; i < currentSelEnd; i++)
    {
      if (!(i >= start && i < end))
      {
	x = (i % ll) * w;
	y = (23 - i / ll) * 16;
	if (TextVisible(x / w, 23 - y / 16))
	{
	  d = wuzthere[(23 - (y / 16)) * 80 + (x / 14) * 2];
	  c = d & 0xff;
	  r.origin.x = (c % 32) * 14 + 448 * ((d & MODE_MOUSE) > 0);
	  r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
	  r.size.width = 14;
	  r.size.height = 16;
	  p.x = x;
	  p.y = y;
	  [col40images composite:NX_COPY fromRect:&r toPoint:&p];
	}
      }
    }
  }

  PSflushgraphics();
  [self unlockFocus];

  currentSelStart = start;
  currentSelEnd = end;

  return self;
}

- getSel:(int *)start :(int *)end
{
  *start = currentSelStart;
  *end = currentSelEnd;

  return self;
}

- (int)getSubstring:(char *)buf start:(int)startPos length:(int)numChars
{
  int                 ll = 40 + 40 * Vid80On;
  int                 i, j;
  static char         pchar[] =
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
  " !\"#$%&'()*+,-./0123456789:;<=>?"
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
  "`abcdefghijklmnopqrstuvwxyz{|}~ "
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
  " !\"#$%&'()*+,-./0123456789:;<=>?"
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
  "`abcdefghijklmnopqrstuvwxyz{|}~ ";


  if (startPos + numChars > ll * 24)
    numChars = ll * 24 - startPos;

  for (j = 0, i = startPos; i < startPos + numChars; i++, j++)
  {
    buf[j] = pchar[wuzthere[i * (1 + !Vid80On)] & 0xff];
    if (!((i + 1) % ll))
    {
      j++;
      buf[j] = '\n';
    }
  }

  return j;
}

- mouseDown:(NXEvent *)theEvent
{
  int                 sx, sy;
  int                 f, t;
  int                 i;
  int                 w = 7 + 7 * !Vid80On;
  int                 ll = 40 + 40 * Vid80On;


  clickState = theEvent->data.mouse.click;
  sx = theEvent->location.x;
  sy = theEvent->location.y;
  dragStart = f = (sx / w) + (23 - sy / 16) * ll;

  switch (clickState)
  {
  case 1:			/* Single click */
    [self setSel:f :f];
    break;

  case 2:			/* Double click */
    for (i = f; i >= 0; i--)
      if ((wuzthere[i * (1 + !Vid80On)] & 0x7f) == 0x20)
	break;
    f = i + 1;
    for (i = dragStart; i < ll * 24; i++)
      if ((wuzthere[i * (1 + !Vid80On)] & 0x7f) == 0x20)
	break;
    t = i;
    [self setSel:f :t];
    break;

  case 3:
    f = (f / ll) * ll;
    t = f + ll;
    [self setSel:f :t];
    break;
  }

  do
  {
    theEvent = [NXApp getNextEvent:NX_MOUSEDRAGGEDMASK | NX_MOUSEENTEREDMASK | NX_MOUSEEXITEDMASK | NX_MOUSEUPMASK];
    if (theEvent->type == NX_MOUSEDRAGGED)
      [self mouseDragged:theEvent];
  } while (theEvent->type != NX_MOUSEUP);

  return self;
}

- mouseDragged:(NXEvent *)theEvent
{
  int                 cx = theEvent->location.x;
  int                 cy = theEvent->location.y;
  int                 f, t;
  int                 i;
  int                 w = 7 + 7 * !Vid80On;
  int                 ll = 40 + 40 * Vid80On;


  f = dragStart;
  t = ((cx + w / 2) / w) + (23 - cy / 16) * ll;
  if (t < f)
  {
    i = t;
    t = f;
    f = i;
  }

  switch (clickState)
  {
  case 1:
    [self setSel:f :t];
    break;

  case 2:
    for (i = f; i >= 0; i--)
      if ((wuzthere[i * (1 + !Vid80On)] & 0x7f) == 0x20)
	break;
    f = i + 1;
    for (i = t; i < ll * 24; i++)
      if ((wuzthere[i * (1 + !Vid80On)] & 0x7f) == 0x20)
	break;
    t = i;
    [self setSel:f :t];
    break;

  case 3:
    f = (f / ll) * ll;
    t = (t / ll) * ll + ll;
    [self setSel:f :t];
    break;
  }

  return self;
}

static BOOL
matchPasteType(const NXAtom * types, NXAtom type)
{
  if (types)
    while (*types)
      if (*types++ == type)
	return YES;
  return NO;
}

- paste:sender
{
  id                  myPB;
  char               *data;
  char               *keyd;
  int                 len;
  int                 i, j;

  void               *malloc();


  [self setSel:0 :0];

  if (matchPasteType([(myPB = [Pasteboard new]) types], NXAsciiPboardType))
  {
    if ([myPB readType:NXAsciiPboardType data:&data length:&len])
    {
      for (i = 0, j = 0; i < len; i++)
	if (data[i] == '\n')
	  j++;
      if (j > maxPLen)
      {
	i = NXRunAlertPanel("Whole lot of data", "The Pasteboard contains %i line%s. Paste it anyway?", "Paste it", "No Don't", NULL, j, j == 1 ? "" : "s");
	if (i != NX_ALERTDEFAULT)
	  return self;
      }

      if (keyboardPos)
      {
	i = keyboardPos + len;
	keyd = malloc(i < 32 ? 32 : i);
	strncpy(keyd, (char *)keyboardBuffer, keyboardPos);
      }
      else
      {
	keyd = malloc(len < 32 ? 32 : i);
      }
      strncpy(&keyd[keyboardPos], data, len);
      for (i = 0; transNL && i < len; i++)
	if (keyd[keyboardPos + i] == '\n')
	  keyd[keyboardPos + i] = '\r';

      keyboardPos += len;
      vm_deallocate((vm_task_t) task_self(), (vm_address_t) data, (vm_size_t) len);
      if (keyboardBuffer)
	free(keyboardBuffer);
      keyboardBuffer = (unsigned char *)keyd;
    }
    else
    {
    /* Book says to pop up a panel here */
    }
  }

  return self;
}

- (BOOL)writeSelectionToPasteboard:pboard types:(NXAtom *) types
{
  const char         *wtypes[3];
  int                 i = 0, length;
  int                 j, k;
  static char         data[2000];


  if (!matchPasteType(types, NXAsciiPboardType))
    return NO;

  wtypes[i++] = NXAsciiPboardType;
  [pboard declareTypes:wtypes num:i owner:NULL];

  [self getSel:&j:&k];
  length = [self getSubstring:data start:j length:k - j];

  [pboard writeType:NXAsciiPboardType data:data length:length];

  return YES;
}

- copy:sender
{
  int                 i = 0;
  const char         *types[3];



  types[i++] = NXAsciiPboardType;
  [self writeSelectionToPasteboard:[Pasteboard new] types:types];

  return self;
}

- validRequestorForSendType:(NXAtom) typeSent andReturnType:(NXAtom) typeReturned
{
  int                 i, j;


  [self getSel:&i:&j];
  if ((typeSent == NXAsciiPboardType || !typeSent) && !typeReturned)
    if ((i != j) || typeSent == NULL)
      return self;

  return[super validRequestorForSendType:typeSent andReturnType:typeReturned];
}

- selectAll:sender
{
  [self setSel:0 :1920];

  return self;
}

- cut:sender
{
  return[self copy:sender];
}

#endif

- drawSelf:(const NXRect *)rects :(int)rectCount
{
  if (sysmem == NULL)
  {
    PSsetgray(NX_BLACK);
    NXRectFill(&bounds);
  }
  else
  {
    [self updateScreen:YES :0.0];
  }
  return self;
}


- setDisplayInfo:(StatusFlags *) f :(const MemoryElement *) s
{
  display_flags = f;
  sysmem = s;
  return self;
}


- computer
{
  return computer;
}


- (BOOL)acceptsFirstResponder
{
  return YES;
}


@end
