/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


/*
 *     execute.h
 */

#ifndef __EXECUTE_H__
#define __EXECUTE_H__

#ifndef USE21
#import <mach/cthreads.h>
#else
#import <cthreads.h>
#endif
#import "opcodes.h"
#import <objc/objc.h>   /* typedef BOOL */

#if defined(mc68000) && !defined __BIG_ENDIAN__
#define __BIG_ENDIAN__
#endif


#ifdef __BIG_ENDIAN__
typedef struct {
  unsigned short opcode;
  union {
    unsigned short s;
    struct {
      char filler;
      unsigned char c;
    } c;
  } u;
} Instruction;

typedef union {
  struct {
    char filler1, filler2;
    unsigned char hi;
    unsigned char c;
  } c;
  struct {
    unsigned short filler;
    unsigned short s;
  } s;
  unsigned long l;
} Reg;

#else

typedef struct {
  union {
    struct {
      unsigned char c;
      char filler;
    } c;
    unsigned short s;
  } u;
  unsigned short opcode;
} Instruction;

typedef union {
  struct {
    unsigned char c;
    unsigned char hi;
    char filler2, filler3;
  } c;
  struct {
    unsigned short s;
    unsigned short filler;
  } s;
  unsigned long l;
} Reg;

#endif  /* not __BIG_ENDIAN__ */


/* Bits in the interrupt_pending byte; set the appropriate bits to signal
 * you want to cause an interrupt.  Before doing this, make sure to lock
 * the interrupt_mutex (and unlock it right after the right bit is set.)
 */
enum {
    IRQ_PENDING=1,
    NMI_PENDING=2,
    RESET_PENDING=4,
    POWEROFF_PENDING=8,
    PAUSE_PENDING=16
};
  
typedef struct {
  struct MemInfo {
    unsigned char byte;         /* Actual value of Apple ][ memory location  */
    unsigned read_special  :1;  /* Boolean: special action on read?          */
    unsigned write_special :1;  /* Boolean: special action on write?         */
    unsigned read_func     :11; /* # of func to call on read                 */
    unsigned write_func    :11; /* # of func to call on write                */
  } mem;
  Instruction inst;             /* Compiled instruction for this address.    */
} MemoryElement;

typedef enum { READ, WRITE, BOTH } AccessType;

typedef struct {
  /* Display */
  unsigned text        :1;
  unsigned mixed       :1;
  unsigned page2       :1;
  unsigned hires       :1;
  unsigned col80       :1;
  unsigned altcharset  :1;

  /* Language card */
  unsigned BSRbank2    :1;
  unsigned BSRreadram  :1;
  unsigned BSRwriteram :1;

  /* Extended memory */
  unsigned alt_zp      :1;
  unsigned aux_read    :1;
  unsigned aux_write   :1;
  unsigned lc_read     :1;
  unsigned lc_write    :1;
  unsigned bank2enable :1;
  unsigned store80     :1;
  unsigned intCXrom    :1;
  unsigned slotC3rom   :1;

  /* Annunciators */
  unsigned an0         :1;
  unsigned an1         :1;
  unsigned an2         :1;
  unsigned an3         :1;

  /* Game controller timer bits */
  unsigned gc0         :1;
  unsigned gc1         :1;
  unsigned gc2         :1;
  unsigned gc3         :1;

  /* Pushbutton states. */
  unsigned pb0         :1;
  unsigned pb1         :1;
  unsigned pb2         :1;
} StatusFlags;

enum {
   CONDITION_MACHINE_PAUSED, CONDITION_MACHINE_RUNNING
};

#define KEY_BUFFER_SIZE 32

typedef struct MachineStateStruct {
  /* Apple // memory.  See memory.h for more info. */
  MemoryElement *sysmem;
  MemoryElement **read_base_p, **write_base_p;

  /* Generic display and other flags. */
  StatusFlags flags;

  /* Keyboard   */
  struct {
    unsigned char buffer[KEY_BUFFER_SIZE];
    int head, tail;
    BOOL key_ready;
    BOOL key_down;
    mutex_t mutex;
  } keyboard;

  /* Approximate # of cycles elapsed.  Does not do page boundary crossings. */
  unsigned long cycles;

  /* Paddle information. */
  volatile unsigned char pdl[2];        /* Current pdl value.           */
  unsigned long pdl_timeout[2];         /* Cycle at which bit drops low */

  /* Speaker private data */
  void *speaker_state;

  /* Ways to communicate with the emulator.  Interrupts are only checked
   * for on instructions which can cause the PC to become smaller.
   */
  volatile unsigned char interrupt_pending;    /* Set from another thread. */
  volatile unsigned char debug;                /* Set from another thread. */
  mutex_t interrupt_mutex;
  id	  pauseLock;

  /* These are for handling trapped softswitches/routines.  Trapped addresses
   * have associated with them a "function number" which specifies which
   * function pointer in the appropriate array to invoke.
   */
  int num_rw_funcs, num_routine_funcs;
  int (**rw_func_list)(AccessType, MemoryElement *,
		       struct MachineStateStruct *);
  BOOL (**routine_func_list)(struct MachineStateStruct *);

  /* These are the id's of the peripheral cards, or nil if the slot is empty.*/
  id peripheral_card[8];    /* slot 0 should always be nil! */

  /* A list of methods used when a peripheral card traps something. */
  int num_card_selectors;
  SEL *card_selector_list;

  /* These are NOT guaranteed to be valid when a softswitch handling
   * routine is called.  However, they are valid for PC handlers.
   * Note that on input to your handler cc bits are only guaranteed to be
   * either 0 or nonzero, but to assure compatibility you are not allowed to
   * assign any value other than 0 or 1 to them.
   */
  unsigned short pc;
  unsigned char sp, a, x, y;
  unsigned char b_flag, c_flag, d_flag, i_flag, n_flag, v_flag, z_flag;
} MachineState;


extern void exec (MachineState *state);

#endif  /* not __EXECUTE_H__ */
