/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#import <objc/Object.h>
#import <appkit/nextstd.h>
#import <stdlib.h>
#import <stdio.h>
#import "execute.h"
#import "opcodes.h"
#import "memory.h"
#import "handlers.h"
#import "softswitches.h"
#import "sound.h"
/*#import "disk.h" */
#import "compile.h"

static const RWHandlerTableEntry rw_handlers[] = {
						  {0x0C000, 0x0C00F, READ, read_softswitch_C000_C00F},
						  {0x0C000, 0x0C00F, WRITE, write_softswitch_C000_C00F},
						  {0x0C010, 0x0C010, BOTH, softswitch_C010},
						  {0x0C011, 0x0C018, READ, softswitch_C011_C018},
						  {0x0C019, 0x0C019, READ, softswitch_C019},
						  {0x0C01A, 0x0C01F, READ, softswitch_C01A_C01F},
						  {0x0C050, 0x0C057, BOTH, softswitch_C050_C057},
						  {0x0C058, 0x0C05F, BOTH, softswitch_C058_C05F},
						  {0x0C061, 0x0C063, READ, softswitch_C061_C063},
						  {0x0C064, 0x0C067, READ, softswitch_C064_C067},
						  {0x0C070, 0x0C070, READ, softswitch_C070},
						  {0x0C080, 0x0C08F, BOTH, softswitch_C080_C08F},
						  {0x0C030, 0x0C030, BOTH, click_speaker},
};

static const RoutineHandlerTableEntry routine_handlers[] = {
/*  { 0x0BD00, 0x0BD00, rwts }, */
};


void
add_routine_handler(MachineState * state, int low_address, int high_address,
		    BOOL (*func) (MachineState *))
{
  MemoryElement      *m;
  int                 i, funcnum;

 /*
  * First see if this function is already in our list of
  * functions. 
  */
  for (funcnum = state->num_routine_funcs - 1; funcnum >= 0; funcnum--)
    if (state->routine_func_list[funcnum] == func)
      break;

 /* If not, grow the list and add this function to it. */
  if (funcnum < 0)
  {
    funcnum = state->num_routine_funcs++;
    NX_REALLOC(state->routine_func_list,
	       typeof(state->routine_func_list[0]),
	       state->num_routine_funcs);
    state->routine_func_list[funcnum] = func;
  }

 /* Now install this routine in all the specified addresses. */
  m = &state->sysmem[low_address];
  for (i = high_address - low_address; i >= 0; m++, i--)
  {
    if (m->inst.opcode != OPC_NOT_COMPILED
	&& m->inst.opcode != OPC_TRAPTHIS
	&& m->inst.opcode != OPC_CARDTRAP)
      decompile_instruction(state, m);
    m->inst.opcode = OPC_TRAPTHIS;
    m->inst.u.s = funcnum;
  }
}


void
remove_routine_handlers(MachineState * m, int low_address, int high_address)
{
  MemoryElement      *mem;
  int                 i;

  mem = &m->sysmem[low_address];
  for (i = high_address - low_address; i >= 0; mem++, i--)
    if (mem->inst.opcode == OPC_TRAPTHIS || mem->inst.opcode == OPC_CARDTRAP)
      mem->inst.opcode = OPC_NOT_COMPILED;
}


void
add_address_handler(MachineState * state, int low_address, int high_address,
		    int (*func) (AccessType, MemoryElement *, MachineState *),
		    AccessType type)
{
  MemoryElement      *m;
  BOOL                do_read = (type != WRITE);	/* Either READ or BOTH */
  BOOL                do_write = (type != READ);	/* Either WRITE or BOTH */
  int                 i, funcnum;

 /*
  * First see if this function is already in our list of
  * functions. 
  */
  for (funcnum = state->num_rw_funcs - 1; funcnum >= 0; funcnum--)
    if (state->rw_func_list[funcnum] == func)
      break;

 /* If not, see if there is an empty slot we can fill in. */
  if (funcnum < 0)
    for (funcnum = state->num_rw_funcs - 1; funcnum >= 0; funcnum--)
      if (state->rw_func_list[funcnum] == NULL)
	break;

 /*
  * If that also fails, grow the list and add this function to
  * it. 
  */
  if (funcnum < 0)
  {
    funcnum = state->num_rw_funcs++;
    NX_REALLOC(state->rw_func_list, typeof(state->rw_func_list[0]),
	       state->num_rw_funcs);
    state->rw_func_list[funcnum] = func;
  }

 /*
  * Now go through and install the handler everywhere that's
  * appropriate. 
  */
  m = &state->sysmem[low_address];
  for (i = high_address - low_address; i >= 0; m++, i--)
  {
    if (do_read)
    {
      m->mem.read_special = YES;
      m->mem.read_func = funcnum;
    }
    if (do_write)
    {
      m->mem.write_special = YES;
      m->mem.write_func = funcnum;
    }
  }
}



/* Be careful deleting address handlers.  Do not remove "write" traps that
 * are used to detect self modifying code, or you risk making the emulator
 * screw up in strange and unusual ways.  This routine does not check for
 * this on the assumption that you know what you are doing.
 */

void
remove_address_handlers(MachineState * state, int low_address,
			int high_address, AccessType type)
{
  MemoryElement      *m;
  BOOL                do_read = (type != WRITE);	/* Either READ or BOTH */
  BOOL                do_write = (type != READ);	/* Either WRITE or BOTH */
  int                 i;

 /*
  * Loop through all affected memory elements and remove the
  * handlers. 
  */
  m = &state->sysmem[low_address];
  for (i = high_address - low_address; i >= 0; m++, i--)
  {
    if (do_read)
      m->mem.read_special = NO;
    if (do_write)
      m->mem.write_special = NO;
  }
}



/* This installs a routine handler that will trap to a given method in the
 * peripheral card in a specified slot whenever any instruction in a range
 * of memory locations is executed.
 */

void
add_slot_routine_handler(int slot_number, MachineState * state,
			 int low_address, int high_address, SEL sel)
{
  MemoryElement      *m;
  int                 i, selnum, n;

 /* First see if this selector is already in our list. */
  for (selnum = state->num_card_selectors - 1; selnum >= 0; selnum--)
    if (state->card_selector_list[selnum] == sel)
      break;

 /* If not, see if there is an empty slot where it will fit. */
  for (selnum = state->num_card_selectors - 1; selnum >= 0; selnum--)
    if (state->card_selector_list[selnum] == NULL)
    {
      state->card_selector_list[selnum] = sel;
      break;
    }

 /*
  * If there are no empty slots, grow the list & add this
  * selector. 
  */
  if (selnum < 0)
  {
    selnum = state->num_card_selectors++;
    NX_REALLOC(state->card_selector_list, SEL, state->num_card_selectors);
    state->card_selector_list[selnum] = sel;
  }

 /* Now install this in the range of addresses specified. */
  m = &state->sysmem[low_address];
  n = (selnum << 3) | slot_number;
  for (i = high_address - low_address; i >= 0; i--)
  {
    if (m->inst.opcode != OPC_NOT_COMPILED
	&& m->inst.opcode != OPC_TRAPTHIS
	&& m->inst.opcode != OPC_CARDTRAP)
      decompile_instruction(state, m);
    m->inst.opcode = OPC_CARDTRAP;
    m->inst.u.s = n;
  }
}


/* This is the "helper" function for address handlers for peripheral cards.
 * It uses the function number corresponding to the trap function just called
 * (which points here) to determine which slot the handling peripheral card
 * is in and which selector to use when invoking it.
 */

static int
slot_address_handler(AccessType type, MemoryElement * m, MachineState * state)
{
  int                 n = (type == READ) ? m->mem.read_func : m->mem.write_func;

  return (int)[state->peripheral_card[n & 7] perform
	       :state->card_selector_list[n >> 3] with:(id)type with:(id)m];
}


/* To add a slot address handler, we use the same basic mechanism we do for
 * "normal" address handlers.  This time, however, the specific function
 * number conveys information about which selector to use and which slot
 * to use; this information is deciphered by the "slot_address_handler"
 * function.  The low 3 bits of the function number specify the slot (0 is not
 * used), and the rest of the bits specify a selector number from the
 * card_selector_list in the MachineState.  The trick is to find an unused
 * function number with both of these properties; to do so we may leave
 * "holes" in the function/selector lists.  These are left NULL so they may
 * be reused later.
 */

void
add_slot_address_handler(int slot_number, MachineState * state,
			 int low_address, int high_address, SEL sel,
			 AccessType type)
{
  MemoryElement      *mem;
  BOOL                do_read = (type != WRITE);	/* Either READ or BOTH */
  BOOL                do_write = (type != READ);	/* Either WRITE or BOTH */
  int                 i, funcnum;
  int                 maxnum = MIN(state->num_rw_funcs, state->num_card_selectors << 3);


 /* Find out if an appropriate function is already in the table. */
  for (funcnum = slot_number; funcnum < maxnum; funcnum += 8)
  {
    if ((state->rw_func_list[funcnum] == NULL
	 || state->rw_func_list[funcnum] == slot_address_handler)
	&& (state->card_selector_list[funcnum >> 3] == sel
	    || state->card_selector_list[funcnum >> 3] == NULL))
      break;
  }

 /* If not, grow the list and add this function/selector to it. */
  if (funcnum >= maxnum)
  {
  /* Find the lowest selector number we can possibly use. */
    int                 selnum = (state->num_rw_funcs - slot_number + 7) >> 3;

  /*
   * Make sure this corresponds to an available selector number;
   * if not, try the next selector number. 
   */
    while (selnum < state->num_card_selectors
	   && state->card_selector_list[selnum] != NULL
	   && state->card_selector_list[selnum] != sel)
    {
      selnum++;
    }

  /* Calculate the function number. */
    funcnum = (selnum << 3) | slot_number;

  /* Do we have to grow the selector list?  If so, do it. */
    if (selnum >= state->num_card_selectors)
    {
      NX_REALLOC(state->card_selector_list, SEL, selnum + 1);
      for (i = state->num_card_selectors; i < selnum; i++)
	state->card_selector_list[i] = NULL;
      state->num_card_selectors = selnum + 1;
    }

  /* Do we have to grow the function list?  If so, do it. */
    if (funcnum >= state->num_rw_funcs)
    {
      state->rw_func_list = realloc(state->rw_func_list,
				    ((funcnum + 1)
				     * sizeof state->rw_func_list[0]));
      for (i = state->num_rw_funcs; i < funcnum; i++)
	state->rw_func_list[i] = NULL;
      state->num_rw_funcs = funcnum + 1;
    }
  }

 /* Add the function & the selector to the relevant lists. */
  state->rw_func_list[funcnum] = slot_address_handler;
  state->card_selector_list[funcnum >> 3] = sel;

 /*
  * Now go through and install the handler everywhere that's
  * appropriate. 
  */
  mem = &state->sysmem[low_address];
  for (i = high_address - low_address; i >= 0; mem++, i--)
  {
    if (do_read)
    {
      mem->mem.read_special = YES;
      mem->mem.read_func = funcnum;
    }
    if (do_write)
    {
      mem->mem.write_special = YES;
      mem->mem.write_func = funcnum;
    }
  }
}


void
setup_default_handlers(MachineState * m)
{
  int                 i;

  for (i = 0; i < sizeof rw_handlers / sizeof rw_handlers[0]; i++)
  {
    const RWHandlerTableEntry *rw = &rw_handlers[i];

    add_address_handler(m, rw->low_address, rw->high_address,
			rw->func, rw->type_trapped);
  }

  for (i = 0; i < sizeof routine_handlers / sizeof routine_handlers[0]; i++)
  {
    const RoutineHandlerTableEntry *r = &routine_handlers[i];

    add_routine_handler(m, r->low_address, r->high_address, r->func);
  }
}
