/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * NOTE: Much of the code in this directory is big-endian specific for
 * performance reasons.  #ifdef __BIG_ENDIAN__'s and versions of the code
 * for little-endian machines should be added; little endian support is a
 * glaring omission.  This code has also been carefully optimized for the
 * 680x0, although many of the optimizations and the general algorithms
 * should be efficient on any machine.
 */



#import <c.h>
#import <string.h>
#import <stdio.h>
#import "hgrconv.h"
#import "execute.h"
#import <appkit/graphics.h>
#import "hgrtable.h"

typedef struct
{
  int                 left_byte, num_bytes;
}                   ChangedRowInfo;
typedef struct
{
  int                 left_byte, num_bytes;
  int                 first_row, num_rows;
}                   ChangedRectInfo;



/* This function locates areas of the screen that have changed, copies the
 * new data into the "shadow" area, and returns a list of strips (one-tall
 * rectangles, really) that have changed, with no more than one strip per row.
 * The strips returned will be in order of increasing y values.  Returns the
 * number of the rows that have changed.
 */

static int
hgr_identify_and_copy_changed_regions(const MemoryElement * screen,
				      unsigned char *shadow,
				      ChangedRectInfo changed[192],
				      BOOL mixed)
{
  ChangedRectInfo    *r;
  const unsigned int *base;
  const MemoryElement *m;
  unsigned char      *c, *basec;
  int                 last_row = mixed ? 159 : 191;
  short               i, j;

  base = &row_to_addr[0];
  for (i = last_row, r = &changed[0]; i >= 0; base++, i--)
  {
    m = &screen[*base];
    c = basec = &shadow[*base];

  /* See if any bytes on this row are different (unwrapped loop) */
    for (j = 9; j >= 0; m += 4, j--)
    {
      if (m[0].mem.byte != *c++ || m[1].mem.byte != *c++
	  || m[2].mem.byte != *c++ || m[3].mem.byte != *c++)
	break;
    }

  /*
   * If we hit a different byte, j should still be >= 0.  If this
   * has happened, find the rightmost byte that has changed, then
   * copy the area from the leftmost to the rightmost changed
   * byte into the shadow area. 
   */
    if (j >= 0)
    {
      unsigned char      *leftc = c - 1;
      int                 index = *base + 40;

    /* Initialize everything. */
      r->left_byte = leftc - basec;
      r->first_row = last_row - i;
      r->num_rows = 1;
      m = &screen[index];
      c = &shadow[index];

    /*
     * Loop in from the right until we hit a changed byte.  Since
     * there is a leftmost changed byte, there must be a
     * rightmost changed byte, so the only termination condition
     * of this loop is finding a changed byte. 
     */
      while (1)
      {
	m -= 4;
	if (m[3].mem.byte != *--c || m[2].mem.byte != *--c
	    || m[1].mem.byte != *--c || m[0].mem.byte != *--c)
	{
	/* Record the number of bytes on this row. */
	  r->num_bytes = c - leftc + 1;

	/* Copy all changed bytes from screen to shadow. */
	  c = leftc;
	  m = &screen[*base + r->left_byte];
	  for (j = r->num_bytes - 1; j >= 0; j--)
	    *c++ = (m++)->mem.byte;

	  break;
	}
      }

    /* Move on to next strip. */
      r++;
    }
  }

 /* Return the number of strips that have changed. */
  return r - &changed[0];
}


/* This is like its hgr counterpart (above) but works with double hires
 * screens.  Note that the "ChangedRectInfo" fields returned by this function
 * refer to byte indices for each bank, not for the total screen.  So, for
 * instance, if something changed on byte 6 of the aux bank and byte 19 of
 * the main bank (and nowhere else), the left_byte field would be 6 and the
 * num_bytes would be 14 (== 19 - 6 + 1).
 */

static int
dblhgr_identify_and_copy_changed_regions(const MemoryElement * screen1,
					 const MemoryElement * screen2,
					 unsigned char *shadow1,
					 unsigned char *shadow2,
					 ChangedRectInfo changed[192],
					 BOOL mixed)
{
  ChangedRectInfo    *r;
  const unsigned int *base;
  const MemoryElement *m1, *m2;
  unsigned char      *c1, *c2, *basec;
  int                 last_row = mixed ? 159 : 191;
  short               i, j;

  base = &row_to_addr[0];
  for (i = last_row, r = &changed[0]; i >= 0; base++, i--)
  {
    m1 = &screen1[*base];
    m2 = &screen2[*base];
    c1 = basec = &shadow1[*base];
    c2 = &shadow2[*base];

  /* See if any bytes on this row are different (unwrapped loop) */
    for (j = 19; j >= 0; m1 += 2, m2 += 2, j--)
    {
      if (m1[0].mem.byte != *c1++ || m2[0].mem.byte != *c2++
	  || m1[1].mem.byte != *c1++ || m2[1].mem.byte != *c2++)
	break;
    }

  /*
   * If we hit a different byte, j should still be >= 0.  If this
   * has happened, find the rightmost byte that has changed, then
   * copy the area from the leftmost to the rightmost changed
   * byte into the shadow area. 
   */
    if (j >= 0)
    {
      unsigned char      *leftc = c1 - 1;
      int                 index = *base + 40;

    /* Initialize everything. */
      r->left_byte = leftc - basec;
      r->first_row = last_row - i;
      r->num_rows = 1;
      m1 = &screen1[index];
      m2 = &screen2[index];
      c1 = &shadow1[index];
      c2 = &shadow2[index];

    /*
     * Loop in from the right until we hit a changed byte.  Since
     * there is a leftmost changed byte, there must be a
     * rightmost changed byte, so the only termination condition
     * of this loop is finding any changed byte. 
     */
      while (1)
      {
	m1 -= 2;
	m2 -= 2;
	if (m1[1].mem.byte != *--c1 || m2[1].mem.byte != *--c2
	    || m1[0].mem.byte != *--c1 || m2[0].mem.byte != *--c2)
	{
	/* Record the number of bytes on this row. */
	  r->num_bytes = c1 - leftc + 1;

	  index = *base + r->left_byte;

	/* Copy all changed bytes from screen to shadow. */
	  c1 = leftc;
	  c2 = &shadow2[index];
	  m1 = &screen1[index];
	  m2 = &screen2[index];
	  for (j = r->num_bytes - 1; j >= 0; j--)
	  {
	    *c1++ = (m1++)->mem.byte;
	    *c2++ = (m2++)->mem.byte;
	  }

	  break;
	}
      }

    /* Move on to next strip. */
      r++;
    }
  }

 /* Return the number of strips that have changed. */
  return r - &changed[0];
}


/* This routine takes the raw changed row information and smooths it out
 * to decrease the number of distinct rectangles to convert and display.
 * For example, two adjacent "thin" rectangles with similar starting and
 * ending points will be "coalesced" into one larger rectangle.
 * The "rectangles" passed to this routine are expected to be sorted in
 * order of increasing y, and to only have a height of 1.
 */

static int
coalesce_changed_rects(int num_rows, ChangedRectInfo * changed)
{
  ChangedRectInfo    *current, *test;
  int                 lasty;
  short               j;

 /* If there are are no changed rows, bail out immediately. */
  if (num_rows == 0)
    return 0;

 /*
  * Set up pointers to the rectangle being worked on (current)
  * and the rectangle (test) that will either be absorbed into it
  * or splintered off into a new rectangle. 
  */
  current = changed;
  test = changed + 1;
  lasty = current->first_row;

 /* Loop through all of the rows merging what seems reasonable. */
  for (j = num_rows - 2; j >= 0; test++, j--)
  {
  /* Can we merge test in with current?  Quick check first. */
    if (test->first_row - lasty <= 4)	/* Near enough vertically? */
    {
    /* Passed that, now perform more complex l/r boundary check. */
      int                 leftdiff = test->left_byte - current->left_byte;
      int                 rightdiff = leftdiff + test->num_bytes - current->num_bytes;

    /* See if the left and right edges are "close enough". */
      if ((leftdiff >= -3 && leftdiff <= 3)
	  && (rightdiff >= -3 && rightdiff <= 3))
      {
	if (leftdiff < 0)
	{
	  current->left_byte = test->left_byte;
	  current->num_bytes += -leftdiff;	/* Adjust right edge. */
	}
	if (rightdiff > 0)
	  current->num_bytes += rightdiff;
	lasty = test->first_row;
	continue;
      }
    }

  /* If we can't merge it in, start a new rectangle. */
    current->num_rows = lasty - current->first_row + 1;
    current++;
    *current = *test;
    lasty = current->first_row;
  }

 /* Clean up. */
  current->num_rows = lasty - current->first_row + 1;

  return (current - changed + 1);
}



/* FIXME!
 * This routine is old but I haven't replaced it yet.  This converts a range
 * of scanlines from Apple ][ format to NeXT grayscale (scaled by 2 in x and
 * y).  Unfortunately, for historical reasons this converts the entire
 * scanline even though it knows which bytes on each row should be converted.
 *
 * Note to anyone attempting to make the obvious fix and only convert the
 * section of the scanline that has changed: remember that colors can "bleed
 * out" from the changed bytes into adjacent screen areas corresponding to
 * unchanged bytes. 
 */


static void
hgr_scanlines_to_NeXT_grayscale(const unsigned char *a2buf,
				unsigned char nextbuf[384][140],
				int first_row, int height)
{
  const unsigned int *row_base = row_to_addr + first_row;
  unsigned long      *nbuf1 = (unsigned long *)&nextbuf[first_row * 2][0];
  unsigned long      *nbuf2 = &nbuf1[140 / 4];	/* Start of next row. */
  const unsigned char *p;
  short               rows_left, n;


 /* NOTE: This assumes a big-endian CPU! */


  for (rows_left = height - 1; rows_left >= 0; rows_left--)
  {
    Reg                 a2bits;
    unsigned long       a, b;

#define NEXT_A2BITS a2bits.s.s &= 0x7F; a2bits.s.s <<= 7; a2bits.c.c = *p++

    p = a2buf + *row_base++;

  /* Left edge - 5 bytes */
  /* Create the first long and ship it out. */
    a2bits.l = *p++;
    a = hgr_one_to_two_bpp_even[a2bits.l] << 16;
    NEXT_A2BITS;
    b = hgr_one_to_two_bpp_odd[a2bits.l];
    *nbuf1++ = *nbuf2++ = a | (b >> 12);

    NEXT_A2BITS;
    a = hgr_one_to_two_bpp_even[a2bits.l];
    *nbuf1++ = *nbuf2++ = (b << 20) | (a >> 8);

    NEXT_A2BITS;
    b = hgr_one_to_two_bpp_odd[a2bits.l];
    *nbuf1++ = *nbuf2++ = (a << 24) | (b >> 4);

    NEXT_A2BITS;
    a = hgr_one_to_two_bpp_even[a2bits.l];
    *nbuf1++ = *nbuf2++ = (b << 28) | a;


  /* Middle - 32 bytes */
    for (n = 3; n >= 0; n--)
    {
      NEXT_A2BITS;
      a = hgr_one_to_two_bpp_odd[a2bits.l] << 4;
      NEXT_A2BITS;
      b = hgr_one_to_two_bpp_even[a2bits.l];
      *nbuf1++ = *nbuf2++ = a | (b >> 24);

      NEXT_A2BITS;
      a = hgr_one_to_two_bpp_odd[a2bits.l];
      *nbuf1++ = *nbuf2++ = (b << 8) | (a >> 20);

      NEXT_A2BITS;
      b = hgr_one_to_two_bpp_even[a2bits.l];
      *nbuf1++ = *nbuf2++ = (a << 12) | (b >> 16);

      NEXT_A2BITS;
      a = hgr_one_to_two_bpp_odd[a2bits.l];
      *nbuf1++ = *nbuf2++ = (b << 16) | (a >> 12);

      NEXT_A2BITS;
      b = hgr_one_to_two_bpp_even[a2bits.l];
      *nbuf1++ = *nbuf2++ = (a << 20) | (b >> 8);

      NEXT_A2BITS;
      a = hgr_one_to_two_bpp_odd[a2bits.l];
      *nbuf1++ = *nbuf2++ = (b << 24) | (a >> 4);

      NEXT_A2BITS;
      b = hgr_one_to_two_bpp_even[a2bits.l];
      *nbuf1++ = *nbuf2++ = (a << 28) | b;
    }

  /* Right edge - 3 bytes. */
    NEXT_A2BITS;
    a = hgr_one_to_two_bpp_odd[a2bits.l] << 4;
    NEXT_A2BITS;
    b = hgr_one_to_two_bpp_even[a2bits.l];
    *nbuf1++ = *nbuf2++ = a | (b >> 24);
    NEXT_A2BITS;
    a = hgr_one_to_two_bpp_odd[a2bits.l];
    *nbuf1++ = *nbuf2++ = (b << 8) | (a >> 20);
    a2bits.s.s &= 0x7F;
    a2bits.s.s <<= 7;
    b = hgr_one_to_two_bpp_even[a2bits.l];
    *nbuf1++ = *nbuf2++ = (a << 12) | (b >> 16);

  /* Move pointers on to the next rows they should fill in. */
    nbuf1 = nbuf2;
    nbuf2 += (140 / 4);
  }
}


/* This routine converts various areas of the hgr screen from Apple ][ pixels
 * to monochrome 2 bits per pixel NeXT pixels, scaling by 2 in x and y.
 * It may convert somewhat more of the screen that it has to, but this
 * should always be OK.
 */

static void
hgr_to_NeXT_monochrome(int num_rects, const ChangedRectInfo * changed,
		       const unsigned char *hgrbuf,
		       unsigned char nextbuf[384][140])
{
  const ChangedRectInfo *r;

  for (r = changed; num_rects--; r++)
  {
    unsigned long      *nbuf1, *nbuf2;
    int                 left_byte = r->left_byte & ~7;
    const unsigned int *row_base = &row_to_addr[r->first_row];
    int                 segs_minus_one;
    unsigned            nbuf_add;
    short               i, j, k;

  /* Initialize pointers */
    nbuf1 = (unsigned long *)&nextbuf[r->first_row * 2][left_byte / 8 * 28];
    nbuf2 = nbuf1 + (140 / sizeof(unsigned long));

  /* Calculate how many 8 byte segments we are going to convert. */
    segs_minus_one = (r->num_bytes + r->left_byte - left_byte - 1) / 8;
    nbuf_add = (4 - segs_minus_one) * 7 + (140 / sizeof(unsigned long));

    for (k = r->num_rows - 1; k >= 0; k--)
    {
      const unsigned char *p = &hgrbuf[*row_base++ + left_byte];

      for (i = segs_minus_one; i >= 0; i--)
      {
	unsigned long       new, old;
	int                 lshift = 0, rshift = 28;

      /*
       * Here is the conversion loop, which converts 8 Apple ][
       * bytes into 28 NeXT bytes.  I used to have this loop
       * unwrapped with appropriate shift constants inserted, but
       * I doubt there's much point...loop overhead is minimal. 
       *
       * gcc is doing suboptimal things with the following loop...
       * look into it. 
       */
	old = hgr_monochrome_one_to_two_bpp[*p++];
	for (j = 6; j >= 0; j--)
	{
	  new = hgr_monochrome_one_to_two_bpp[*p++];
	  *nbuf1++ = *nbuf2++ = (old << lshift) | (new >> rshift);
	  lshift += 4;
	  rshift -= 4;
	  old = new;
	}
      }

      nbuf1 += nbuf_add;
      nbuf2 += nbuf_add;
    }
  }
}



/* This is the key routine for converting from an Apple //e screen to NeXT
 * 16 bit color.  I've added inline assembly in the key parts of the
 * inner loops where gcc wasn't doing a good job.  I wouldn't *NORMALLY*
 * do this, but I imagine this eats lots of CPU time.
 */

static void
hgr_to_NeXT_color(int num_rects, const ChangedRectInfo * changed,
		  unsigned char *hgrbuf,
		  unsigned short nextbuf[384][560])
{
  const ChangedRectInfo *r;

  for (r = changed; num_rects--; r++)
  {
    unsigned long      *nbuf1, *nbuf2;
    int                 left_byte = r->left_byte & ~1;	/* Always start on even byte. */
    const unsigned int *row_base = &row_to_addr[r->first_row];
    unsigned            row;
    unsigned            nbuf_add;
    int                 pairs_minus_one;
    short               k;

  /* Figure out leftmost byte.  Account for color bleeding. */
    left_byte = r->left_byte;
    if (left_byte > 0)
      left_byte--;
    left_byte &= ~1;		/* Assure that we always start on an even byte. */

  /*
   * Figure out how many pairs of bytes to do.  Make sure we
   * don't go past the right edge.  Note that pairs_minus_one
   * implicitly adds one byte to make sure it accounts for "color
   * bleeds" to the right. 
   */
    pairs_minus_one = (r->num_bytes + r->left_byte - left_byte) >> 1;
    if (left_byte + 2 * pairs_minus_one > 38)
      pairs_minus_one--;

  /* Initialize pointers and calculate row increment. */
#define LONGS_PER_ROW ((560 * sizeof (unsigned short)) \
		       / sizeof (unsigned long))
    nbuf1 = (unsigned long *)&nextbuf[r->first_row * 2][left_byte * 14];
    nbuf2 = nbuf1 + LONGS_PER_ROW;
    nbuf_add = (2 * LONGS_PER_ROW
		- ((pairs_minus_one + 1) * 14 * 2 * sizeof(unsigned short)
		   / sizeof(unsigned long)));
#undef LONGS_PER_ROW

    for (k = r->num_rows - 1; k >= 0; k--)
    {
      unsigned char      *base = &hgrbuf[*row_base++];
      const unsigned char *p;
      unsigned            odd, even;
      unsigned char       old_end;
      short               j;

    /* Save end byte. */
      p = base + left_byte;
      old_end = base[40];
      base[40] = 0;

    /* Set up starting conditions. */
      odd = (left_byte == 0) ? 0 : p[-1];
      even = *p++;

      for (j = pairs_minus_one; j >= 0; j--)
      {
	const unsigned long *data;
	const unsigned long (*even_lookup)[4], (*odd_lookup)[4];

      /* Slam in first four pixels. */
	even_lookup = hgr_to_color[even >> 7];
	even &= ~0x80;

	data = &even_lookup[(even << 3) | ((odd >> 4) & 7)][0];
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;

      /* Slam in next 3 pixels. */
	odd = *p++;
	odd_lookup = hgr_to_color[odd >> 7];
	odd &= ~0x80;

	data = &even_lookup[((even >> 1) & 0x3F) | ((odd & 7) << 6)][0];
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;

      /* Slam in next 3 pixels. */
	data = &odd_lookup[(even >> 3) | ((odd & 0x3F) << 4)][1];
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;

      /*
       * This was temporarily set to be 0 off the right edge of
       * the row, so the boundary case works correctly. 
       */
	even = *p++;

      /* Slam in final 4 pixels. */
	data = &odd_lookup[odd | ((even & 7) << 7)][0];
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
	*nbuf1++ = *data, *nbuf2++ = *data++;
      }

      nbuf1 += nbuf_add;
      nbuf2 += nbuf_add;
      base[40] = old_end;
    }
  }
}


static void
hgr_to_NeXT_16_bit_monochrome(int num_rects, const ChangedRectInfo * changed,
			      const unsigned char *hgrbuf,
			      unsigned short nextbuf[384][560])
{
  const ChangedRectInfo *r;

  for (r = changed; num_rects--; r++)
  {
    unsigned long      *nbuf1, *nbuf2;
    const unsigned int *row_base = &row_to_addr[r->first_row];
    const unsigned char *bufbase = hgrbuf + r->left_byte;
    unsigned            row;
    unsigned            nbuf_add;
    short               k;

  /* Initialize pointers and calculate row increment. */
#define LONGS_PER_ROW ((560 * sizeof (unsigned short)) \
		       / sizeof (unsigned long))
    nbuf1 = ((unsigned long *)
	     &nextbuf[r->first_row * 2][(r->left_byte + r->num_bytes) * 14]);
    nbuf2 = nbuf1 + LONGS_PER_ROW;
    nbuf_add = 2 * LONGS_PER_ROW + r->num_bytes * 7;
#undef LONGS_PER_ROW

    for (k = r->num_rows - 1; k >= 0; k--)
    {
      const unsigned char *p = &bufbase[*row_base++];
      short               j;

    /* Toss out 7 double-width pixels per byte. */
      for (j = r->num_bytes - 1; j >= 0; j--)
      {
	const unsigned long *data;
	unsigned int        n;

#ifdef mc68000
      /*
       * This efficiently expresses extraction of 7 bits into a
       * long 
       */
	asm("bfextu %1@(%2:w){#1:#7},%0":"=d"(n) :"a"(p), "d"(j));
#else
	n = p[j] & 0x7F;
#endif

	data = &hgr_monochrome_one_to_16_bpp[n][7];


      /*
       * Here we have to move 28 bytes from a table into our
       * frame buffer.  I special cased it for the 680x0 so it
       * uses moveml (we can't use move16 for a variety of
       * reasons), but benchmarks showed that I wasn't getting
       * any significant performance increase with moveml over a
       * bunch of movel's (at least not on the 68040).  I've
       * deleted the moveml code. 
       */

      /*
       * Yes, this looks weird but it produces faster code than
       * *--nbuf1 = *--nbuf2 = *--data. 
       */
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
	*--nbuf1 = *--data, *--nbuf2 = *data;
      }

      nbuf1 += nbuf_add;
      nbuf2 += nbuf_add;
    }
  }
}

/* This routine converts various areas of the double hires screen from Apple ][
 * pixels to grayscale 2 bits per pixel NeXT pixels.  It uses 8x8 dithering
 * patterns to try to make 16 colors look good on a 2 bpp screen.
 * It may convert somewhat more of the screen that it has to, but this
 * should always be OK.
 */

static void
dblhgr_to_NeXT_grayscale(int num_rects, const ChangedRectInfo * changed,
			 const unsigned char *main_hgrbuf,
			 const unsigned char *aux_hgrbuf,
			 unsigned char nextbuf[384][140])
{
  const ChangedRectInfo *r;

  for (r = changed; num_rects--; r++)
  {
    unsigned long      *nbuf1, *nbuf2;
    int                 left_byte = r->left_byte & ~7;
    const unsigned int *row_base = &row_to_addr[r->first_row];
    unsigned            row;
    int                 segs_minus_one;
    unsigned            nbuf_add;
    short               k;

  /* Initialize pointers */
    nbuf1 = (unsigned long *)&nextbuf[r->first_row * 2][left_byte / 8 * 28];
    nbuf2 = nbuf1 + (140 / sizeof(unsigned long));

  /* Calculate how many 8 byte segments we are going to convert. */
    segs_minus_one = (r->num_bytes + r->left_byte - left_byte - 1) / 8;
    nbuf_add = (4 - segs_minus_one) * 7 + (140 / sizeof(unsigned long));

    row = r->first_row;
    for (k = r->num_rows - 1; k >= 0; k--)
    {
      unsigned long       a, b, n;
      register const unsigned long *lookup;
      const unsigned long *p1, *p2;
      short               j;

      lookup = &dblhgr_grayscale[row++ % 8][0];

    /* Fetch ptrs to first elements of the rows. */
      p1 = (const unsigned long *)&aux_hgrbuf[*row_base + left_byte];
      p2 = (const unsigned long *)&main_hgrbuf[*row_base + left_byte];
      row_base++;

      for (j = segs_minus_one; j >= 0; j--)
      {

      /*
       * Here we actually convert the data.  I used to have an
       * interesting union of bitfield structs hack here, but it
       * was far too bizarre. This is really really annoying! 
       * Basically we have an array of 4 bit pixels...but odd
       * numbered bytes are in one array, even numbered bytes are
       * in another.  Worse yet, bit 7 of every byte must be
       * skipped!  Nightmare! gcc seems to do decently with the
       * following code.  It may look to you like (((a >> 24) &
       * 0x7F) << 1) is suboptimal code, but it turns out that
       * gcc uses bfextu to efficiently extract the high bits if
       * it sees expressions involving a shift and an and with
       * something of the form 2^n - 1. 
       */

	a = *p1++;
	b = *p2++;

      /* First 16 NeXT pixels. */
	n = ((lookup[(((a >> 24) & 0x7F) << 1) | ((b >> 30) & 1)] << 16)
	     | lookup[(((b >> 24) & 0x3F) << 2) | ((a >> 21) & 0x3)]);
	*nbuf1++ = *nbuf2++ = n;

      /* Another 16 pixels */
	n = ((lookup[(((a >> 16) & 0x1F) << 3) | ((b >> 20) & 7)] << 16)
	     | lookup[(((b >> 16) & 0xF) << 4) | ((a >> 11) & 0xF)]);
	*nbuf1++ = *nbuf2++ = n;

      /* Another 16 pixels */
	n = ((lookup[(((a >> 8) & 0x7) << 5) | ((b >> 10) & 0x1F)] << 16)
	     | lookup[(((b >> 8) & 0x3) << 6) | ((a >> 1) & 0x3F)]);
	*nbuf1++ = *nbuf2++ = n;

      /* Get the trailing info and save it. */
	n = lookup[((a & 1) << 7) | (b & 0x7F)] << 16;

      /* Move on to next set of 56 pixels. */
	a = *p1++;
	b = *p2++;

      /* Another 16 pixels. */
	n |= lookup[(((a >> 24) & 0x7F) << 1) | ((b >> 30) & 0x1)];
	*nbuf1++ = *nbuf2++ = n;

      /* Another 16 pixels. */
	n = ((lookup[(((b >> 24) & 0x3F) << 2) | ((a >> 21) & 3)] << 16)
	     | lookup[(((a >> 16) & 0x1F) << 3) | ((b >> 20) & 0x7)]);
	*nbuf1++ = *nbuf2++ = n;

      /* Another 16 pixels. */
	n = ((lookup[(((b >> 16) & 0xF) << 4) | ((a >> 11) & 0xF)] << 16)
	     | lookup[(((a >> 8) & 0x7) << 5) | ((b >> 10) & 0x1F)]);
	*nbuf1++ = *nbuf2++ = n;

      /* Another 16 pixels. */
	n = ((lookup[(((b >> 8) & 0x3) << 6) | ((a >> 1) & 0x3F)] << 16)
	     | lookup[((a & 1) << 7) | (b & 0x7F)]);
	*nbuf1++ = *nbuf2++ = n;
      }
      nbuf1 += nbuf_add;
      nbuf2 += nbuf_add;
    }
  }
}



/* This routine converts various areas of the double hires screen from Apple ][
 * pixels to monochrome 2 bits per pixel NeXT pixels.
 * It may convert somewhat more of the screen that it has to, but this
 * should always be OK.
 */

static void
dblhgr_to_NeXT_monochrome(int num_rects, const ChangedRectInfo * changed,
			  const unsigned char *main_hgrbuf,
			  const unsigned char *aux_hgrbuf,
			  unsigned char nextbuf[384][140])
{
  const ChangedRectInfo *r;

  for (r = changed; num_rects--; r++)
  {
    unsigned long      *nbuf1, *nbuf2;
    const unsigned char *p1, *p2;
    int                 left_byte = r->left_byte & ~7;
    const unsigned int *row_base = &row_to_addr[r->first_row];
    int                 segs_minus_one;
    unsigned            nbuf_add;
    short               i, j, k;

  /* Initialize pointers */
    nbuf1 = (unsigned long *)&nextbuf[r->first_row * 2][left_byte / 8 * 28];
    nbuf2 = nbuf1 + (140 / sizeof(unsigned long));

  /* Calculate how many 8 byte segments we are going to convert. */
    segs_minus_one = (r->num_bytes + r->left_byte - left_byte - 1) / 8;
    nbuf_add = (4 - segs_minus_one) * 7 + (140 / sizeof(unsigned long));

    for (k = r->num_rows - 1; k >= 0; k--)
    {
      p1 = &aux_hgrbuf[*row_base + left_byte];
      p2 = &main_hgrbuf[*row_base + left_byte];
      row_base++;

      for (i = segs_minus_one; i >= 0; i--)
      {
	unsigned long       new, old;
	int                 lshift = 2, rshift = 26;

      /*
       * Here is the conversion code, which converts 16 Apple ][
       * bytes into 28 NeXT bytes. 
       */
	old = ((dblhgr_monochrome_one_to_two_bpp[*p1++] << 14)
	       | dblhgr_monochrome_one_to_two_bpp[*p2++]);
	for (j = 6; j >= 0; j--)
	{
	  new = ((dblhgr_monochrome_one_to_two_bpp[*p1++] << 14)
		 | dblhgr_monochrome_one_to_two_bpp[*p2++]);
	  *nbuf1++ = *nbuf2++ = (old << lshift) | (new >> rshift);
	  lshift += 4;
	  rshift -= 4;
	  old = new;
	}
      }

      nbuf1 += nbuf_add;
      nbuf2 += nbuf_add;
    }
  }
}



/* This routine takes a list of subrectangles of the Apple ][ hires page.
 * It converts them to NeXT grayscale format (scaled by 2 in x and y) and
 * places the updated information into a raw NeXT bitmap.  This routine may
 * recalculate somewhat more of the screen than is specified, but that
 * should always be OK since updating too much can only be redundant,
 * never wrong.
 */

static void
hgr_to_NeXT_grayscale(int num_rects, ChangedRectInfo * changed,
		      const unsigned char *hgrbuf,
		      unsigned char nextbuf[384][140])
{
  for (; num_rects--; changed++)
    hgr_scanlines_to_NeXT_grayscale(hgrbuf, nextbuf, changed->first_row,
				    changed->num_rows);
}


/* This is the general purpose internal hgr/dblhgr update function called by
 * all of the hires update variants.
 */
static int
generic_hgr_update(const MemoryElement * screen,
		   const MemoryElement * screen2,	/* only used for dblhgr */
		   unsigned char *shadow,
		   unsigned char *shadow2,	/* only used for dblhgr */
		   void *fbuf,
		   NXRect changed_rect[MAX_UPDATE_RECTS],
		   BOOL update_all, BOOL mixed,
		   int (*identify_func) (),
		   void (*convert) (),
		   int alignment_minus_one,
		   BOOL is_double_hires)
{
  ChangedRectInfo     changed[192], *cri;
  NXRect             *r;
  int                 num_rows_changed, num_rects;
  short               j;

 /* Figure out what it is we have to convert. */
  if (!update_all)
  {
    if (is_double_hires)
      num_rows_changed = identify_func(screen, screen2, shadow, shadow2,
				       changed, mixed);
    else
      num_rows_changed = identify_func(screen, shadow, changed, mixed);
    num_rects = coalesce_changed_rects(num_rows_changed, &changed[0]);
  }
  else
  {
  /* Set up a rectangle enclosing the whole screen. */
    changed[0].left_byte = 0;
    changed[0].num_bytes = 40;
    changed[0].first_row = 0;
    changed[0].num_rows = mixed ? 160 : 192;
    num_rects = 1;

  /* Set up the shadow hgr screen correctly. */

    if (is_double_hires)
    {
      for (j = 191; j >= 0; j--)
      {
	int                 base = row_to_addr[j];
	const MemoryElement *m1 = &screen[base];
	const MemoryElement *m2 = &screen2[base];
	unsigned char      *c1 = &shadow[base];
	unsigned char      *c2 = &shadow2[base];
	short               i;

      /*
       * Who really cares about speed here; forced updates of the
       * entire screen should be rare.  And when they happen,
       * they'll take a while anyway. 
       */
	for (i = 39; i >= 0; i--)
	{
	  *c1++ = (m1++)->mem.byte;
	  *c2++ = (m2++)->mem.byte;
	}
      }
    }
    else
    {
      for (j = 191; j >= 0; j--)
      {
	const MemoryElement *m = &screen[row_to_addr[j]];
	unsigned char      *c = &shadow[row_to_addr[j]];
	short               i;

	for (i = 39; i >= 0; i--)
	  *c++ = (m++)->mem.byte;
      }
    }
  }

  if (num_rects > 0)
  {
  /* Convert the appropriate areas of the screen. */
    if (is_double_hires)
      convert(num_rects, changed, shadow2, shadow, fbuf);
    else
      convert(num_rects, changed, shadow, fbuf);

  /*
   * Generate an NXRect for every region that should be
   * displayed. 
   */
    r = &changed_rect[0];
    cri = &changed[0];
    for (j = num_rects - 1; j >= 0; r++, cri++, j--)
    {
    /*
     * Make sure the rectangle falls on even n byte boundaries in
     * the NeXT raw bitmap. 
     */

      int                 x, orig_x = cri->left_byte * 14, width, d;

      x = orig_x;
      if (x > 0)
	x -= 4;			/* Compensate for "bleeding" */
      x &= ~alignment_minus_one;

      r->origin.x = x;
      r->origin.y = 384 - (cri->first_row + cri->num_rows) * 2;

    /* Add to width to assure that we catch bleeds to the right. */
      width = cri->num_bytes * 14 + 4 + (orig_x - x);

    /* Round up to nearest n bytes. */
      width = (width + alignment_minus_one) & ~alignment_minus_one;

    /* See if we go off the right edge. */
      d = x + width - 560;
      if (d > 0)		/* If so, move back so we stop at the right edge. */
	width -= d;
      r->size.width = width;
      r->size.height = cri->num_rows * 2;
    }
  }

  return num_rects;
}


/* This is a routine to be called from the emulator's display routines.
 * It takes a pointer to the base of hires memory, the "shadow"
 * screen (the last hgr screen drawn), an array of rectangles that will be
 * partially filled in with the areas of the screen that should be redrawn,
 * and a flag indicating whether to reconvert the entire screen.
 * It returns the number of rectangles to be blitted to the screen.
 */

int
hgr_update_framebuffer(const MemoryElement * screen, unsigned char *shadow,
		       unsigned char *fbuf,
		       NXRect changed_rect[MAX_UPDATE_RECTS],
		       BOOL update_all, BOOL mixed, BOOL monochrome)
{
  return generic_hgr_update(screen, NULL, shadow, NULL, fbuf, changed_rect,
			    update_all, mixed,
			    (int (*) ())hgr_identify_and_copy_changed_regions,
			    (monochrome ?
			     (void (*) ())hgr_to_NeXT_monochrome
			     : (void (*) ())hgr_to_NeXT_grayscale),
			    15, NO);
}


/* This is a routine to be called from the emulator's display routines.
 * It takes two pointer to the bases of hires memory (aux and main),
 * the "shadow" screens (the last dblhgr screens drawn, one for aux and
 * one for main), an array of rectangles that will be partially filled in
 * with the areas of the screen that should be redrawn, and a flag indicating
 * whether to reconvert the entire screen.  It returns the number of
 * rectangles to be blitted to the screen.
 */

int
dblhgr_update_framebuffer(const MemoryElement * main_screen,
			  const MemoryElement * aux_screen,
			  unsigned char *main_shadow,
			  unsigned char *aux_shadow,
			  unsigned char *fbuf,
			  NXRect changed_rect[MAX_UPDATE_RECTS],
			  BOOL update_all, BOOL mixed, BOOL monochrome)
{
  return generic_hgr_update(aux_screen, main_screen, aux_shadow,
			    main_shadow, fbuf, changed_rect,
			    update_all, mixed,
			    ((int (*) ())
			     dblhgr_identify_and_copy_changed_regions),
			    (monochrome ?
			     (void (*) ())dblhgr_to_NeXT_monochrome
			     : (void (*) ())dblhgr_to_NeXT_grayscale),
			    15, YES);
}


/* This is a routine to be called from the emulator's display routines.
 * It takes a pointer to the base of hires memory, the "shadow"
 * screen (the last hgr screen drawn), an array of rectangles that will be
 * partially filled in with the areas of the screen that should be redrawn,
 * and a flag indicating whether to reconvert the entire screen.
 * It returns the number of rectangles to be blitted to the screen.
 */

int
hgr_update_color_framebuffer(const MemoryElement * screen,
			     unsigned char *shadow,
			     unsigned char *fbuf,
			     NXRect changed_rect[MAX_UPDATE_RECTS],
			     BOOL update_all, BOOL mixed, BOOL monochrome)
{
  return generic_hgr_update(screen, NULL, shadow, NULL, fbuf, changed_rect,
			    update_all, mixed,
			    (int (*) ())hgr_identify_and_copy_changed_regions,
			    (monochrome ?
			     (void (*) ())hgr_to_NeXT_16_bit_monochrome
			     : (void (*) ())hgr_to_NeXT_color),
			    7, NO);
}
