/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#import "memory.h"
#import "execute.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))


/* Writes one byte to the byte occupying a given Apple ][ address.
 * This handles writes to trapped addresses and this (or read_mem_byte)
 * is the preferred way to flip a softswitch.
 */

void
write_mem_byte(MachineState * s, unsigned short addr, int x)
{
  MemoryElement      *m = &s->write_base_p[addr >> 8][addr % 256];

  m->mem.byte = x;
  if (m->mem.write_special)
    s->rw_func_list[m->mem.write_func] (WRITE, m, s);
}


/* Reads one byte from the byte occupying a given Apple ][ address.
 * This handles reads from trapped addresses and this (or write_mem_byte)
 * is the preferred way to flip a softswitch.
 */

unsigned char
read_mem_byte(MachineState * s, unsigned short addr)
{
  MemoryElement      *m = &s->read_base_p[addr >> 8][addr % 256];

  if (m->mem.read_special)
    return s->rw_func_list[m->mem.read_func] (READ, m, s);
  else
    return m->mem.byte;
}


/* Convenience function.  Returns the two byte, LSB-first value stored
 * at addr.
 */
unsigned short
read_mem_short(MachineState * s, unsigned short addr)
{
  unsigned short      ret = read_mem_byte(s, addr);

  ret |= read_mem_byte(s, (addr + 1) & 0xFFFF) << 8;
  return ret;
}


/* Convenience function.  Writes a two-byte value LSB first to a specified
 * address.
 */
void
write_mem_short(MachineState * s, unsigned short addr, int value)
{
  write_mem_byte(s, addr, value & 0xFF);
  write_mem_byte(s, (addr + 1) & 0xFFFF, (value >> 8) & 0xFF);
}


/* Reads from a given MemoryElement, trapping iff necessary. */
unsigned char
read_memory_element(MachineState * s, MemoryElement * m)
{
  if (m->mem.read_special)
    return s->rw_func_list[m->mem.read_func] (READ, m, s);
  else
    return m->mem.byte;
}


/* Writes to a given MemoryElement, trapping iff necessary. */
void
write_memory_element(MachineState * s, MemoryElement * m, int x)
{
  m->mem.byte = x;
  if (m->mem.write_special)
    s->rw_func_list[m->mem.write_func] (WRITE, m, s);
}


/* Similar to memcpy(), but this copies memory from outside the Apple ][
 * to Apple ][ memory occupying a given range of addresses.  Handles all
 * trapped addresses.  This is faster than multiple calls to
 * write_mem_byte().
 */

void
copy_bytes_to_apple(MachineState * s, const unsigned char *src,
		    unsigned short dst_addr, int num_bytes)
{
 /*
  * Here we speed things up greatly by not doing the
  * virtual->physical address indirection every time we get the
  * next MemoryElement, but instead every time we cross a page. 
  */

  while (num_bytes > 0)
  {
    int                 to_do = MIN(num_bytes, 256 - (dst_addr & 0xFF));
    MemoryElement      *m = &s->write_base_p[dst_addr >> 8][dst_addr % 256];

    num_bytes -= to_do;
    dst_addr += to_do;

    for (; to_do--; m++)
    {
      m->mem.byte = *src++;
      if (m->mem.write_special)
	s->rw_func_list[m->mem.write_func] (WRITE, m, s);
    }
  }
}


/* Similar to memcpy(), but this copies memory from the emulator to a buffer
 * you provide.  Handles all trapped addresses.  This is faster than
 * multiple calls to read_mem_byte().
 */

void
copy_bytes_from_apple(MachineState * s, unsigned short src_addr,
		      unsigned char *dst, int num_bytes)
{
 /*
  * Here we speed things up greatly by not doing the
  * virtual->physical address indirection every time we get the
  * next MemoryElement, but instead every time we cross a page. 
  */

  while (num_bytes > 0)
  {
    int                 to_do = MIN(num_bytes, 256 - (src_addr & 0xFF));
    MemoryElement      *m = &s->read_base_p[src_addr >> 8][src_addr % 256];

    num_bytes -= to_do;
    src_addr += to_do;

    for (; to_do--; m++)
    {
      if (m->mem.read_special)
	*dst++ = s->rw_func_list[m->mem.read_func] (READ, m, s);
      else
	*dst++ = m->mem.byte;
    }
  }
}

/* This gives you a pointer to the memory element occupying a given address
 * referenced by either READ or WRITE.
 */
MemoryElement      *
get_memory_element(MachineState * state, unsigned short addr, AccessType type)
{
  MemoryElement     **list = ((type == READ)
			      ? state->read_base_p : state->write_base_p);

  return &list[addr >> 8][addr % 256];
}
