/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


/*
 *    opcodes.h
 */

#ifndef __OPCODES_H__
#define __OPCODES_H__


/* Opcode bits:
 * xxxx xxxx xxxC cnvz
 *
 * x - these bits distinguish the opcode type (eg, between "ADC", "AND", etc.)
 * C - "check"    Child instruction(s) may be in a different bank; check!
 * c - "carry",    \   Each of these bits signifies that a given 6502 CC bit
 * n - "negative",  |  is guaranteed to be "valid" after this instruction has
 * v - "overflow",  |  executed.  This means either this instruction set it,
 * z - "zero",     /   or a previous instruction set it and it's still valid.
 */


/* Note: these instructions have not been correctly numbered yet. */

#define BLOCKSIZE 32
#define NEWBANK 16
#define CBIT    8
#define NBIT    4
#define VBIT    2
#define ZBIT    1
#define NONE    0
#define ALLBITS (CBIT | NBIT | VBIT | ZBIT)

typedef enum {
  OPC_NOT_COMPILED   = 0,

  OPC_INVALID        = OPC_NOT_COMPILED + BLOCKSIZE,

  OPC_ADC_IMM        = OPC_INVALID + BLOCKSIZE,
  OPC_ADC_IMM_C      = OPC_ADC_IMM | CBIT,
  OPC_ADC_IMM_N      = OPC_ADC_IMM | NBIT,
  OPC_ADC_IMM_V      = OPC_ADC_IMM | VBIT,
  OPC_ADC_IMM_Z      = OPC_ADC_IMM | ZBIT,
  OPC_ADC_IMM_ALL    = OPC_ADC_IMM | ALLBITS,

  OPC_ADC_ZP         = OPC_ADC_IMM + BLOCKSIZE,
  OPC_ADC_ZP_C       = OPC_ADC_ZP | CBIT,
  OPC_ADC_ZP_N       = OPC_ADC_ZP | NBIT,
  OPC_ADC_ZP_V       = OPC_ADC_ZP | VBIT,
  OPC_ADC_ZP_Z       = OPC_ADC_ZP | ZBIT,
  OPC_ADC_ZP_ALL     = OPC_ADC_ZP | ALLBITS,

  OPC_ADC_ZP_X       = OPC_ADC_ZP + BLOCKSIZE,
  OPC_ADC_ZP_X_C     = OPC_ADC_ZP_X | CBIT,
  OPC_ADC_ZP_X_N     = OPC_ADC_ZP_X | NBIT,
  OPC_ADC_ZP_X_V     = OPC_ADC_ZP_X | VBIT,
  OPC_ADC_ZP_X_Z     = OPC_ADC_ZP_X | ZBIT,
  OPC_ADC_ZP_X_ALL   = OPC_ADC_ZP_X | ALLBITS,

  OPC_ADC_ABS        = OPC_ADC_ZP_X + BLOCKSIZE,
  OPC_ADC_ABS_C      = OPC_ADC_ABS | CBIT,
  OPC_ADC_ABS_N      = OPC_ADC_ABS | NBIT,
  OPC_ADC_ABS_V      = OPC_ADC_ABS | VBIT,
  OPC_ADC_ABS_Z      = OPC_ADC_ABS | ZBIT,
  OPC_ADC_ABS_ALL    = OPC_ADC_ABS | ALLBITS,

  OPC_ADCSAFE_ABS       = OPC_ADC_ABS + BLOCKSIZE,
  OPC_ADCSAFE_ABS_C     = OPC_ADCSAFE_ABS | CBIT,
  OPC_ADCSAFE_ABS_N     = OPC_ADCSAFE_ABS | NBIT,
  OPC_ADCSAFE_ABS_V     = OPC_ADCSAFE_ABS | VBIT,
  OPC_ADCSAFE_ABS_Z     = OPC_ADCSAFE_ABS | ZBIT,
  OPC_ADCSAFE_ABS_ALL   = OPC_ADCSAFE_ABS | ALLBITS,

  OPC_ADC_ABS_X      = OPC_ADCSAFE_ABS + BLOCKSIZE,
  OPC_ADC_ABS_X_C    = OPC_ADC_ABS_X | CBIT,
  OPC_ADC_ABS_X_N    = OPC_ADC_ABS_X | NBIT,
  OPC_ADC_ABS_X_V    = OPC_ADC_ABS_X | VBIT,
  OPC_ADC_ABS_X_Z    = OPC_ADC_ABS_X | ZBIT,
  OPC_ADC_ABS_X_ALL  = OPC_ADC_ABS_X | ALLBITS,

  OPC_ADCSAFE_ABS_X     = OPC_ADC_ABS_X + BLOCKSIZE,
  OPC_ADCSAFE_ABS_X_C   = OPC_ADCSAFE_ABS_X | CBIT,
  OPC_ADCSAFE_ABS_X_N   = OPC_ADCSAFE_ABS_X | NBIT,
  OPC_ADCSAFE_ABS_X_V   = OPC_ADCSAFE_ABS_X | VBIT,
  OPC_ADCSAFE_ABS_X_Z   = OPC_ADCSAFE_ABS_X | ZBIT,
  OPC_ADCSAFE_ABS_X_ALL = OPC_ADCSAFE_ABS_X | ALLBITS,

  OPC_ADC_ABS_Y      = OPC_ADCSAFE_ABS_X + BLOCKSIZE,
  OPC_ADC_ABS_Y_C    = OPC_ADC_ABS_Y | CBIT,
  OPC_ADC_ABS_Y_N    = OPC_ADC_ABS_Y | NBIT,
  OPC_ADC_ABS_Y_V    = OPC_ADC_ABS_Y | VBIT,
  OPC_ADC_ABS_Y_Z    = OPC_ADC_ABS_Y | ZBIT,
  OPC_ADC_ABS_Y_ALL  = OPC_ADC_ABS_Y | ALLBITS,

  OPC_ADCSAFE_ABS_Y     = OPC_ADC_ABS_Y + BLOCKSIZE,
  OPC_ADCSAFE_ABS_Y_C   = OPC_ADCSAFE_ABS_Y | CBIT,
  OPC_ADCSAFE_ABS_Y_N   = OPC_ADCSAFE_ABS_Y | NBIT,
  OPC_ADCSAFE_ABS_Y_V   = OPC_ADCSAFE_ABS_Y | VBIT,
  OPC_ADCSAFE_ABS_Y_Z   = OPC_ADCSAFE_ABS_Y | ZBIT,
  OPC_ADCSAFE_ABS_Y_ALL = OPC_ADCSAFE_ABS_Y | ALLBITS,

  OPC_ADC_IND        = OPC_ADCSAFE_ABS_Y + BLOCKSIZE,
  OPC_ADC_IND_C      = OPC_ADC_IND | CBIT,
  OPC_ADC_IND_N      = OPC_ADC_IND | NBIT,
  OPC_ADC_IND_V      = OPC_ADC_IND | VBIT,
  OPC_ADC_IND_Z      = OPC_ADC_IND | ZBIT,
  OPC_ADC_IND_ALL    = OPC_ADC_IND | ALLBITS,

  OPC_ADC_IND_X      = OPC_ADC_IND + BLOCKSIZE,
  OPC_ADC_IND_X_C    = OPC_ADC_IND_X | CBIT,
  OPC_ADC_IND_X_N    = OPC_ADC_IND_X | NBIT,
  OPC_ADC_IND_X_V    = OPC_ADC_IND_X | VBIT,
  OPC_ADC_IND_X_Z    = OPC_ADC_IND_X | ZBIT,
  OPC_ADC_IND_X_ALL  = OPC_ADC_IND_X | ALLBITS,

  OPC_ADC_IND_Y      = OPC_ADC_IND_X + BLOCKSIZE,
  OPC_ADC_IND_Y_C    = OPC_ADC_IND_Y | CBIT,
  OPC_ADC_IND_Y_N    = OPC_ADC_IND_Y | NBIT,
  OPC_ADC_IND_Y_V    = OPC_ADC_IND_Y | VBIT,
  OPC_ADC_IND_Y_Z    = OPC_ADC_IND_Y | ZBIT,
  OPC_ADC_IND_Y_ALL  = OPC_ADC_IND_Y | ALLBITS,

  OPC_AND_IMM        = OPC_ADC_IND_Y + BLOCKSIZE,
  OPC_AND_IMM_N      = OPC_AND_IMM | NBIT,
  OPC_AND_IMM_Z      = OPC_AND_IMM | ZBIT,
  OPC_AND_IMM_ALL    = OPC_AND_IMM | ALLBITS,

  OPC_AND_ZP         = OPC_AND_IMM + BLOCKSIZE,
  OPC_AND_ZP_N       = OPC_AND_ZP | NBIT,
  OPC_AND_ZP_Z       = OPC_AND_ZP | ZBIT,
  OPC_AND_ZP_ALL     = OPC_AND_ZP | ALLBITS,

  OPC_AND_ZP_X       = OPC_AND_ZP + BLOCKSIZE,
  OPC_AND_ZP_X_N     = OPC_AND_ZP_X | NBIT,
  OPC_AND_ZP_X_Z     = OPC_AND_ZP_X | ZBIT,
  OPC_AND_ZP_X_ALL   = OPC_AND_ZP_X | ALLBITS,

  OPC_AND_ABS        = OPC_AND_ZP_X + BLOCKSIZE,
  OPC_AND_ABS_N      = OPC_AND_ABS | NBIT,
  OPC_AND_ABS_Z      = OPC_AND_ABS | ZBIT,
  OPC_AND_ABS_ALL    = OPC_AND_ABS | ALLBITS,

  OPC_AND_ABS_X      = OPC_AND_ABS + BLOCKSIZE,
  OPC_AND_ABS_X_N    = OPC_AND_ABS_X | NBIT,
  OPC_AND_ABS_X_Z    = OPC_AND_ABS_X | ZBIT,
  OPC_AND_ABS_X_ALL  = OPC_AND_ABS_X | ALLBITS,

  OPC_AND_ABS_Y      = OPC_AND_ABS_X + BLOCKSIZE,
  OPC_AND_ABS_Y_N    = OPC_AND_ABS_Y | NBIT,
  OPC_AND_ABS_Y_Z    = OPC_AND_ABS_Y | ZBIT,
  OPC_AND_ABS_Y_ALL  = OPC_AND_ABS_Y | ALLBITS,

  OPC_AND_IND        = OPC_AND_ABS_Y + BLOCKSIZE,
  OPC_AND_IND_N      = OPC_AND_IND | NBIT,
  OPC_AND_IND_Z      = OPC_AND_IND | ZBIT,
  OPC_AND_IND_ALL    = OPC_AND_IND | ALLBITS,

  OPC_AND_IND_X      = OPC_AND_IND + BLOCKSIZE,
  OPC_AND_IND_X_N    = OPC_AND_IND_X | NBIT,
  OPC_AND_IND_X_Z    = OPC_AND_IND_X | ZBIT,
  OPC_AND_IND_X_ALL  = OPC_AND_IND_X | ALLBITS,

  OPC_AND_IND_Y      = OPC_AND_IND_X + BLOCKSIZE,
  OPC_AND_IND_Y_N    = OPC_AND_IND_Y | NBIT,
  OPC_AND_IND_Y_Z    = OPC_AND_IND_Y | ZBIT,
  OPC_AND_IND_Y_ALL  = OPC_AND_IND_Y | ALLBITS,

  OPC_ASL_ACC        = OPC_AND_IND_Y + BLOCKSIZE,
  OPC_ASL_ACC_C      = OPC_ASL_ACC | CBIT,
  OPC_ASL_ACC_N      = OPC_ASL_ACC | NBIT,
  OPC_ASL_ACC_Z      = OPC_ASL_ACC | ZBIT,
  OPC_ASL_ACC_ALL    = OPC_ASL_ACC | ALLBITS,

  OPC_ASL_ZP         = OPC_ASL_ACC + BLOCKSIZE,
  OPC_ASL_ZP_C       = OPC_ASL_ZP | CBIT,
  OPC_ASL_ZP_N       = OPC_ASL_ZP | NBIT,
  OPC_ASL_ZP_Z       = OPC_ASL_ZP | ZBIT,
  OPC_ASL_ZP_ALL     = OPC_ASL_ZP | ALLBITS,

  OPC_ASL_ZP_X       = OPC_ASL_ZP + BLOCKSIZE,
  OPC_ASL_ZP_X_C     = OPC_ASL_ZP_X | CBIT,
  OPC_ASL_ZP_X_N     = OPC_ASL_ZP_X | NBIT,
  OPC_ASL_ZP_X_Z     = OPC_ASL_ZP_X | ZBIT,
  OPC_ASL_ZP_X_ALL   = OPC_ASL_ZP_X | ALLBITS,

  OPC_ASL_ABS        = OPC_ASL_ZP_X + BLOCKSIZE,
  OPC_ASL_ABS_C      = OPC_ASL_ABS | CBIT,
  OPC_ASL_ABS_N      = OPC_ASL_ABS | NBIT,
  OPC_ASL_ABS_Z      = OPC_ASL_ABS | ZBIT,
  OPC_ASL_ABS_ALL    = OPC_ASL_ABS | ALLBITS,

  OPC_ASL_ABS_X      = OPC_ASL_ABS + BLOCKSIZE,
  OPC_ASL_ABS_X_C    = OPC_ASL_ABS_X | CBIT,
  OPC_ASL_ABS_X_N    = OPC_ASL_ABS_X | NBIT,
  OPC_ASL_ABS_X_Z    = OPC_ASL_ABS_X | ZBIT,
  OPC_ASL_ABS_X_ALL  = OPC_ASL_ABS_X | ALLBITS,

  OPC_BCC            = OPC_ASL_ABS_X + BLOCKSIZE,

  OPC_BCS            = OPC_BCC + BLOCKSIZE,

  OPC_BEQ            = OPC_BCS + BLOCKSIZE,

  OPC_BMI            = OPC_BEQ + BLOCKSIZE,

  OPC_BNE            = OPC_BMI + BLOCKSIZE,

  OPC_BPL            = OPC_BNE + BLOCKSIZE,

  OPC_BRA            = OPC_BPL + BLOCKSIZE,

  OPC_BVC            = OPC_BRA + BLOCKSIZE,

  OPC_BVS            = OPC_BVC + BLOCKSIZE,

  OPC_BIT_IMM        = OPC_BVS + BLOCKSIZE,
  OPC_BIT_IMM_N      = OPC_BIT_IMM | NBIT,
  OPC_BIT_IMM_Z      = OPC_BIT_IMM | ZBIT,
  OPC_BIT_IMM_V      = OPC_BIT_IMM | VBIT,
  OPC_BIT_IMM_ALL    = OPC_BIT_IMM | ALLBITS,

  OPC_BIT_ZP         = OPC_BIT_IMM + BLOCKSIZE,
  OPC_BIT_ZP_N       = OPC_BIT_ZP | NBIT,
  OPC_BIT_ZP_Z       = OPC_BIT_ZP | ZBIT,
  OPC_BIT_ZP_V       = OPC_BIT_ZP | VBIT,
  OPC_BIT_ZP_ALL     = OPC_BIT_ZP | ALLBITS,

  OPC_BIT_ABS        = OPC_BIT_ZP + BLOCKSIZE,
  OPC_BIT_ABS_N      = OPC_BIT_ABS | NBIT,
  OPC_BIT_ABS_Z      = OPC_BIT_ABS | ZBIT,
  OPC_BIT_ABS_V      = OPC_BIT_ABS | VBIT,
  OPC_BIT_ABS_ALL    = OPC_BIT_ABS | ALLBITS,

  OPC_BIT_ZP_X       = OPC_BIT_ABS + BLOCKSIZE,
  OPC_BIT_ZP_X_N     = OPC_BIT_ZP_X | NBIT,
  OPC_BIT_ZP_X_Z     = OPC_BIT_ZP_X | ZBIT,
  OPC_BIT_ZP_X_V     = OPC_BIT_ZP_X | VBIT,
  OPC_BIT_ZP_X_ALL   = OPC_BIT_ZP_X | ALLBITS,

  OPC_BIT_ABS_X      = OPC_BIT_ZP_X + BLOCKSIZE,
  OPC_BIT_ABS_X_N    = OPC_BIT_ABS_X | NBIT,
  OPC_BIT_ABS_X_Z    = OPC_BIT_ABS_X | ZBIT,
  OPC_BIT_ABS_X_V    = OPC_BIT_ABS_X | VBIT,
  OPC_BIT_ABS_X_ALL  = OPC_BIT_ABS_X | ALLBITS,

  OPC_BRK            = OPC_BIT_ABS_X + BLOCKSIZE,

  OPC_CLC            = OPC_BRK + BLOCKSIZE,

  OPC_CLD            = OPC_CLC + BLOCKSIZE,

  OPC_CLI            = OPC_CLD + BLOCKSIZE,

  OPC_CLV            = OPC_CLI + BLOCKSIZE,

  OPC_CMP_IMM        = OPC_CLV + BLOCKSIZE,
  OPC_CMP_IMM_C      = OPC_CMP_IMM | CBIT,
  OPC_CMP_IMM_N      = OPC_CMP_IMM | NBIT,
  OPC_CMP_IMM_Z      = OPC_CMP_IMM | ZBIT,
  OPC_CMP_IMM_ALL    = OPC_CMP_IMM | ALLBITS,

  OPC_CMP_ZP         = OPC_CMP_IMM + BLOCKSIZE,
  OPC_CMP_ZP_C       = OPC_CMP_ZP | CBIT,
  OPC_CMP_ZP_N       = OPC_CMP_ZP | NBIT,
  OPC_CMP_ZP_Z       = OPC_CMP_ZP | ZBIT,
  OPC_CMP_ZP_ALL     = OPC_CMP_ZP | ALLBITS,

  OPC_CMP_ZP_X       = OPC_CMP_ZP + BLOCKSIZE,
  OPC_CMP_ZP_X_C     = OPC_CMP_ZP_X | CBIT,
  OPC_CMP_ZP_X_N     = OPC_CMP_ZP_X | NBIT,
  OPC_CMP_ZP_X_Z     = OPC_CMP_ZP_X | ZBIT,
  OPC_CMP_ZP_X_ALL   = OPC_CMP_ZP_X | ALLBITS,

  OPC_CMP_ABS        = OPC_CMP_ZP_X + BLOCKSIZE,
  OPC_CMP_ABS_C      = OPC_CMP_ABS | CBIT,
  OPC_CMP_ABS_N      = OPC_CMP_ABS | NBIT,
  OPC_CMP_ABS_Z      = OPC_CMP_ABS | ZBIT,
  OPC_CMP_ABS_ALL    = OPC_CMP_ABS | ALLBITS,

  OPC_CMPSAFE_ABS       = OPC_CMP_ABS + BLOCKSIZE,
  OPC_CMPSAFE_ABS_C     = OPC_CMPSAFE_ABS | CBIT,
  OPC_CMPSAFE_ABS_N     = OPC_CMPSAFE_ABS | NBIT,
  OPC_CMPSAFE_ABS_Z     = OPC_CMPSAFE_ABS | ZBIT,
  OPC_CMPSAFE_ABS_ALL   = OPC_CMPSAFE_ABS | ALLBITS,

  OPC_CMP_ABS_X      = OPC_CMPSAFE_ABS + BLOCKSIZE,
  OPC_CMP_ABS_X_C    = OPC_CMP_ABS_X | CBIT,
  OPC_CMP_ABS_X_N    = OPC_CMP_ABS_X | NBIT,
  OPC_CMP_ABS_X_Z    = OPC_CMP_ABS_X | ZBIT,
  OPC_CMP_ABS_X_ALL  = OPC_CMP_ABS_X | ALLBITS,

  OPC_CMPSAFE_ABS_X     = OPC_CMP_ABS_X + BLOCKSIZE,
  OPC_CMPSAFE_ABS_X_C   = OPC_CMPSAFE_ABS_X | CBIT,
  OPC_CMPSAFE_ABS_X_N   = OPC_CMPSAFE_ABS_X | NBIT,
  OPC_CMPSAFE_ABS_X_Z   = OPC_CMPSAFE_ABS_X | ZBIT,
  OPC_CMPSAFE_ABS_X_ALL = OPC_CMPSAFE_ABS_X | ALLBITS,

  OPC_CMP_ABS_Y      = OPC_CMPSAFE_ABS_X + BLOCKSIZE,
  OPC_CMP_ABS_Y_C    = OPC_CMP_ABS_Y | CBIT,
  OPC_CMP_ABS_Y_N    = OPC_CMP_ABS_Y | NBIT,
  OPC_CMP_ABS_Y_Z    = OPC_CMP_ABS_Y | ZBIT,
  OPC_CMP_ABS_Y_ALL  = OPC_CMP_ABS_Y | ALLBITS,

  OPC_CMPSAFE_ABS_Y     = OPC_CMP_ABS_Y + BLOCKSIZE,
  OPC_CMPSAFE_ABS_Y_C   = OPC_CMPSAFE_ABS_Y | CBIT,
  OPC_CMPSAFE_ABS_Y_N   = OPC_CMPSAFE_ABS_Y | NBIT,
  OPC_CMPSAFE_ABS_Y_Z   = OPC_CMPSAFE_ABS_Y | ZBIT,
  OPC_CMPSAFE_ABS_Y_ALL = OPC_CMPSAFE_ABS_Y | ALLBITS,

  OPC_CMP_IND        = OPC_CMPSAFE_ABS_Y + BLOCKSIZE,
  OPC_CMP_IND_C      = OPC_CMP_IND | CBIT,
  OPC_CMP_IND_N      = OPC_CMP_IND | NBIT,
  OPC_CMP_IND_Z      = OPC_CMP_IND | ZBIT,
  OPC_CMP_IND_ALL    = OPC_CMP_IND | ALLBITS,

  OPC_CMP_IND_X      = OPC_CMP_IND + BLOCKSIZE,
  OPC_CMP_IND_X_C    = OPC_CMP_IND_X | CBIT,
  OPC_CMP_IND_X_N    = OPC_CMP_IND_X | NBIT,
  OPC_CMP_IND_X_Z    = OPC_CMP_IND_X | ZBIT,
  OPC_CMP_IND_X_ALL  = OPC_CMP_IND_X | ALLBITS,

  OPC_CMP_IND_Y      = OPC_CMP_IND_X + BLOCKSIZE,
  OPC_CMP_IND_Y_C    = OPC_CMP_IND_Y | CBIT,
  OPC_CMP_IND_Y_N    = OPC_CMP_IND_Y | NBIT,
  OPC_CMP_IND_Y_Z    = OPC_CMP_IND_Y | ZBIT,
  OPC_CMP_IND_Y_ALL  = OPC_CMP_IND_Y | ALLBITS,

  OPC_CPX_IMM        = OPC_CMP_IND_Y + BLOCKSIZE,
  OPC_CPX_IMM_C      = OPC_CPX_IMM | CBIT,
  OPC_CPX_IMM_N      = OPC_CPX_IMM | NBIT,
  OPC_CPX_IMM_Z      = OPC_CPX_IMM | ZBIT,
  OPC_CPX_IMM_ALL    = OPC_CPX_IMM | ALLBITS,

  OPC_CPX_ZP         = OPC_CPX_IMM + BLOCKSIZE,
  OPC_CPX_ZP_C       = OPC_CPX_ZP | CBIT,
  OPC_CPX_ZP_N       = OPC_CPX_ZP | NBIT,
  OPC_CPX_ZP_Z       = OPC_CPX_ZP | ZBIT,
  OPC_CPX_ZP_ALL     = OPC_CPX_ZP | ALLBITS,

  OPC_CPX_ABS        = OPC_CPX_ZP + BLOCKSIZE,
  OPC_CPX_ABS_C      = OPC_CPX_ABS | CBIT,
  OPC_CPX_ABS_N      = OPC_CPX_ABS | NBIT,
  OPC_CPX_ABS_Z      = OPC_CPX_ABS | ZBIT,
  OPC_CPX_ABS_ALL    = OPC_CPX_ABS | ALLBITS,

  OPC_CPXSAFE_ABS     = OPC_CPX_ABS + BLOCKSIZE,
  OPC_CPXSAFE_ABS_C   = OPC_CPXSAFE_ABS | CBIT,
  OPC_CPXSAFE_ABS_N   = OPC_CPXSAFE_ABS | NBIT,
  OPC_CPXSAFE_ABS_Z   = OPC_CPXSAFE_ABS | ZBIT,
  OPC_CPXSAFE_ABS_ALL = OPC_CPXSAFE_ABS | ALLBITS,

  OPC_CPY_IMM        = OPC_CPXSAFE_ABS + BLOCKSIZE,
  OPC_CPY_IMM_C      = OPC_CPY_IMM | CBIT,
  OPC_CPY_IMM_N      = OPC_CPY_IMM | NBIT,
  OPC_CPY_IMM_Z      = OPC_CPY_IMM | ZBIT,
  OPC_CPY_IMM_ALL    = OPC_CPY_IMM | ALLBITS,

  OPC_CPY_ZP         = OPC_CPY_IMM + BLOCKSIZE,
  OPC_CPY_ZP_C       = OPC_CPY_ZP | CBIT,
  OPC_CPY_ZP_N       = OPC_CPY_ZP | NBIT,
  OPC_CPY_ZP_Z       = OPC_CPY_ZP | ZBIT,
  OPC_CPY_ZP_ALL     = OPC_CPY_ZP | ALLBITS,

  OPC_CPY_ABS        = OPC_CPY_ZP + BLOCKSIZE,
  OPC_CPY_ABS_C      = OPC_CPY_ABS | CBIT,
  OPC_CPY_ABS_N      = OPC_CPY_ABS | NBIT,
  OPC_CPY_ABS_Z      = OPC_CPY_ABS | ZBIT,
  OPC_CPY_ABS_ALL    = OPC_CPY_ABS | ALLBITS,

  OPC_CPYSAFE_ABS       = OPC_CPY_ABS + BLOCKSIZE,
  OPC_CPYSAFE_ABS_C     = OPC_CPYSAFE_ABS | CBIT,
  OPC_CPYSAFE_ABS_N     = OPC_CPYSAFE_ABS | NBIT,
  OPC_CPYSAFE_ABS_Z     = OPC_CPYSAFE_ABS | ZBIT,
  OPC_CPYSAFE_ABS_ALL   = OPC_CPYSAFE_ABS | ALLBITS,

  OPC_DEC_ACC        = OPC_CPYSAFE_ABS + BLOCKSIZE,
  OPC_DEC_ACC_N      = OPC_DEC_ACC | NBIT,
  OPC_DEC_ACC_Z      = OPC_DEC_ACC | ZBIT,
  OPC_DEC_ACC_ALL    = OPC_DEC_ACC | ALLBITS,

  OPC_DEC_ZP         = OPC_DEC_ACC + BLOCKSIZE,
  OPC_DEC_ZP_N       = OPC_DEC_ZP | NBIT,
  OPC_DEC_ZP_Z       = OPC_DEC_ZP | ZBIT,
  OPC_DEC_ZP_ALL     = OPC_DEC_ZP | ALLBITS,

  OPC_DEC_ZP_X       = OPC_DEC_ZP + BLOCKSIZE,
  OPC_DEC_ZP_X_N     = OPC_DEC_ZP_X | NBIT,
  OPC_DEC_ZP_X_Z     = OPC_DEC_ZP_X | ZBIT,
  OPC_DEC_ZP_X_ALL   = OPC_DEC_ZP_X | ALLBITS,

  OPC_DEC_ABS        = OPC_DEC_ZP_X + BLOCKSIZE,
  OPC_DEC_ABS_N      = OPC_DEC_ABS | NBIT,
  OPC_DEC_ABS_Z      = OPC_DEC_ABS | ZBIT,
  OPC_DEC_ABS_ALL    = OPC_DEC_ABS | ALLBITS,

  OPC_DEC_ABS_X      = OPC_DEC_ABS + BLOCKSIZE,
  OPC_DEC_ABS_X_N    = OPC_DEC_ABS_X | NBIT,
  OPC_DEC_ABS_X_Z    = OPC_DEC_ABS_X | ZBIT,
  OPC_DEC_ABS_X_ALL  = OPC_DEC_ABS_X | ALLBITS,

  OPC_DEX            = OPC_DEC_ABS_X + BLOCKSIZE,
  OPC_DEX_N          = OPC_DEX | NBIT,
  OPC_DEX_Z          = OPC_DEX | ZBIT,
  OPC_DEX_ALL        = OPC_DEX | ALLBITS,

  OPC_DEY            = OPC_DEX + BLOCKSIZE,
  OPC_DEY_N          = OPC_DEY | NBIT,
  OPC_DEY_Z          = OPC_DEY | ZBIT,
  OPC_DEY_ALL        = OPC_DEY | ALLBITS,

  OPC_EOR_IMM        = OPC_DEY + BLOCKSIZE,
  OPC_EOR_IMM_N      = OPC_EOR_IMM | NBIT,
  OPC_EOR_IMM_Z      = OPC_EOR_IMM | ZBIT,
  OPC_EOR_IMM_ALL    = OPC_EOR_IMM | ALLBITS,

  OPC_EOR_ZP         = OPC_EOR_IMM + BLOCKSIZE,
  OPC_EOR_ZP_N       = OPC_EOR_ZP | NBIT,
  OPC_EOR_ZP_Z       = OPC_EOR_ZP | ZBIT,
  OPC_EOR_ZP_ALL     = OPC_EOR_ZP | ALLBITS,

  OPC_EOR_ZP_X       = OPC_EOR_ZP + BLOCKSIZE,
  OPC_EOR_ZP_X_N     = OPC_EOR_ZP_X | NBIT,
  OPC_EOR_ZP_X_Z     = OPC_EOR_ZP_X | ZBIT,
  OPC_EOR_ZP_X_ALL   = OPC_EOR_ZP_X | ALLBITS,

  OPC_EOR_ABS        = OPC_EOR_ZP_X + BLOCKSIZE,
  OPC_EOR_ABS_N      = OPC_EOR_ABS | NBIT,
  OPC_EOR_ABS_Z      = OPC_EOR_ABS | ZBIT,
  OPC_EOR_ABS_ALL    = OPC_EOR_ABS | ALLBITS,

  OPC_EOR_ABS_X      = OPC_EOR_ABS + BLOCKSIZE,
  OPC_EOR_ABS_X_N    = OPC_EOR_ABS_X | NBIT,
  OPC_EOR_ABS_X_Z    = OPC_EOR_ABS_X | ZBIT,
  OPC_EOR_ABS_X_ALL  = OPC_EOR_ABS_X | ALLBITS,

  OPC_EOR_ABS_Y      = OPC_EOR_ABS_X + BLOCKSIZE,
  OPC_EOR_ABS_Y_N    = OPC_EOR_ABS_Y | NBIT,
  OPC_EOR_ABS_Y_Z    = OPC_EOR_ABS_Y | ZBIT,
  OPC_EOR_ABS_Y_ALL  = OPC_EOR_ABS_Y | ALLBITS,

  OPC_EOR_IND        = OPC_EOR_ABS_Y + BLOCKSIZE,
  OPC_EOR_IND_N      = OPC_EOR_IND | NBIT,
  OPC_EOR_IND_Z      = OPC_EOR_IND | ZBIT,
  OPC_EOR_IND_ALL    = OPC_EOR_IND | ALLBITS,

  OPC_EOR_IND_X      = OPC_EOR_IND + BLOCKSIZE,
  OPC_EOR_IND_X_N    = OPC_EOR_IND_X | NBIT,
  OPC_EOR_IND_X_Z    = OPC_EOR_IND_X | ZBIT,
  OPC_EOR_IND_X_ALL  = OPC_EOR_IND_X | ALLBITS,

  OPC_EOR_IND_Y      = OPC_EOR_IND_X + BLOCKSIZE,
  OPC_EOR_IND_Y_N    = OPC_EOR_IND_Y | NBIT,
  OPC_EOR_IND_Y_Z    = OPC_EOR_IND_Y | ZBIT,
  OPC_EOR_IND_Y_ALL  = OPC_EOR_IND_Y | ALLBITS,

  OPC_INC_ZP         = OPC_EOR_IND_Y + BLOCKSIZE,
  OPC_INC_ZP_N       = OPC_INC_ZP | NBIT,
  OPC_INC_ZP_Z       = OPC_INC_ZP | ZBIT,
  OPC_INC_ZP_ALL     = OPC_INC_ZP | ALLBITS,

  OPC_INC_ACC        = OPC_INC_ZP + BLOCKSIZE,
  OPC_INC_ACC_N      = OPC_INC_ACC | NBIT,
  OPC_INC_ACC_Z      = OPC_INC_ACC | ZBIT,
  OPC_INC_ACC_ALL    = OPC_INC_ACC | ALLBITS,

  OPC_INC_ZP_X       = OPC_INC_ACC + BLOCKSIZE,
  OPC_INC_ZP_X_N     = OPC_INC_ZP_X | NBIT,
  OPC_INC_ZP_X_Z     = OPC_INC_ZP_X | ZBIT,
  OPC_INC_ZP_X_ALL   = OPC_INC_ZP_X | ALLBITS,

  OPC_INC_ABS        = OPC_INC_ZP_X + BLOCKSIZE,
  OPC_INC_ABS_N      = OPC_INC_ABS | NBIT,
  OPC_INC_ABS_Z      = OPC_INC_ABS | ZBIT,
  OPC_INC_ABS_ALL    = OPC_INC_ABS | ALLBITS,

  OPC_INC_ABS_X      = OPC_INC_ABS + BLOCKSIZE,
  OPC_INC_ABS_X_N    = OPC_INC_ABS_X | NBIT,
  OPC_INC_ABS_X_Z    = OPC_INC_ABS_X | ZBIT,
  OPC_INC_ABS_X_ALL  = OPC_INC_ABS_X | ALLBITS,

  OPC_INX            = OPC_INC_ABS_X + BLOCKSIZE,
  OPC_INX_N          = OPC_INX | NBIT,
  OPC_INX_Z          = OPC_INX | ZBIT,
  OPC_INX_ALL        = OPC_INX | ALLBITS,

  OPC_INY            = OPC_INX + BLOCKSIZE,
  OPC_INY_N          = OPC_INY | NBIT,
  OPC_INY_Z          = OPC_INY | ZBIT,
  OPC_INY_ALL        = OPC_INY | ALLBITS,

  OPC_JMP_ABS        = OPC_INY + BLOCKSIZE,

  OPC_JMP_ABSIND     = OPC_JMP_ABS + BLOCKSIZE,

  OPC_JMP_ABSIND_X   = OPC_JMP_ABSIND + BLOCKSIZE,

  OPC_JSR            = OPC_JMP_ABSIND_X + BLOCKSIZE,

  OPC_LDA_IMM        = OPC_JSR + BLOCKSIZE,
  OPC_LDA_IMM_N      = OPC_LDA_IMM | NBIT,
  OPC_LDA_IMM_Z      = OPC_LDA_IMM | ZBIT,
  OPC_LDA_IMM_ALL    = OPC_LDA_IMM | ALLBITS,

  OPC_LDA_ZP         = OPC_LDA_IMM + BLOCKSIZE,
  OPC_LDA_ZP_N       = OPC_LDA_ZP | NBIT,
  OPC_LDA_ZP_Z       = OPC_LDA_ZP | ZBIT,
  OPC_LDA_ZP_ALL     = OPC_LDA_ZP | ALLBITS,

  OPC_LDA_ZP_X       = OPC_LDA_ZP + BLOCKSIZE,
  OPC_LDA_ZP_X_N     = OPC_LDA_ZP_X | NBIT,
  OPC_LDA_ZP_X_Z     = OPC_LDA_ZP_X | ZBIT,
  OPC_LDA_ZP_X_ALL   = OPC_LDA_ZP_X | ALLBITS,

  OPC_LDA_ABS        = OPC_LDA_ZP_X + BLOCKSIZE,
  OPC_LDA_ABS_N      = OPC_LDA_ABS | NBIT,
  OPC_LDA_ABS_Z      = OPC_LDA_ABS | ZBIT,
  OPC_LDA_ABS_ALL    = OPC_LDA_ABS | ALLBITS,

  OPC_LDASAFE_ABS     = OPC_LDA_ABS + BLOCKSIZE,
  OPC_LDASAFE_ABS_N   = OPC_LDASAFE_ABS | NBIT,
  OPC_LDASAFE_ABS_Z   = OPC_LDASAFE_ABS | ZBIT,
  OPC_LDASAFE_ABS_ALL = OPC_LDASAFE_ABS | ALLBITS,

  OPC_LDA_ABS_X      = OPC_LDASAFE_ABS + BLOCKSIZE,
  OPC_LDA_ABS_X_N    = OPC_LDA_ABS_X | NBIT,
  OPC_LDA_ABS_X_Z    = OPC_LDA_ABS_X | ZBIT,
  OPC_LDA_ABS_X_ALL  = OPC_LDA_ABS_X | ALLBITS,

  OPC_LDASAFE_ABS_X     = OPC_LDA_ABS_X + BLOCKSIZE,
  OPC_LDASAFE_ABS_X_N   = OPC_LDASAFE_ABS_X | NBIT,
  OPC_LDASAFE_ABS_X_Z   = OPC_LDASAFE_ABS_X | ZBIT,
  OPC_LDASAFE_ABS_X_ALL = OPC_LDASAFE_ABS_X | ALLBITS,

  OPC_LDA_ABS_Y      = OPC_LDASAFE_ABS_X + BLOCKSIZE,
  OPC_LDA_ABS_Y_N    = OPC_LDA_ABS_Y | NBIT,
  OPC_LDA_ABS_Y_Z    = OPC_LDA_ABS_Y | ZBIT,
  OPC_LDA_ABS_Y_ALL  = OPC_LDA_ABS_Y | ALLBITS,

  OPC_LDASAFE_ABS_Y     = OPC_LDA_ABS_Y + BLOCKSIZE,
  OPC_LDASAFE_ABS_Y_N   = OPC_LDASAFE_ABS_Y | NBIT,
  OPC_LDASAFE_ABS_Y_Z   = OPC_LDASAFE_ABS_Y | ZBIT,
  OPC_LDASAFE_ABS_Y_ALL = OPC_LDASAFE_ABS_Y | ALLBITS,

  OPC_LDA_IND        = OPC_LDASAFE_ABS_Y + BLOCKSIZE,
  OPC_LDA_IND_N      = OPC_LDA_IND | NBIT,
  OPC_LDA_IND_Z      = OPC_LDA_IND | ZBIT,
  OPC_LDA_IND_ALL    = OPC_LDA_IND | ALLBITS,

  OPC_LDA_IND_X      = OPC_LDA_IND + BLOCKSIZE,
  OPC_LDA_IND_X_N    = OPC_LDA_IND_X | NBIT,
  OPC_LDA_IND_X_Z    = OPC_LDA_IND_X | ZBIT,
  OPC_LDA_IND_X_ALL  = OPC_LDA_IND_X | ALLBITS,

  OPC_LDA_IND_Y      = OPC_LDA_IND_X + BLOCKSIZE,
  OPC_LDA_IND_Y_N    = OPC_LDA_IND_Y | NBIT,
  OPC_LDA_IND_Y_Z    = OPC_LDA_IND_Y | ZBIT,
  OPC_LDA_IND_Y_ALL  = OPC_LDA_IND_Y | ALLBITS,

  OPC_LDX_IMM        = OPC_LDA_IND_Y + BLOCKSIZE,
  OPC_LDX_IMM_N      = OPC_LDX_IMM | NBIT,
  OPC_LDX_IMM_Z      = OPC_LDX_IMM | ZBIT,
  OPC_LDX_IMM_ALL    = OPC_LDX_IMM | ALLBITS,

  OPC_LDX_ZP         = OPC_LDX_IMM + BLOCKSIZE,
  OPC_LDX_ZP_N       = OPC_LDX_ZP | NBIT,
  OPC_LDX_ZP_Z       = OPC_LDX_ZP | ZBIT,
  OPC_LDX_ZP_ALL     = OPC_LDX_ZP | ALLBITS,

  OPC_LDX_ZP_Y       = OPC_LDX_ZP + BLOCKSIZE,
  OPC_LDX_ZP_Y_N     = OPC_LDX_ZP_Y | NBIT,
  OPC_LDX_ZP_Y_Z     = OPC_LDX_ZP_Y | ZBIT,
  OPC_LDX_ZP_Y_ALL   = OPC_LDX_ZP_Y | ALLBITS,

  OPC_LDX_ABS        = OPC_LDX_ZP_Y + BLOCKSIZE,
  OPC_LDX_ABS_N      = OPC_LDX_ABS | NBIT,
  OPC_LDX_ABS_Z      = OPC_LDX_ABS | ZBIT,
  OPC_LDX_ABS_ALL    = OPC_LDX_ABS | ALLBITS,

  OPC_LDXSAFE_ABS     = OPC_LDX_ABS + BLOCKSIZE,
  OPC_LDXSAFE_ABS_N   = OPC_LDXSAFE_ABS | NBIT,
  OPC_LDXSAFE_ABS_Z   = OPC_LDXSAFE_ABS | ZBIT,
  OPC_LDXSAFE_ABS_ALL = OPC_LDXSAFE_ABS | ALLBITS,

  OPC_LDX_ABS_Y      = OPC_LDXSAFE_ABS + BLOCKSIZE,
  OPC_LDX_ABS_Y_N    = OPC_LDX_ABS_Y | NBIT,
  OPC_LDX_ABS_Y_Z    = OPC_LDX_ABS_Y | ZBIT,
  OPC_LDX_ABS_Y_ALL  = OPC_LDX_ABS_Y | ALLBITS,

  OPC_LDXSAFE_ABS_Y     = OPC_LDX_ABS_Y + BLOCKSIZE,
  OPC_LDXSAFE_ABS_Y_N   = OPC_LDXSAFE_ABS_Y | NBIT,
  OPC_LDXSAFE_ABS_Y_Z   = OPC_LDXSAFE_ABS_Y | ZBIT,
  OPC_LDXSAFE_ABS_Y_ALL = OPC_LDXSAFE_ABS_Y | ALLBITS,

  OPC_LDY_IMM        = OPC_LDXSAFE_ABS_Y + BLOCKSIZE,
  OPC_LDY_IMM_N      = OPC_LDY_IMM | NBIT,
  OPC_LDY_IMM_Z      = OPC_LDY_IMM | ZBIT,
  OPC_LDY_IMM_ALL    = OPC_LDY_IMM | ALLBITS,

  OPC_LDY_ZP         = OPC_LDY_IMM + BLOCKSIZE,
  OPC_LDY_ZP_N       = OPC_LDY_ZP | NBIT,
  OPC_LDY_ZP_Z       = OPC_LDY_ZP | ZBIT,
  OPC_LDY_ZP_ALL     = OPC_LDY_ZP | ALLBITS,

  OPC_LDY_ZP_X       = OPC_LDY_ZP + BLOCKSIZE,
  OPC_LDY_ZP_X_N     = OPC_LDY_ZP_X | NBIT,
  OPC_LDY_ZP_X_Z     = OPC_LDY_ZP_X | ZBIT,
  OPC_LDY_ZP_X_ALL   = OPC_LDY_ZP_X | ALLBITS,

  OPC_LDY_ABS        = OPC_LDY_ZP_X + BLOCKSIZE,
  OPC_LDY_ABS_N      = OPC_LDY_ABS | NBIT,
  OPC_LDY_ABS_Z      = OPC_LDY_ABS | ZBIT,
  OPC_LDY_ABS_ALL    = OPC_LDY_ABS | ALLBITS,

  OPC_LDYSAFE_ABS     = OPC_LDY_ABS + BLOCKSIZE,
  OPC_LDYSAFE_ABS_N   = OPC_LDYSAFE_ABS | NBIT,
  OPC_LDYSAFE_ABS_Z   = OPC_LDYSAFE_ABS | ZBIT,
  OPC_LDYSAFE_ABS_ALL = OPC_LDYSAFE_ABS | ALLBITS,

  OPC_LDY_ABS_X      = OPC_LDYSAFE_ABS + BLOCKSIZE,
  OPC_LDY_ABS_X_N    = OPC_LDY_ABS_X | NBIT,
  OPC_LDY_ABS_X_Z    = OPC_LDY_ABS_X | ZBIT,
  OPC_LDY_ABS_X_ALL  = OPC_LDY_ABS_X | ALLBITS,

  OPC_LDYSAFE_ABS_X     = OPC_LDY_ABS_X + BLOCKSIZE,
  OPC_LDYSAFE_ABS_X_N   = OPC_LDYSAFE_ABS_X | NBIT,
  OPC_LDYSAFE_ABS_X_Z   = OPC_LDYSAFE_ABS_X | ZBIT,
  OPC_LDYSAFE_ABS_X_ALL = OPC_LDYSAFE_ABS_X | ALLBITS,

  OPC_LSR_ACC        = OPC_LDYSAFE_ABS_X + BLOCKSIZE,
  OPC_LSR_ACC_C      = OPC_LSR_ACC | CBIT,
  OPC_LSR_ACC_N      = OPC_LSR_ACC | NBIT,
  OPC_LSR_ACC_Z      = OPC_LSR_ACC | ZBIT,
  OPC_LSR_ACC_ALL    = OPC_LSR_ACC | ALLBITS,

  OPC_LSR_ZP         = OPC_LSR_ACC + BLOCKSIZE,
  OPC_LSR_ZP_C       = OPC_LSR_ZP | CBIT,
  OPC_LSR_ZP_N       = OPC_LSR_ZP | NBIT,
  OPC_LSR_ZP_Z       = OPC_LSR_ZP | ZBIT,
  OPC_LSR_ZP_ALL     = OPC_LSR_ZP | ALLBITS,

  OPC_LSR_ZP_X       = OPC_LSR_ZP + BLOCKSIZE,
  OPC_LSR_ZP_X_C     = OPC_LSR_ZP_X | CBIT,
  OPC_LSR_ZP_X_N     = OPC_LSR_ZP_X | NBIT,
  OPC_LSR_ZP_X_Z     = OPC_LSR_ZP_X | ZBIT,
  OPC_LSR_ZP_X_ALL   = OPC_LSR_ZP_X | ALLBITS,

  OPC_LSR_ABS        = OPC_LSR_ZP_X + BLOCKSIZE,
  OPC_LSR_ABS_C      = OPC_LSR_ABS | CBIT,
  OPC_LSR_ABS_N      = OPC_LSR_ABS | NBIT,
  OPC_LSR_ABS_Z      = OPC_LSR_ABS | ZBIT,
  OPC_LSR_ABS_ALL    = OPC_LSR_ABS | ALLBITS,

  OPC_LSR_ABS_X      = OPC_LSR_ABS + BLOCKSIZE,
  OPC_LSR_ABS_X_C    = OPC_LSR_ABS_X | CBIT,
  OPC_LSR_ABS_X_N    = OPC_LSR_ABS_X | NBIT,
  OPC_LSR_ABS_X_Z    = OPC_LSR_ABS_X | ZBIT,
  OPC_LSR_ABS_X_ALL  = OPC_LSR_ABS_X | ALLBITS,

  OPC_NOP            = OPC_LSR_ABS_X + BLOCKSIZE,

  OPC_OR_IMM         = OPC_NOP + BLOCKSIZE,
  OPC_OR_IMM_N       = OPC_OR_IMM | NBIT,
  OPC_OR_IMM_Z       = OPC_OR_IMM | ZBIT,
  OPC_OR_IMM_ALL     = OPC_OR_IMM | ALLBITS,

  OPC_OR_ZP          = OPC_OR_IMM + BLOCKSIZE,
  OPC_OR_ZP_N        = OPC_OR_ZP | NBIT,
  OPC_OR_ZP_Z        = OPC_OR_ZP | ZBIT,
  OPC_OR_ZP_ALL      = OPC_OR_ZP | ALLBITS,

  OPC_OR_ZP_X        = OPC_OR_ZP + BLOCKSIZE,
  OPC_OR_ZP_X_N      = OPC_OR_ZP_X | NBIT,
  OPC_OR_ZP_X_Z      = OPC_OR_ZP_X | ZBIT,
  OPC_OR_ZP_X_ALL    = OPC_OR_ZP_X | ALLBITS,

  OPC_OR_ABS         = OPC_OR_ZP_X + BLOCKSIZE,
  OPC_OR_ABS_N       = OPC_OR_ABS | NBIT,
  OPC_OR_ABS_Z       = OPC_OR_ABS | ZBIT,
  OPC_OR_ABS_ALL     = OPC_OR_ABS | ALLBITS,

  OPC_OR_ABS_X       = OPC_OR_ABS + BLOCKSIZE,
  OPC_OR_ABS_X_N     = OPC_OR_ABS_X | NBIT,
  OPC_OR_ABS_X_Z     = OPC_OR_ABS_X | ZBIT,
  OPC_OR_ABS_X_ALL   = OPC_OR_ABS_X | ALLBITS,

  OPC_OR_ABS_Y       = OPC_OR_ABS_X + BLOCKSIZE,
  OPC_OR_ABS_Y_N     = OPC_OR_ABS_Y | NBIT,
  OPC_OR_ABS_Y_Z     = OPC_OR_ABS_Y | ZBIT,
  OPC_OR_ABS_Y_ALL   = OPC_OR_ABS_Y | ALLBITS,

  OPC_OR_IND         = OPC_OR_ABS_Y + BLOCKSIZE,
  OPC_OR_IND_N       = OPC_OR_IND | NBIT,
  OPC_OR_IND_Z       = OPC_OR_IND | ZBIT,
  OPC_OR_IND_ALL     = OPC_OR_IND | ALLBITS,

  OPC_OR_IND_X       = OPC_OR_IND + BLOCKSIZE,
  OPC_OR_IND_X_N     = OPC_OR_IND_X | NBIT,
  OPC_OR_IND_X_Z     = OPC_OR_IND_X | ZBIT,
  OPC_OR_IND_X_ALL   = OPC_OR_IND_X | ALLBITS,

  OPC_OR_IND_Y       = OPC_OR_IND_X + BLOCKSIZE,
  OPC_OR_IND_Y_N     = OPC_OR_IND_Y | NBIT,
  OPC_OR_IND_Y_Z     = OPC_OR_IND_Y | ZBIT,
  OPC_OR_IND_Y_ALL   = OPC_OR_IND_Y | ALLBITS,

  OPC_PHA            = OPC_OR_IND_Y + BLOCKSIZE,

  OPC_PHX            = OPC_PHA + BLOCKSIZE,

  OPC_PHY            = OPC_PHX + BLOCKSIZE,

  OPC_PLX            = OPC_PHY + BLOCKSIZE,
  OPC_PLX_N          = OPC_PLX | NBIT,
  OPC_PLX_Z          = OPC_PLX | ZBIT,
  OPC_PLX_ALL        = OPC_PLX | ALLBITS,

  OPC_PLY            = OPC_PLX + BLOCKSIZE,
  OPC_PLY_N          = OPC_PLY | NBIT,
  OPC_PLY_Z          = OPC_PLY | ZBIT,
  OPC_PLY_ALL        = OPC_PLY | ALLBITS,

  OPC_PHP            = OPC_PLY + BLOCKSIZE,

  OPC_PLA            = OPC_PHP + BLOCKSIZE,
  OPC_PLA_N          = OPC_PLA | NBIT,
  OPC_PLA_Z          = OPC_PLA | ZBIT,
  OPC_PLA_ALL        = OPC_PLA | ALLBITS,

  OPC_PLP            = OPC_PLA + BLOCKSIZE,
  OPC_PLP_C          = OPC_PLP | CBIT,
  OPC_PLP_N          = OPC_PLP | NBIT,
  OPC_PLP_V          = OPC_PLP | VBIT,
  OPC_PLP_Z          = OPC_PLP | ZBIT,
  OPC_PLP_ALL        = OPC_PLP | ALLBITS,

  OPC_ROL_ACC        = OPC_PLP + BLOCKSIZE,
  OPC_ROL_ACC_C      = OPC_ROL_ACC | CBIT,
  OPC_ROL_ACC_N      = OPC_ROL_ACC | NBIT,
  OPC_ROL_ACC_Z      = OPC_ROL_ACC | ZBIT,
  OPC_ROL_ACC_ALL    = OPC_ROL_ACC | ALLBITS,

  OPC_ROL_ZP         = OPC_ROL_ACC + BLOCKSIZE,
  OPC_ROL_ZP_C       = OPC_ROL_ZP | CBIT,
  OPC_ROL_ZP_N       = OPC_ROL_ZP | NBIT,
  OPC_ROL_ZP_Z       = OPC_ROL_ZP | ZBIT,
  OPC_ROL_ZP_ALL     = OPC_ROL_ZP | ALLBITS,

  OPC_ROL_ZP_X       = OPC_ROL_ZP + BLOCKSIZE,
  OPC_ROL_ZP_X_C     = OPC_ROL_ZP_X | CBIT,
  OPC_ROL_ZP_X_N     = OPC_ROL_ZP_X | NBIT,
  OPC_ROL_ZP_X_Z     = OPC_ROL_ZP_X | ZBIT,
  OPC_ROL_ZP_X_ALL   = OPC_ROL_ZP_X | ALLBITS,

  OPC_ROL_ABS        = OPC_ROL_ZP_X + BLOCKSIZE,
  OPC_ROL_ABS_C      = OPC_ROL_ABS | CBIT,
  OPC_ROL_ABS_N      = OPC_ROL_ABS | NBIT,
  OPC_ROL_ABS_Z      = OPC_ROL_ABS | ZBIT,
  OPC_ROL_ABS_ALL    = OPC_ROL_ABS | ALLBITS,

  OPC_ROL_ABS_X      = OPC_ROL_ABS + BLOCKSIZE,
  OPC_ROL_ABS_X_C    = OPC_ROL_ABS_X | CBIT,
  OPC_ROL_ABS_X_N    = OPC_ROL_ABS_X | NBIT,
  OPC_ROL_ABS_X_Z    = OPC_ROL_ABS_X | ZBIT,
  OPC_ROL_ABS_X_ALL  = OPC_ROL_ABS_X | ALLBITS,

  OPC_ROR_ACC        = OPC_ROL_ABS_X + BLOCKSIZE,
  OPC_ROR_ACC_C      = OPC_ROR_ACC | CBIT,
  OPC_ROR_ACC_N      = OPC_ROR_ACC | NBIT,
  OPC_ROR_ACC_Z      = OPC_ROR_ACC | ZBIT,
  OPC_ROR_ACC_ALL    = OPC_ROR_ACC | ALLBITS,

  OPC_ROR_ZP         = OPC_ROR_ACC + BLOCKSIZE,
  OPC_ROR_ZP_C       = OPC_ROR_ZP | CBIT,
  OPC_ROR_ZP_N       = OPC_ROR_ZP | NBIT,
  OPC_ROR_ZP_Z       = OPC_ROR_ZP | ZBIT,
  OPC_ROR_ZP_ALL     = OPC_ROR_ZP | ALLBITS,

  OPC_ROR_ZP_X       = OPC_ROR_ZP + BLOCKSIZE,
  OPC_ROR_ZP_X_C     = OPC_ROR_ZP_X | CBIT,
  OPC_ROR_ZP_X_N     = OPC_ROR_ZP_X | NBIT,
  OPC_ROR_ZP_X_Z     = OPC_ROR_ZP_X | ZBIT,
  OPC_ROR_ZP_X_ALL   = OPC_ROR_ZP_X | ALLBITS,

  OPC_ROR_ABS        = OPC_ROR_ZP_X + BLOCKSIZE,
  OPC_ROR_ABS_C      = OPC_ROR_ABS | CBIT,
  OPC_ROR_ABS_N      = OPC_ROR_ABS | NBIT,
  OPC_ROR_ABS_Z      = OPC_ROR_ABS | ZBIT,
  OPC_ROR_ABS_ALL    = OPC_ROR_ABS | ALLBITS,

  OPC_ROR_ABS_X      = OPC_ROR_ABS + BLOCKSIZE,
  OPC_ROR_ABS_X_C    = OPC_ROR_ABS_X | CBIT,
  OPC_ROR_ABS_X_N    = OPC_ROR_ABS_X | NBIT,
  OPC_ROR_ABS_X_Z    = OPC_ROR_ABS_X | ZBIT,
  OPC_ROR_ABS_X_ALL  = OPC_ROR_ABS_X | ALLBITS,

  OPC_RTI            = OPC_ROR_ABS_X + BLOCKSIZE,

  OPC_RTS            = OPC_RTI + BLOCKSIZE,

  OPC_SBC_IMM        = OPC_RTS + BLOCKSIZE,
  OPC_SBC_IMM_C      = OPC_SBC_IMM | CBIT,
  OPC_SBC_IMM_N      = OPC_SBC_IMM | NBIT,
  OPC_SBC_IMM_V      = OPC_SBC_IMM | VBIT,
  OPC_SBC_IMM_Z      = OPC_SBC_IMM | ZBIT,
  OPC_SBC_IMM_ALL    = OPC_SBC_IMM | ALLBITS,

  OPC_SBC_ZP         = OPC_SBC_IMM + BLOCKSIZE,
  OPC_SBC_ZP_C       = OPC_SBC_ZP | CBIT,
  OPC_SBC_ZP_N       = OPC_SBC_ZP | NBIT,
  OPC_SBC_ZP_V       = OPC_SBC_ZP | VBIT,
  OPC_SBC_ZP_Z       = OPC_SBC_ZP | ZBIT,
  OPC_SBC_ZP_ALL     = OPC_SBC_ZP | ALLBITS,

  OPC_SBC_ZP_X       = OPC_SBC_ZP + BLOCKSIZE,
  OPC_SBC_ZP_X_C     = OPC_SBC_ZP_X | CBIT,
  OPC_SBC_ZP_X_N     = OPC_SBC_ZP_X | NBIT,
  OPC_SBC_ZP_X_V     = OPC_SBC_ZP_X | VBIT,
  OPC_SBC_ZP_X_Z     = OPC_SBC_ZP_X | ZBIT,
  OPC_SBC_ZP_X_ALL   = OPC_SBC_ZP_X | ALLBITS,

  OPC_SBC_ABS        = OPC_SBC_ZP_X + BLOCKSIZE,
  OPC_SBC_ABS_C      = OPC_SBC_ABS | CBIT,
  OPC_SBC_ABS_N      = OPC_SBC_ABS | NBIT,
  OPC_SBC_ABS_V      = OPC_SBC_ABS | VBIT,
  OPC_SBC_ABS_Z      = OPC_SBC_ABS | ZBIT,
  OPC_SBC_ABS_ALL    = OPC_SBC_ABS | ALLBITS,

  OPC_SBCSAFE_ABS    = OPC_SBC_ABS + BLOCKSIZE,
  OPC_SBCSAFE_ABS_C  = OPC_SBCSAFE_ABS | CBIT,
  OPC_SBCSAFE_ABS_N  = OPC_SBCSAFE_ABS | NBIT,
  OPC_SBCSAFE_ABS_V  = OPC_SBCSAFE_ABS | VBIT,
  OPC_SBCSAFE_ABS_Z  = OPC_SBCSAFE_ABS | ZBIT,
  OPC_SBCSAFE_ABS_ALL= OPC_SBCSAFE_ABS | ALLBITS,

  OPC_SBC_ABS_X      = OPC_SBCSAFE_ABS + BLOCKSIZE,
  OPC_SBC_ABS_X_C    = OPC_SBC_ABS_X | CBIT,
  OPC_SBC_ABS_X_N    = OPC_SBC_ABS_X | NBIT,
  OPC_SBC_ABS_X_V    = OPC_SBC_ABS_X | VBIT,
  OPC_SBC_ABS_X_Z    = OPC_SBC_ABS_X | ZBIT,
  OPC_SBC_ABS_X_ALL  = OPC_SBC_ABS_X | ALLBITS,

  OPC_SBCSAFE_ABS_X     = OPC_SBC_ABS_X + BLOCKSIZE,
  OPC_SBCSAFE_ABS_X_C   = OPC_SBCSAFE_ABS_X | CBIT,
  OPC_SBCSAFE_ABS_X_N   = OPC_SBCSAFE_ABS_X | NBIT,
  OPC_SBCSAFE_ABS_X_V   = OPC_SBCSAFE_ABS_X | VBIT,
  OPC_SBCSAFE_ABS_X_Z   = OPC_SBCSAFE_ABS_X | ZBIT,
  OPC_SBCSAFE_ABS_X_ALL = OPC_SBCSAFE_ABS_X | ALLBITS,

  OPC_SBC_ABS_Y      = OPC_SBCSAFE_ABS_X + BLOCKSIZE,
  OPC_SBC_ABS_Y_C    = OPC_SBC_ABS_Y | CBIT,
  OPC_SBC_ABS_Y_N    = OPC_SBC_ABS_Y | NBIT,
  OPC_SBC_ABS_Y_V    = OPC_SBC_ABS_Y | VBIT,
  OPC_SBC_ABS_Y_Z    = OPC_SBC_ABS_Y | ZBIT,
  OPC_SBC_ABS_Y_ALL  = OPC_SBC_ABS_Y | ALLBITS,

  OPC_SBCSAFE_ABS_Y     = OPC_SBC_ABS_Y + BLOCKSIZE,
  OPC_SBCSAFE_ABS_Y_C   = OPC_SBCSAFE_ABS_Y | CBIT,
  OPC_SBCSAFE_ABS_Y_N   = OPC_SBCSAFE_ABS_Y | NBIT,
  OPC_SBCSAFE_ABS_Y_V   = OPC_SBCSAFE_ABS_Y | VBIT,
  OPC_SBCSAFE_ABS_Y_Z   = OPC_SBCSAFE_ABS_Y | ZBIT,
  OPC_SBCSAFE_ABS_Y_ALL = OPC_SBCSAFE_ABS_Y | ALLBITS,

  OPC_SBC_IND        = OPC_SBCSAFE_ABS_Y + BLOCKSIZE,
  OPC_SBC_IND_C      = OPC_SBC_IND | CBIT,
  OPC_SBC_IND_N      = OPC_SBC_IND | NBIT,
  OPC_SBC_IND_V      = OPC_SBC_IND | VBIT,
  OPC_SBC_IND_Z      = OPC_SBC_IND | ZBIT,
  OPC_SBC_IND_ALL    = OPC_SBC_IND | ALLBITS,

  OPC_SBC_IND_X      = OPC_SBC_IND + BLOCKSIZE,
  OPC_SBC_IND_X_C    = OPC_SBC_IND_X | CBIT,
  OPC_SBC_IND_X_N    = OPC_SBC_IND_X | NBIT,
  OPC_SBC_IND_X_V    = OPC_SBC_IND_X | VBIT,
  OPC_SBC_IND_X_Z    = OPC_SBC_IND_X | ZBIT,
  OPC_SBC_IND_X_ALL  = OPC_SBC_IND_X | ALLBITS,

  OPC_SBC_IND_Y      = OPC_SBC_IND_X + BLOCKSIZE,
  OPC_SBC_IND_Y_C    = OPC_SBC_IND_Y | CBIT,
  OPC_SBC_IND_Y_N    = OPC_SBC_IND_Y | NBIT,
  OPC_SBC_IND_Y_V    = OPC_SBC_IND_Y | VBIT,
  OPC_SBC_IND_Y_Z    = OPC_SBC_IND_Y | ZBIT,
  OPC_SBC_IND_Y_ALL  = OPC_SBC_IND_Y | ALLBITS,

  OPC_SEC            = OPC_SBC_IND_Y + BLOCKSIZE,

  OPC_SED            = OPC_SEC + BLOCKSIZE,

  OPC_SEI            = OPC_SED + BLOCKSIZE,

  OPC_STA_ZP         = OPC_SEI + BLOCKSIZE,

  OPC_STA_ZP_X       = OPC_STA_ZP + BLOCKSIZE,

  OPC_STA_ABS        = OPC_STA_ZP_X + BLOCKSIZE,

  OPC_STA_ABS_X      = OPC_STA_ABS + BLOCKSIZE,

  OPC_STA_ABS_Y      = OPC_STA_ABS_X + BLOCKSIZE,

  OPC_STA_IND        = OPC_STA_ABS_Y + BLOCKSIZE,

  OPC_STA_IND_X      = OPC_STA_IND + BLOCKSIZE,

  OPC_STA_IND_Y      = OPC_STA_IND_X + BLOCKSIZE,

  OPC_STX_ZP         = OPC_STA_IND_Y + BLOCKSIZE,

  OPC_STX_ZP_Y       = OPC_STX_ZP + BLOCKSIZE,

  OPC_STX_ABS        = OPC_STX_ZP_Y + BLOCKSIZE,

  OPC_STY_ZP         = OPC_STX_ABS + BLOCKSIZE,

  OPC_STY_ZP_X       = OPC_STY_ZP + BLOCKSIZE,

  OPC_STY_ABS        = OPC_STY_ZP_X + BLOCKSIZE,

  OPC_STZ_ZP         = OPC_STY_ABS + BLOCKSIZE,

  OPC_STZ_ZP_X       = OPC_STZ_ZP + BLOCKSIZE,

  OPC_STZ_ABS        = OPC_STZ_ZP_X + BLOCKSIZE,

  OPC_STZ_ABS_X      = OPC_STZ_ABS + BLOCKSIZE,

  OPC_TAX            = OPC_STZ_ABS_X + BLOCKSIZE,
  OPC_TAX_N          = OPC_TAX | NBIT,
  OPC_TAX_Z          = OPC_TAX | ZBIT,
  OPC_TAX_ALL        = OPC_TAX | ALLBITS,

  OPC_TAY            = OPC_TAX + BLOCKSIZE,
  OPC_TAY_N          = OPC_TAY | NBIT,
  OPC_TAY_Z          = OPC_TAY | ZBIT,
  OPC_TAY_ALL        = OPC_TAY | ALLBITS,

  OPC_TRB_ZP         = OPC_TAY + BLOCKSIZE,
  OPC_TRB_ZP_ALL     = OPC_TRB_ZP | ALLBITS,

  OPC_TRB_ABS        = OPC_TRB_ZP + BLOCKSIZE,
  OPC_TRB_ABS_ALL    = OPC_TRB_ABS | ALLBITS,

  OPC_TSB_ZP         = OPC_TRB_ABS + BLOCKSIZE,
  OPC_TSB_ZP_ALL     = OPC_TSB_ZP | ALLBITS,

  OPC_TSB_ABS        = OPC_TSB_ZP + BLOCKSIZE,
  OPC_TSB_ABS_ALL    = OPC_TSB_ABS | ALLBITS,

  OPC_TSX            = OPC_TSB_ABS + BLOCKSIZE,
  OPC_TSX_N          = OPC_TSX | NBIT,
  OPC_TSX_Z          = OPC_TSX | ZBIT,
  OPC_TSX_ALL        = OPC_TSX | ALLBITS,

  OPC_TXA            = OPC_TSX + BLOCKSIZE,
  OPC_TXA_N          = OPC_TXA | NBIT,
  OPC_TXA_Z          = OPC_TXA | ZBIT,
  OPC_TXA_ALL        = OPC_TXA | ALLBITS,

  OPC_TXS            = OPC_TXA + BLOCKSIZE,

  OPC_TYA            = OPC_TXS + BLOCKSIZE,
  OPC_TYA_N          = OPC_TYA | NBIT,
  OPC_TYA_Z          = OPC_TYA | ZBIT,
  OPC_TYA_ALL        = OPC_TYA | ALLBITS,

  OPC_TRAPTHIS       = OPC_TYA + BLOCKSIZE,

  OPC_CARDTRAP       = OPC_TRAPTHIS + BLOCKSIZE
} Opcode;

#endif  /* not __OPCODES_H__ */
