/* This program takes a .a2disk is DOS 3.3 format and converts it to a disk
 * in nibbleized (raw) format.  This is sometimes necessary when programs
 * on the drive access the drive hardware directly.  This program will no
 * longer be necessary in a future release of zaniWok.
 */


#include <stdio.h>
#include <stdlib.h>


#define GAP2LENGTH 7
#define GAP3LENGTH 19
#define RAW_TRACK_LENGTH ((342 + 7 + GAP2LENGTH + GAP3LENGTH + 14) * 16)

static void convert_to_raw (unsigned char nice[258],
			    unsigned char raw[343]);


static inline int
four_and_four_encode (int val)
{
  return 0xAAAA | val | (val << 7);
}


static inline void
put2 (int n, FILE *fp)
{
  putc (n >> 8, fp);
  putc (n & 0xFF, fp);
}




int
main (int argc, char *argv[])
{
  FILE *in, *out;
  unsigned long header, first_track_offset;
  int track, sector;
  
  if (argc != 3)
    {
      fprintf (stderr, "Usage: %s <indisk> <outdisk>\n", argv[0]);
      exit (-1);
    }

  in = fopen (argv[1], "rb");
  if (in == NULL)
    {
      fprintf (stderr, "Unable to read file \"%s\".\n", argv[1]);
      exit (-9);
    }

  header = ((getc (in) << 24) | (getc (in) << 16)
	    | (getc (in) << 8) | getc (in));
  if (header != 0x04136359)
    {
      fprintf (stderr, "Missing magic header word.  Aborting.\n");
      exit (-2);
    }

  first_track_offset = ((getc (in) << 24) | (getc (in) << 16)
			| (getc (in) << 8) | getc (in));

  if (getc (in) != 2)
    {
      fprintf (stderr, "Not a DOS 3.3 disk.  Aborting.\n");
      exit (-3);
    }

  out = fopen (argv[2], "wb");
  if (out == NULL)
    {
      fprintf (stderr, "Unable to open file \"%s\" for writing.\n");
      exit (-15);
    }

  /* Output header. */
  putc (0x04, out);
  putc (0x13, out);
  putc (0x63, out);
  putc (0x59, out);
  
  /* Output offset to data. */
  putc (0, out);
  putc (0, out);
  putc (0, out);
  putc (9, out);

  /* Output disk type. */
  putc (3, out);

  /* Output raw data for each track. */
  for (track = 0; track < 35; track++)
    {
      /* Output qtrack #. */
      putc (track * 4, out);
      
      /* Output track length. */
      putc (RAW_TRACK_LENGTH >> 8, out);
      putc (RAW_TRACK_LENGTH & 0xFF, out);

      for (sector = 0; sector < 16; sector++)
	{
	  unsigned char nice[258], raw[343];
	  int i, v, t, s;
	  static const int physical_to_logical[] = {
	    0x0, 0x7, 0xE, 0x6, 0xD, 0x5, 0xC, 0x4,
	    0xB, 0x3, 0xA, 0x2, 0x9, 0x1, 0x8, 0xF
	    };

	  /* Read in the sector data and convert it. */
	  fseek (in, (first_track_offset + track * 16 * 256
		      + physical_to_logical[sector] * 256),
		 SEEK_SET);
	  fread (nice, 1, 256, in);
	  convert_to_raw (nice, raw);

	  /* Output address field. */
	  putc (0xD5, out), putc (0xAA, out), putc (0x96, out);  /* prologue */
	  v = four_and_four_encode (254);
	  t = four_and_four_encode (track);
	  s = four_and_four_encode (sector);
	  put2 (v, out);
	  put2 (t, out);
	  put2 (s, out);
	  put2 (v ^ t ^ s, out);
	  putc (0xDE, out), putc (0xAA, out), putc (0xEB, out);  /* epilogue */

	  /* Output sync bytes. */
	  for (i = 0; i < GAP2LENGTH; i++)
	    putc (0xFF, out);
	  
	  /* Output data field prologue. */
	  putc (0xD5, out), putc (0xAA, out), putc (0xAD, out);

	  /* Output the main data. */
	  fwrite (raw, 1, 343, out);

	  /* Output data field epilogue. */
	  putc (0xDE, out), putc (0xAA, out), putc (0xEB, out);

	  /* Output trailing sync bytes. */
	  for (i = 0; i < GAP3LENGTH; i++)
	    putc (0xFF, out);
	}
    }

  fclose (in);
  fclose (out);

  return EXIT_SUCCESS;
}


/* Lookup tables to help us encode the data. */

static const unsigned char reverse_bit_pairs[] = {
  0x00, 0x02, 0x01, 0x03, 0x08, 0x0A, 0x09, 0x0B,
  0x04, 0x06, 0x05, 0x07, 0x0C, 0x0E, 0x0D, 0x0F,
  0x20, 0x22, 0x21, 0x23, 0x28, 0x2A, 0x29, 0x2B,
  0x24, 0x26, 0x25, 0x27, 0x2C, 0x2E, 0x2D, 0x2F,
  0x10, 0x12, 0x11, 0x13, 0x18, 0x1A, 0x19, 0x1B,
  0x14, 0x16, 0x15, 0x17, 0x1C, 0x1E, 0x1D, 0x1F,
  0x30, 0x32, 0x31, 0x33, 0x38, 0x3A, 0x39, 0x3B,
  0x34, 0x36, 0x35, 0x37, 0x3C, 0x3E, 0x3D, 0x3F,
};

static const unsigned char write_translate_table[] = {
  0x96, 0x97, 0x9A, 0x9B, 0x9D, 0x9E, 0x9F, 0xA6,
  0xA7, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB2, 0xB3,
  0xB4, 0xB5, 0xB6, 0xB7, 0xB9, 0xBA, 0xBB, 0xBC,
  0xBD, 0xBE, 0xBF, 0xCB, 0xCD, 0xCE, 0xCF, 0xD3,
  0xD6, 0xD7, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE,
  0xDF, 0xE5, 0xE6, 0xE7, 0xE9, 0xEA, 0xEB, 0xEC,
  0xED, 0xEE, 0xEF, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6,
  0xF7, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF,
};


static void
convert_to_raw (unsigned char nice[258], unsigned char raw[343])
{
  const unsigned char *in;
  unsigned char *out;
  int i;

  nice[256] = nice[257] = 0;

  /* Do 6 bit sequences. */
  for (i = 255, in = &nice[0], out = &raw[86]; i >= 0; i--)
    *out++ = *in++ >> 2;

  /* Do 2 bit munging. */
  for (i = 85, in = &nice[0], out = &raw[0]; i >= 0; in++, i--)
    *out++ = reverse_bit_pairs[(in[0] & 3) | ((in[86] & 3) << 2)
			       | ((in[86 * 2] & 3) << 4)];
  
  /* Do xors and calculate checksum. */
  raw[342] = 0;
  for (i = 341, out = &raw[342]; i >= 0; out--, i--)
    out[0] ^= out[-1];

  /* Encode everything. */
  for (i = 342, out = &raw[0]; i >= 0; i--)
    *out++ = write_translate_table[*out];
}
